﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.CheckoutHandlers.ChargeLogicConnect.chargelogic.transact;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.Security.UserManagement;
using System;
using Order = Dynamicweb.Ecommerce.Orders.Order;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.ChargeLogicConnect
{
    /// <summary>
    /// ChargeLogicConnect Checkout Handler
    /// </summary>
    [
        AddInName("ChargeLogic Connect"),
        AddInDescription("ChargeLogic Connect Hosted Payment Frame")
    ]
    public class ChargeLogicConnect : CheckoutHandler, ISavedCard
    {
        private const string ApplicationNo = "DYNAMICWEB";
        private const string ApplicationVersion = "1.0";

        private enum TransactionType
        {
            Authorize,
            Capture,
            Charge
        }

        #region Addin parameters

        [AddInParameter("Store No./Username"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string StoreNo { get; set; }

        [AddInParameter("API Key/Password"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string APIKey { get; set; }

        [AddInParameter("Merchant Resource file"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string MerchantResource { get; set; }

        [AddInParameter("Require CVV"), AddInParameterEditor(typeof(YesNoParameterEditor), ""), AddInDescription("")]
        public bool RequireCVV { get; set; }

        [AddInParameter("Hosted payment URL"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
        public string HostedPaymentURL { get; set; }

        [AddInParameter("Payment template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/ChargeLogic/Payment")]
        public string PaymentTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/ChargeLogic/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/ChargeLogic/Error")]
        public string ErrorTemplate { get; set; }

        #endregion

        #region Form properties

        private string MerchantResourceURL()
        {
            return string.Format("https://{0}{1}", Context.Current.Request.Url.Host, MerchantResource);
        }

        private string getCardTokenName(Order order)
        {
            return Converter.ToString(Context.Current.Session[string.Format("ChargeLogic.CardTokenName:{0}", order.Id)]);
        }

        private void setCardTokenName(Order order, string tokenName)
        {
            Context.Current.Session[string.Format("ChargeLogic.CardTokenName:{0}", order.Id)] = tokenName;
        }

        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="ChargeLogicConnect" /> class.
        /// </summary>
        public ChargeLogicConnect()
        {
            MerchantResource = "/Admin/Public/Gateways/ChargeLogic/MerchantResource.html";
            HostedPaymentURL = "https://connect.chargelogic.com";

            PaymentTemplate = "eCom7/CheckoutHandler/ChargeLogic/Payment/Payment.html";
            CancelTemplate = "eCom7/CheckoutHandler/ChargeLogic/Cancel/checkouthandler_cancel.html";
            ErrorTemplate = "eCom7/CheckoutHandler/ChargeLogic/Error/checkouthandler_error.html";
        }

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");

                return RenderPaymentFrame(order);
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order, ex.Message);
            }
        }

        private string RenderPaymentFrame(Order order)
        {
            string hostedPaymentID = SetupHostedPayment(order, TransactionType.Authorize);

            var formTemplate = new Template(PaymentTemplate);

            formTemplate.SetTag("ChargeLogic.HostedPaymentURL", HostedPaymentURL);
            formTemplate.SetTag("ChargeLogic.HostedPaymentID", hostedPaymentID);

            // Draft saved card name
            if (order.DoSaveCardToken || !string.IsNullOrWhiteSpace(order.SavedCardDraftName))
            {
                setCardTokenName(order, string.IsNullOrWhiteSpace(order.SavedCardDraftName) ? order.Id : order.SavedCardDraftName);
            }

            return Render(order, formTemplate);
        }

        /// <summary>
        /// Redirects user to payment step or canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(order, "Redirected to ChargeLogicConnect CheckoutHandler");

                string action = Dynamicweb.Context.Current.Request["?Action"];

                switch (action)
                {
                    case "Approve":
                        return StateOk(order);

                    case "SaveToken":
                        return StateSaveToken(order);

                    default:
                        return StateCancel(order);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order, ex.Message);
            }
        }

        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");

            if (string.Compare(order.GatewayResult, Dynamicweb.Context.Current.Request["ConfirmationID"], true) != 0)
            {
                throw new Exception(string.Format("ConfirmationID is wrong: {0}", Converter.ToString(Context.Current.Request["ConfirmationID"])));
            }

            order.GatewayPaymentStatus = GatewayPaymentStatus.Accepted;
            order.TransactionCardNumber = HideCardNumber(Converter.ToString(Context.Current.Request["MaskedAccountNumber"]));
            order.TransactionCardType = Converter.ToString(Context.Current.Request["AccountType"]);
            order.TransactionAmount = order.Price.Price;
            order.TransactionStatus = "Succeeded";

            string cardName = getCardTokenName(order);
            if (!string.IsNullOrEmpty(cardName))
            {
                var user = User.GetCurrentUser(PagePermissionLevels.Frontend);
                if (user != null)
                {
                    Services.PaymentCard.CreatePaymentCard(user.ID, order.PaymentMethodId, order.LanguageId, cardName, order.TransactionCardType, order.TransactionCardNumber, Context.Current.Request["Token"]);
                    //set card Id to order???
                }
                else
                {
                    order.TransactionToken = Context.Current.Request["Token"];
                    Services.Orders.Save(order);
                }
            }

            if (!order.Complete)
            {
                SetOrderComplete(order);
                CheckoutDone(order);
            }

            RedirectToCart(order);
            return null;
        }

        private string StateSaveToken(Order order)
        {
            setCardTokenName(order, Converter.ToString(Context.Current.Request["CardTokenName"]));
            return null;
        }

        private void UseSavedCardInternal(Order order, int tokenId, string cardholderName, string cardCVV)
        {
            if (tokenId == 0)
            {
                throw new Exception("Token id is 0.");
            }

            var token = Services.PaymentCard.GetById(tokenId);
            if (token == null || order.CustomerAccessUserId != token.UserID)
            {
                throw new Exception("Token is incorrect.");
            }

            if (string.IsNullOrEmpty(order.GatewayResult))
            {
                string orderGuid = Guid.NewGuid().ToString();
                order.GatewayResult = orderGuid;
                Services.Orders.UpdateGatewayResult(order, false);
            }

            Credential credential = PrepareCredential(order);
            Transaction transaction = PrepareTransaction(order);
            Address billingAddress = PrepareBillingAddress(order);

            var card = new Card();
            card.Token = token.Token;
            card.CardholderName = cardholderName;

            if (RequireCVV)
            {
                card.CardVerificationValue = cardCVV;
            }

            var response = new Response();
            var api = new EFT_API_2_PortClient();
            api.ClientCredentials.UserName.UserName = StoreNo;
            api.ClientCredentials.UserName.Password = APIKey;

            order.TransactionNumber = api.CreditCardAuthorize(credential, card, transaction, billingAddress, ref response);

            if (response.TransactionStatus == "Approved")
            {
                order.GatewayPaymentStatus = GatewayPaymentStatus.Accepted;
                order.TransactionCardNumber = token.Identifier;
                order.TransactionCardType = token.CardType;
                order.TransactionAmount = order.Price.Price;
                order.TransactionStatus = "Succeeded";

                if (!order.Complete)
                {
                    SetOrderComplete(order);
                    CheckoutDone(order);
                }

                RedirectToCart(order);
            }
            else if (!string.IsNullOrEmpty(response.Message))
            {
                throw new Exception(string.Format("Problem with saved card: {0}", response.Message));
            }
            else
            {
                throw new Exception(string.Format("Internal error. TransactionStatus = {0}", response.TransactionStatus));
            }
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            return this.Render(order, cancelTemplate);
        }

        private string PrintErrorTemplate(Order order, string msg)
        {
            order.TransactionAmount = 0;
            order.TransactionStatus = "Failed";
            order.Errors.Add(msg);
            Services.Orders.Save(order);

            Services.Orders.DowngradeToCart(order);
            order.CartV2StepIndex = 0;
            order.TransactionStatus = "";
            Common.Context.SetCart(order);

            if (string.IsNullOrWhiteSpace(this.ErrorTemplate))
            {
                RedirectToCart(order);
            }

            var errorTemplate = new Template(ErrorTemplate);
            errorTemplate.SetTag("CheckoutHandler:ErrorMessage", msg);

            return Render(order, errorTemplate);
        }

        private string SetupHostedPayment(Order order, TransactionType transactionType)
        {
            string orderGuid = Guid.NewGuid().ToString();
            order.GatewayResult = orderGuid;
            Services.Orders.UpdateGatewayResult(order, false);

            Credential credential = PrepareCredential(order);
            Transaction transaction = PrepareTransaction(order);
            Address billingAddress = PrepareBillingAddress(order);
            Address shippingAddress = PrepareShippingAddress(order);

            var hostedPayment = new HostedPayment();
            hostedPayment.Language = "ENG";
            hostedPayment.RequireCVV = RequireCVV ? "Yes" : "No";
            hostedPayment.Embedded = "Yes";
            hostedPayment.MerchantResourceURL = MerchantResourceURL();
            hostedPayment.ConfirmationID = orderGuid;
            hostedPayment.ReturnURL = this.GetBaseUrl(order) + "&redirect=false&";

            string рostedPaymentID;
            var api = new EFT_API_2_PortClient();
            api.ClientCredentials.UserName.UserName = StoreNo;
            api.ClientCredentials.UserName.Password = APIKey;

            switch (transactionType)
            {
                case TransactionType.Authorize:
                    рostedPaymentID = api.SetupHostedCreditCardAuthorize(credential, transaction, billingAddress, shippingAddress, hostedPayment);
                    break;

                case TransactionType.Capture:
                    рostedPaymentID = api.SetupHostedCreditCardCapture(credential, transaction, billingAddress, shippingAddress, hostedPayment);
                    break;

                case TransactionType.Charge:
                    рostedPaymentID = api.SetupHostedCreditCardCharge(credential, transaction, billingAddress, shippingAddress, hostedPayment);
                    break;

                default:
                    throw new Exception("Unknown transaction type");
            }

            return рostedPaymentID;
        }

        #region ChargeLogicHelpers

        private Credential PrepareCredential(Order order)
        {
            var credential = new Credential();
            credential.StoreNo = StoreNo;
            credential.APIKey = APIKey;
            credential.ApplicationNo = ApplicationNo;
            credential.ApplicationVersion = ApplicationVersion;

            return credential;
        }

        private Transaction PrepareTransaction(Order order)
        {
            Func<double, string> formatPrice = (x) =>
            {
                int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
                return Math.Round(x, decimals).ToString("0.00", System.Globalization.CultureInfo.CreateSpecificCulture("en-US"));
            };

            var transaction = new Transaction();
            transaction.Amount = formatPrice(order.Price.Price);
            transaction.ConfirmationID = order.GatewayResult;
            transaction.Currency = order.Currency.Code;
            transaction.ExternalReferenceNumber = order.Id;

            return transaction;
        }

        private Address PrepareBillingAddress(Order order)
        {
            var billingAddress = new Address();
            billingAddress.StreetAddress = order.CustomerAddress;
            billingAddress.StreetAddress2 = order.CustomerAddress2;
            billingAddress.Name = string.IsNullOrEmpty(order.CustomerName) ? order.CustomerFirstName + " " + order.CustomerSurname : order.CustomerName;
            billingAddress.PhoneNumber = order.CustomerPhone;
            billingAddress.PostCode = order.CustomerZip;
            billingAddress.State = order.CustomerRegion;
            billingAddress.City = order.CustomerCity;
            billingAddress.Country = order.CustomerCountryCode;
            billingAddress.Email = order.CustomerEmail;

            return billingAddress;
        }

        private Address PrepareShippingAddress(Order order)
        {
            Address shippingAddress;
            bool isDeliveryFieldsFilled = !string.IsNullOrEmpty(order.DeliveryAddress) && !string.IsNullOrEmpty(order.DeliveryZip);
            if (isDeliveryFieldsFilled)
            {
                shippingAddress = new Address();
                shippingAddress.StreetAddress = order.DeliveryAddress;
                shippingAddress.StreetAddress2 = order.DeliveryAddress2;
                shippingAddress.Name = string.IsNullOrEmpty(order.DeliveryName) ? order.DeliveryFirstName + " " + order.DeliverySurname : order.DeliveryName;
                shippingAddress.PhoneNumber = order.DeliveryPhone;
                shippingAddress.PostCode = order.DeliveryZip;
                shippingAddress.State = order.DeliveryRegion;
                shippingAddress.City = order.DeliveryCity;
                shippingAddress.Country = order.DeliveryCountryCode;
                shippingAddress.Email = order.DeliveryEmail;
            }
            else
            {
                shippingAddress = PrepareBillingAddress(order);
            }

            return shippingAddress;
        }

        #endregion

        #region ISavedCard interface

        /// <remarks>Not supported</remarks>
        public void DeleteSavedCard(int savedCardID) { }

        /// <summary>
        /// Directs checkout handler to use saved card
        /// </summary>
        /// <param name="order">Order that should be processed using saved card information</param>
        /// <returns>Empty string, if operation succeeded, otherwise string template with exception mesage</returns>
        public string UseSavedCard(Order order)
        {
            try
            {
                var cardHolder = string.IsNullOrEmpty(order.CustomerName) ? order.CustomerFirstName + " " + order.CustomerSurname : order.CustomerName;
                UseSavedCardInternal(order, order.SavedCardId, cardHolder, string.Empty);

                return string.Empty;
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogEvent(order, ex.Message, DebuggingInfoType.UseSavedCard);
                return PrintErrorTemplate(order, ex.Message);
            }
        }

        /// <summary>
        /// Shows if order supports saving card
        /// </summary>
        /// <param name="order">Instance of order</param>
        /// <returns>True, if saving card is supported</returns>
        public bool SavedCardSupported(Orders.Order order)
        {
            return true;
        }

        #endregion 
    }
}
