﻿using Dynamicweb.Core;
using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Xml;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.AltaPay
{
    /// <summary>
    /// Wrapper around HttpClient which helps to form alta pay request and send it to the server
    /// </summary>
    internal class AltaPayRequest
    {
        /// <summary>
        /// Parameters for x-www-form-urlencoded format which is used in AltaPay
        /// </summary>
        public Dictionary<string, string> Parameters { get; set; } = new Dictionary<string, string>();

        /// <summary>
        /// Contains the user name, password, gateway address and path to the API
        /// </summary>
        public RequestConfiguration Configuration { get; set; }

        /// <summary>
        /// Set parameter value
        /// </summary>
        /// <param name="key">Key</param>
        /// <param name="value">Value</param>
        public void SetParameter(string key, object value)
        {
            Parameters[key] = value?.ToString() ?? string.Empty;
        }

        /// <summary>
        /// Set field value of some complex parameter
        /// </summary>
        /// <param name="parameterName">Parameter name</param>
        /// <param name="fieldName">Field of parameter</param>
        /// <param name="value">Field value</param>
        public void SetField(string parameterName, string fieldName, object value)
        {
            string key = $"{parameterName}[{fieldName}]";
            SetParameter(key, value);
        }

        /// <summary>
        /// Set field value of some parameter which is an element of array
        /// </summary>
        /// <param name="parameterName">Parameter name</param>
        /// <param name="fieldName">Field of array item</param>
        /// <param name="value">Value of array item</param>
        /// <param name="index">Index of the array item</param>
        public void SetArrayField(string parameterName, string fieldName, object value, int index)
        {
            string key = $"{parameterName}[{index}][{fieldName}]";
            SetParameter(key, value);
        }

        /// <summary>
        /// Send the request to the AltaPay API
        /// </summary>
        /// <param name="methodName">API method name</param>
        /// <returns>The server response presented as wrapper around xml-document</returns>
        public AltaPayResponse SendRequest(string methodName)
        {
            Ensure.NotNullOrEmpty(Configuration.UserName, "User name should not be empty");
            Ensure.NotNullOrEmpty(Configuration.Password, "Password should not be empty");

            using (var messageHandler = GetMessageHandler())
            {
                using (var client = new HttpClient(messageHandler))
                {
                    string authenticationScheme = "Basic";
                    string authenticationParameter = Convert.ToBase64String(Encoding.Default.GetBytes($"{Configuration.UserName}:{Configuration.Password}"));
                    client.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue(authenticationScheme, authenticationParameter);    
                    client.BaseAddress = new Uri(Configuration.Gateway);

                    var content = new FormUrlEncodedContent(Parameters);
                    string apiCommand = Configuration.ApiPath + methodName;
                    using (var response = client.PostAsync(apiCommand, content).GetAwaiter().GetResult())
                    {
                        string responseText = response.Content.ReadAsStringAsync().GetAwaiter().GetResult();
                        var responseDocument = new XmlDocument();
                        responseDocument.LoadXml(responseText);

                        return new AltaPayResponse(responseDocument);
                    }
                }
            }
        }

        private HttpMessageHandler GetMessageHandler()
        {
            return new HttpClientHandler()
            {
                AutomaticDecompression = DecompressionMethods.Deflate | DecompressionMethods.GZip
            };
        }
    }
}
