﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Security.Cryptography;
using System.Text;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.AuthorizeNet
{
    /// <summary>
    /// AuthorizeNet Checkout Handler
    /// </summary>
    [
        AddInName("AuthorizeNet"),
        AddInDescription("AuthorizeNet Checkout Handler")
    ]
    public class AuthorizeNet : CheckoutHandlerWithStatusPage, IDropDownOptions, IRemotePartialFinalOnlyCapture
    {
        private enum PostModes { Auto, Template }
        private enum WorkModes { Developer, Test, Production }
        private enum ResponseCodes { Approved = 1, Declined = 2, Error = 3, Held = 4 }

        private PostModes postMode;
        private WorkModes workMode;
        private char delimeterChar;
        private string transactionType;

        #region Addin parameters

        [AddInParameter("API login ID"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string ApiLogin { get; set; }

        [AddInParameter("Transaction key"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string TransactionKey { get; set; }

        [AddInParameter("Signature key"), AddInParameterEditor(typeof(TextParameterEditor), "TextArea=true")]
        public string SignatureKey { get; set; }

        [AddInParameter("The type of credit card transaction"), AddInParameterEditor(typeof(RadioParameterEditor), "")]
        public string TransactionType
        {
            get { return transactionType; }
            set { transactionType = value; }
        }

        [AddInParameter("Merchant hash"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string MerchantHash { get; set; }

        [AddInParameter("Post mode"), AddInParameterEditor(typeof(DropDownParameterEditor), "none=false; SortBy=Value;")]
        public string PostModeSelection
        {
            get
            {
                return postMode.ToString();
            }
            set
            {
                switch (value)
                {
                    case "Auto":
                        postMode = PostModes.Auto;
                        break;

                    case "Template":
                        postMode = PostModes.Template;
                        break;
                }
            }
        }

        [AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Post")]
        public string PostTemplate { get; set; }

        [AddInParameter("Relay response template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Relay")]
        public string RelayTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Work Mode"), AddInParameterEditor(typeof(RadioParameterEditor), "")]
        public string WorkMode
        {
            get
            {
                return workMode.ToString();
            }
            set
            {
                switch (value)
                {
                    case "Developer":
                        workMode = WorkModes.Developer;
                        break;

                    case "Test":
                        workMode = WorkModes.Test;
                        break;

                    case "Production":
                        workMode = WorkModes.Production;
                        break;
                }
            }
        }

        #endregion

        #region Form properties

        private static string DefaultUrl()
        {
            var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);

            return string.Format("{1}://{0}{2}/Default.aspx", Context.Current.Request.Url.Host, Context.Current.Request.Url.Scheme, portString);
        }

        private static string BaseUrl(Order order, int pageId = 0)
        {
            string pageIdParam;
            if (pageId == 0)
            {
                pageIdParam = Dynamicweb.Context.Current.Request["ID"] == null ? string.Empty : string.Format("ID={0}&", Dynamicweb.Context.Current.Request["ID"]);
            }
            else
            {
                pageIdParam = string.Format("ID={0}&", pageId);
            }

            return string.Format("{0}?{1}{3}={2}", DefaultUrl(), pageIdParam, order.Id, OrderIdRequestName);
        }

        private static string CancelUrl(Order order)
        {
            return string.Format("{0}&AuthState=Cancel", BaseUrl(order));
        }

        private static string CallbackUrl(Order order)
        {
            var portString = Context.Current.Request.Url.IsDefaultPort ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);

            return string.Format("{1}://{0}{2}/Admin/Public/Gateways/CheckoutHandler/Authorizenet/AuthorizenetFormCallBack.aspx", Context.Current.Request.Url.Host, Context.Current.Request.Url.Scheme, portString);
        }

        private static string AcceptUrl(Order order, int pageId)
        {
            return string.Format("{0}&AuthState=Accept", BaseUrl(order, pageId));
        }

        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthorizeNet" /> class
        /// </summary>
        public AuthorizeNet()
        {
            ServicePointManager.SecurityProtocol = ServicePointManager.SecurityProtocol | SecurityProtocolType.Tls12;
            HasServerCallback = true;
            TransactionType = "AUTH_CAPTURE";
            delimeterChar = '|';
            postMode = PostModes.Auto;
        }

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            string sequence = CryptoSha.GenerateSequence();
            int timeStamp = CryptoSha.GenerateTimestamp();
            int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == string.Empty ? 2 : order.Currency.Rounding.Decimals;
            string amount = Math.Round(order.Price.Price, decimals).ToString("0.00", System.Globalization.CultureInfo.CreateSpecificCulture("en-US"));
            string fingerPrint = string.IsNullOrEmpty(SignatureKey) ?
                Crypto.GenerateFingerprint(TransactionKey, ApiLogin, amount, sequence, timeStamp.ToString()) :
                CryptoSha.GenerateFingerprint(SignatureKey, ApiLogin, amount, sequence, timeStamp.ToString());
            string formAction = string.Empty;
            string testMode = string.Empty;

            switch (workMode)
            {
                case WorkModes.Developer:
                    formAction = "https://test.authorize.net/gateway/transact.dll";
                    testMode = "0";
                    break;

                case WorkModes.Test:
                    formAction = "https://secure2.authorize.net/gateway/transact.dll";
                    testMode = "1";
                    break;

                case WorkModes.Production:
                    formAction = "https://secure2.authorize.net/gateway/transact.dll";
                    testMode = "0";
                    break;
            }

            bool isDeliveryFieldsFilled = !string.IsNullOrEmpty(order.DeliveryAddress) && !string.IsNullOrEmpty(order.DeliveryZip);
            string customerName = Converter.ToString(order.CustomerName);
            int delPos = customerName.IndexOf(' ');
            string firstName = delPos >= 0 ? customerName.Substring(0, delPos) : customerName;
            string lastName = delPos >= 0 ? customerName.Substring(delPos + 1) : string.Empty;
            string deliveryName = Converter.ToString(order.DeliveryName);
            delPos = deliveryName.IndexOf(' ');
            string deliveryFirstName = string.IsNullOrEmpty(order.DeliveryFirstName) ? (delPos >= 0 ? deliveryName.Substring(0, delPos) : deliveryName) : order.DeliveryFirstName;
            string deliveryLastName = string.IsNullOrEmpty(order.DeliverySurname) ? (delPos >= 0 ? deliveryName.Substring(delPos + 1) : "") : order.DeliverySurname;

            try
            {
                LogEvent(order, "Checkout started");

                if (order.CurrencyCode != "USD")
                {
                    throw new Exception($"Only USD currency is allowed. Order currency: {order.CurrencyCode}");
                }

                var formValues = new Dictionary<string, string> {
                    { "x_fp_hash", fingerPrint },
                    { "x_fp_sequence", sequence },
                    { "x_fp_timestamp", timeStamp.ToString() },
                    { "x_show_form", "PAYMENT_FORM" },
                    { "x_method", "CC" },
                    { "x_version", "3.1" },
                    { "x_login", ApiLogin },
                    { "x_amount", amount },
                    { "x_test_request", testMode },
                    { "x_invoice_num", order.Id },
                    { "x_description", $"Payment to {Context.Current.Request.Url.Host}" },
                    { "x_first_name", firstName },
                    { "x_last_name", lastName },
                    { "x_company", Converter.ToString(order.CustomerCompany) },
                    { "x_address", string.Format("{0} {1}", Converter.ToString(order.CustomerAddress), Converter.ToString(order.CustomerAddress2)) },
                    { "x_city", Converter.ToString(order.CustomerCity) },
                    { "x_state", Converter.ToString(order.CustomerRegion) },
                    { "x_zip", Converter.ToString(order.CustomerZip) },
                    { "x_country", Converter.ToString(order.CustomerCountry) },
                    {"x_ship_to_first_name", Converter.ToString(isDeliveryFieldsFilled ? deliveryFirstName : firstName)},
                    {"x_ship_to_last_name", Converter.ToString(isDeliveryFieldsFilled ? deliveryLastName : lastName)},
                    {"x_ship_to_company", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryCompany : order.CustomerCompany)},
                    {"x_ship_to_address", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryAddress : order.CustomerAddress) +" "+ Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryAddress2 : order.CustomerAddress2)},
                    {"x_ship_to_city", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryCity : order.CustomerCity)},
                    {"x_ship_to_state", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryRegion : order.CustomerRegion)},
                    {"x_ship_to_zip", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryZip : order.CustomerZip)},
                    {"x_ship_to_country", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryCountry : order.CustomerCountry)},
                    { "x_email", Converter.ToString(order.CustomerEmail) },
                    { "x_cancel_url", CancelUrl(order) },
                    { "x_delim_data", "0" },
                    { "x_relay_response", "1" },
                    { "x_relay_URL", CallbackUrl(order) },
                    { "x_type", TransactionType },
                                     }.ToDictionary(x => x.Key, x => x.Value);

                switch (postMode)
                {
                    case PostModes.Auto:
                        LogEvent(order, "Auto post to AuthorizeNet");
                        SubmitForm(formAction, formValues);
                        return string.Empty;

                    case PostModes.Template:
                        LogEvent(order, "Render template");

                        var formTemplate = new Template(PostTemplate);
                        formTemplate.SetTag("Authorizenet.x_action", formAction);

                        foreach (var formValue in formValues)
                        {
                            formTemplate.SetTag(string.Format("Authorizenet.{0}", formValue.Key), formValue.Value);
                        }

                        var renderer = new Dynamicweb.Ecommerce.Frontend.Renderer(Dynamicweb.Frontend.PageView.Current());
                        renderer.RenderOrder(order, formTemplate);
                        return formTemplate.Output();

                    default:
                        LogError(order, "Unhandled post mode: '{0}'", postMode);
                        throw new ArgumentException(string.Format("Unhandled post mode: '{0}'", postMode));
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order, ex.Message);
            }
            finally
            {
                order.GatewayResult = PrepareGatewayResult(Dynamicweb.Context.Current.Request["ID"]);
                order.UpdateGatewayResult(false);
            }
        }

        /// <summary>
        /// Redirects user to payment step or to canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(order, "Redirected to AuthorizeNet CheckoutHandler");

                switch (Converter.ToString(Context.Current.Request["AuthState"]))
                {
                    case "Accept":
                        return StateOk(order);

                    case "Cancel":
                        return StateCancel(order);

                    default:
                        return Callback(order);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order, ex.Message);
            }
        }

        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");

            if (!order.Complete)
            {
                SetOrderComplete(order);
                CheckoutDone(order);
            }

            RedirectToCart(order);
            return null;
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            order.Save();
            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(cancelTemplate, order, true);

            return cancelTemplate.Output();
        }

        private string Callback(Order order)
        {
            int pageID = 0;
            bool orderWasCompleted = order.Complete;
            ResponseCodes responseCode = ResponseCodes.Error;

            try
            {
                LogEvent(order, "Callback started");

                bool errorOccured = false;
                var request = Context.Current.Request;

                responseCode = (ResponseCodes)Converter.ToInt32(request["x_response_code"]);

                string amount = request["x_amount"];

                string transactionID = request["x_trans_id"];
                if (string.IsNullOrEmpty(transactionID))
                {
                    errorOccured = true;
                    LogEvent(order, "No transaction number sent to callback");
                }
                else if (transactionID == "0" && workMode == WorkModes.Production)
                {
                    errorOccured = true;
                    LogEvent(order, "Transaction number is 0, but gateway is in Production mode!");
                }

                if (!string.IsNullOrEmpty(SignatureKey))
                {
                    string shaHash = request["x_SHA2_Hash"];
                    if (!CryptoSha.IsMatch(request, SignatureKey, shaHash))
                    {
                        errorOccured = true;
                        LogEvent(order, "SHA hash do not match");
                    }
                }
                else
                {
                    string md5Hash = request["x_MD5_Hash"];
                    if (!Crypto.IsMatch(MerchantHash, ApiLogin, transactionID, amount, md5Hash))
                    {
                        errorOccured = true;
                        LogEvent(order, "MD5 sums do not match");
                    }
                }

                if (errorOccured)
                {
                    responseCode = ResponseCodes.Error;
                    LogError(order, "At least one callback validation error exists - exiting callback routine.");
                }
                else
                {
                    order.GatewayPaymentStatus = ConvertAuthorizeNetToDwStatus(responseCode);

                    if (order.GatewayPaymentStatus == GatewayPaymentStatus.Paid)
                    {
                        LogEvent(order, "Payment succeeded with transaction number: {0}", transactionID);
                        pageID = RetrievePageID(order.GatewayResult);

                        SetOrderCompleteSafety(order);

                        if (TransactionType == "AUTH_CAPTURE")
                        {
                            string infoTxt = "Capture succeeded";
                            LogEvent(order, infoTxt);
                            order.CaptureInfo = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, infoTxt);
                        }

                        SetTransactionState(order, true, order.Price.Price, transactionID, request["x_card_type"]);
                    }
                    else
                    {
                        LogEvent(order, "Payment failed with error {0}/{1} - {2}", request["x_response_code"], request["x_response_reason_code"], request["x_response_reason_text"]);
                        SetTransactionState(order, false, 0, "", "");
                    }
                }
            }
            finally
            {
                if (!orderWasCompleted && order.Complete)
                {
                    CheckoutDone(order, pageID);
                }
            }

            return PrintRelayTemplate(order, pageID, responseCode);
        }

        private new void SetOrderComplete(Order order)
        {
            if (!order.Complete)
            {
                order.Complete = true;
                order.SaveOrderComplete(true);
            }
        }

        private string ConvertAuthorizeNetToDwStatus(ResponseCodes gatewayPaymentStatusResult)
        {
            switch (gatewayPaymentStatusResult)
            {
                case ResponseCodes.Approved:
                    return GatewayPaymentStatus.Paid;
                case ResponseCodes.Declined:
                    return GatewayPaymentStatus.Declined;
                case ResponseCodes.Error:
                    return GatewayPaymentStatus.Rejected;
                case ResponseCodes.Held:
                    return GatewayPaymentStatus.Hold;
                default:
                    return GatewayPaymentStatus.Unknown;
            }
        }

        private string PrintRelayTemplate(Order order, int pageId, ResponseCodes responseCode)
        {
            LogEvent(order, "Printing relay template");
            var relayTemplate = new Template(RelayTemplate);
            var request = Context.Current.Request;

            if (responseCode == ResponseCodes.Approved && request["x_trans_id"] == "0")
            {
                relayTemplate.SetTag("Authorizenet.x_test", true);
            }
            relayTemplate.SetTag("Authorizenet.x_amount", request["x_amount"]);
            relayTemplate.SetTag("Authorizenet.x_trans_id", request["x_trans_id"]);
            relayTemplate.SetTag("Authorizenet.x_response_code", request["x_response_code"]);
            relayTemplate.SetTag("Authorizenet.x_response_reason_code", request["x_response_reason_code"]);
            relayTemplate.SetTag("Authorizenet.x_response_reason_text", request["x_response_reason_text"]);

            if (responseCode == ResponseCodes.Approved)
            {
                relayTemplate.SetTag("Authorizenet.x_return_URL", AcceptUrl(order, pageId));
            }
            else
            {
                relayTemplate.SetTag("Authorizenet.x_return_URL", DefaultUrl());
            }

            return relayTemplate.Output();
        }

        private string PrintErrorTemplate(Order order, string errMessage)
        {
            LogEvent(order, "Printing error template");

            var errorTemplate = new Template(ErrorTemplate);
            errorTemplate.SetTag("CheckoutHandler:ErrorMessage", errMessage);

            var orderRenderer = new Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(errorTemplate, order, true);

            return errorTemplate.Output();
        }

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="name"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string name)
        {
            try
            {
                switch (name)
                {
                    case "Post mode":
                        return new Hashtable
                        {
                            { "Auto", Translate.Translate("Auto post (does not use the template)") },
                            { "Template", Translate.Translate("Render template") }
                        };

                    case "The type of credit card transaction":
                        return new Hashtable
                        {
                            { "AUTH_CAPTURE", Translate.Translate("Authorization and Capture") },
                            { "AUTH_ONLY", Translate.Translate("Authorization only") }
                        };

                    case "Work Mode":
                        return new Hashtable
                        {
                            { "Developer", Translate.Translate("Developer") },
                            { "Test", Translate.Translate("Test") },
                            { "Production", Translate.Translate("Production") }
                        };

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", name));
                }
            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }
        }

        #region IRemoteCapture

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order)
        {
            return Capture(order, order.Price.PricePIP, false, false);
        }

        private OrderCaptureInfo Capture(Order order, long amount, bool final, bool isSplitCapture)
        {
            try
            {
                string errorText = string.Empty;

                if (order == null || string.IsNullOrEmpty(order.Id))
                {
                    errorText = "No valid Order object set";
                }
                else if (string.IsNullOrEmpty(order.TransactionNumber))
                {
                    errorText = "No transaction number set on the order";
                }
                else if (order.TransactionNumber.Trim() == "0")
                {
                    errorText = "Order with 0 transaction number can't be captured";
                }
                else if (order.Price.PricePIP < amount)
                {
                    LogError(null, "Amount to capture should be less of order total");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Amount to capture should be less of order total");
                }

                if (!string.IsNullOrEmpty(errorText))
                {
                    LogEvent(order, errorText);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
                }

                string serviceUrl = string.Empty;
                switch (workMode)
                {
                    case WorkModes.Developer:
                        serviceUrl = "https://test.authorize.net/gateway/transact.dll";
                        break;
                    case WorkModes.Test:
                        serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
                        break;
                    case WorkModes.Production:
                        serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
                        break;
                }
                decimal dAmount = amount / 100m;
                int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
                string strAmount = Math.Round(dAmount, decimals).ToString("0.00", System.Globalization.CultureInfo.CreateSpecificCulture("en-US"));

                var postDataValues = new StringBuilder();
                postDataValues.Append("x_type=PRIOR_AUTH_CAPTURE");
                postDataValues.AppendFormat("&x_login={0}", ApiLogin);
                postDataValues.AppendFormat("&x_tran_key={0}", TransactionKey);
                postDataValues.AppendFormat("&x_amount={0}", strAmount);
                postDataValues.AppendFormat("&x_delim_data={0}", "T");
                postDataValues.AppendFormat("&x_delim_char={0}", delimeterChar.ToString());
                postDataValues.AppendFormat("&x_trans_id={0}", order.TransactionNumber);
                postDataValues.Append("&x_relay_response=false");

                var response = SendRequest(serviceUrl, postDataValues.ToString()).Split(delimeterChar);
                LogEvent(order, "Remote capture response from AuthorizeNet");

                return ProcessCaptureResponse(order, response);
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);

                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        private string SendRequest(string serviceUrl, string postData)
        {
            var result = string.Empty;

            var webRequest = (HttpWebRequest)WebRequest.Create(serviceUrl);
            webRequest.Method = "POST";
            webRequest.ContentLength = postData.Length;
            webRequest.ContentType = "application/x-www-form-urlencoded";

            var myWriter = new StreamWriter(webRequest.GetRequestStream());
            myWriter.Write(postData);
            myWriter.Close();

            var response = (HttpWebResponse)webRequest.GetResponse();
            using (var responseStream = new StreamReader(response.GetResponseStream()))
            {
                result = responseStream.ReadToEnd();
                responseStream.Close();
            }

            return result;
        }

        private OrderCaptureInfo ProcessCaptureResponse(Order order, string[] response)
        {
            if (response.Length == 1)
            {
                throw new InvalidDataException(
                    string.Format(
                        "There was an error returned from AuthorizeNet: {0}; this usually means your data sent along was incorrect. Please recheck that all dates and amounts are formatted correctly",
                        response[0]
                    )
                );
            }

            ResponseCodes responseCode = (ResponseCodes)Converter.ToInt32(response[(int)AuthorizeNetResponseFields.ResponseCode]);
            order.GatewayPaymentStatus = ConvertAuthorizeNetToDwStatus(responseCode);
            if (responseCode == ResponseCodes.Approved)
            {
                LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
                order.CaptureAmount = Converter.ToDouble(response[(int)AuthorizeNetResponseFields.Amount]);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
            }
            else
            {
                string infoTxt = string.Format(
                    "Payment was unsucceeded with error {0}/{1} - {2}",
                    responseCode, response[(int)AuthorizeNetResponseFields.ResponseReasonCode], response[(int)AuthorizeNetResponseFields.ResponseReasonText]
                );
                LogEvent(order, infoTxt);

                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        #endregion

        #region private class Crypto

        [Obsolete("Md5 is removed in February 2019")]
        private class Crypto
        {
            private static string EncryptHMAC(string key, string value)
            {
                byte[] HMACkey = new ASCIIEncoding().GetBytes(key);
                byte[] HMACdata = new ASCIIEncoding().GetBytes(value);
                byte[] HMAChash = new HMACMD5(HMACkey).ComputeHash(HMACdata);
                string fingerprint = string.Empty;
                for (int i = 0; i < HMAChash.Length; i++)
                {
                    fingerprint = fingerprint + HMAChash[i].ToString("x").PadLeft(2, '0');
                }

                return fingerprint;
            }

            public static string GenerateFingerprint(string transactionKey, string login, string amount, string sequence, string timeStamp)
            {
                string keyString = string.Format("{0}^{1}^{2}^{3}^", new object[] { login, sequence, timeStamp.ToString(), amount });
                return EncryptHMAC(transactionKey, keyString);
            }

            public static bool IsMatch(string merchantHash, string apiLogin, string transactionID, string amount, string expectedMD5Hash)
            {
                string unencrypted = string.Format("{0}{1}{2}{3}", new object[] { merchantHash, apiLogin, transactionID, amount });
                var cryptoService = new MD5CryptoServiceProvider();
                var computedHash = BitConverter.ToString(cryptoService.ComputeHash(Encoding.Default.GetBytes(unencrypted))).Replace("-", string.Empty);
                return computedHash.Equals(expectedMD5Hash);
            }
        }

        private static readonly string[] Responsefields = new string[] {
            "x_trans_id",
            "x_test_request",
            "x_response_code",
            "x_auth_code",
            "x_cvv2_resp_code",
            "x_cavv_response",
            "x_avs_code",
            "x_method",
            "x_account_number",
            "x_amount",
            "x_company",
            "x_first_name",
            "x_last_name",
            "x_address",
            "x_city",
            "x_state",
            "x_zip",
            "x_country",
            "x_phone",
            "x_fax",
            "x_email",
            "x_ship_to_company",
            "x_ship_to_first_name",
            "x_ship_to_last_name",
            "x_ship_to_address",
            "x_ship_to_city",
            "x_ship_to_state",
            "x_ship_to_zip",
            "x_ship_to_country",
            "x_invoice_num"
        };

        private class CryptoSha
        {
            public static string GenerateFingerprint(string signatureKey, string login, string amount, string sequence, string timeStamp)
            {
                string keyString = string.Format("{0}^{1}^{2}^{3}^", new object[] { login, sequence, timeStamp.ToString(), amount });
                return EncryptHMACSHA512(signatureKey, keyString);
            }

            public static string EncryptHMACSHA512(string key, string textToHash)
            {
                if (key.Length % 2 != 0 || key.Trim().Length < 2)
                {
                    throw new ArgumentNullException("HMACSHA512: key", "Parameter cannot be odd or less than 2 characters.");
                }

                byte[] k = Enumerable.Range(0, key.Length)
                            .Where(x => x % 2 == 0)
                            .Select(x => Convert.ToByte(key.Substring(x, 2), 16))
                            .ToArray();
                var hmac = new HMACSHA512(k);
                byte[] HashedValue = hmac.ComputeHash((new ASCIIEncoding()).GetBytes(textToHash));
                return BitConverter.ToString(HashedValue).Replace("-", string.Empty);
            }

            public static string GenerateSequence()
            {
                Random random = new Random();
                return random.Next(0, 0x3e8).ToString();
            }

            public static int GenerateTimestamp()
            {
                TimeSpan ts = DateTime.UtcNow - new DateTime(0x7b2, 1, 1);
                return (int)ts.TotalSeconds;
            }

            public static bool IsMatch(Environment.IRequest request, string signatureKey, string expectedHash)
            {
                var computedHash = EncryptHMACSHA512(signatureKey, "^" + string.Join("^", Responsefields.Select(x => request[x])) + "^");
                return computedHash.Equals(expectedHash);
            }
        }

        #endregion

        #region IRemotePartialCapture

        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            return Capture(order, amount, final, true);
        }

        public bool CaptureSupported(Order order)
        {
            return true;
        }

        public bool SplitCaptureSupported(Order order)
        {
            return true;
        }
        #endregion
    }


}
