﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Security.Cryptography;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.AuthorizeNetAIM
{
    /// <summary>
    /// AuthorizeNet Checkout Handler
    /// </summary>
    [
		AddInName("AuthorizeNetAIM"),
		AddInDescription("AuthorizeNet (Advanced Integration Method)")
	]
	public class AuthorizeNetAIM : CheckoutHandlerWithStatusPage, IDropDownOptions, IRemoteCapture
	{
		private enum WorkModes { Developer, Test, Production }
		private enum ResponseCodes { Approved = 1, Declined = 2, Error = 3, Held = 4 }

		private char delimeterChar;
		private string paymentError;
		private WorkModes workMode;

        #region Addin parameters

        [AddInParameter("API login ID"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string ApiLogin { get; set; }

		[AddInParameter("Transaction key"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string TransactionKey { get; set; }

        [AddInParameter("Signature key"), AddInParameterEditor(typeof(TextParameterEditor), "TextArea=true")]
        public string SignatureKey { get; set; }

        [AddInParameter("The type of credit card transaction"), AddInParameterEditor(typeof(RadioParameterEditor), "")]
        public string TransactionType { get; set; }

        [AddInParameter("Merchant email"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string MerchantEmail { get; set; }

		[AddInParameter("Send receipt to the customer"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool SendReceipt { get; set; }

		[AddInParameter("The email receipt header"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string ReceiptHeader { get; set; }

		[AddInParameter("The email receipt footer"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string ReceiptFooter { get; set; }

		[AddInParameter("Payment template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Payment")]
		public string PaymentTemplate { get; set; }

		[AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Cancel")]
		public string CancelTemplate { get; set; }

		[AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/AuthorizeNet/Error")]
		public string ErrorTemplate { get; set; }

		[AddInParameter("Work Mode"), AddInParameterEditor(typeof(RadioParameterEditor), "")]
		public string WorkMode
		{
			get { return workMode.ToString(); }
			set
			{
				switch (value)
				{
					case "Developer":
						workMode = WorkModes.Developer;
						break;

					case "Test":
						workMode = WorkModes.Test;
						break;

					case "Production":
						workMode = WorkModes.Production;
						break;
				}
			}
		}

		[AddInParameter("Check secure connection in Test mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool CheckSecureConnection { get; set; }

		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref="AuthorizeNet" /> class
		/// </summary>
		public AuthorizeNetAIM()
		{
		    ServicePointManager.SecurityProtocol = ServicePointManager.SecurityProtocol | SecurityProtocolType.Tls12;
            TransactionType = "AUTH_CAPTURE";
			delimeterChar = '|';
			CheckSecureConnection = true;
		}

		/// <summary>
		/// Starts order checkout procedure
		/// </summary>
		/// <param name="order">Order to be checked out</param>
		/// <returns>String representation of template output</returns>
		public override string StartCheckout(Order order)
		{
			try
			{
				LogEvent(order, "Checkout started");
				LogEvent(order, "Render AuthorizeNet template");

				if (workMode == WorkModes.Production || CheckSecureConnection)
				{
					if (!Context.Current.Request.IsSecureConnection)
					{
						throw new Exception("AuthorizeNet AIM Payment Gateway requires secure connection!");
					}
				}
				if (order.CurrencyCode != "USD" && order.CurrencyCode != "CAD" && order.CurrencyCode != "GBP")
				{
					throw new Exception($"Only USD, CAD, or GBP currency is allowed. Order currency: {order.CurrencyCode}");
				}

				var formTemplate = new Template(PaymentTemplate);

				var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
				orderRenderer.RenderOrderDetails(formTemplate, order, true);
				formTemplate.SetTag("Authorizenet.x_card_num", WebUtility.UrlEncode(Dynamicweb.Context.Current.Request["x_card_num"]));
				formTemplate.SetTag("Authorizenet.x_exp_date", WebUtility.UrlEncode(Dynamicweb.Context.Current.Request["x_exp_date"]));
				formTemplate.SetTag("Authorizenet.x_card_code", WebUtility.UrlEncode(Dynamicweb.Context.Current.Request["x_card_code"]));
				if (!string.IsNullOrEmpty(paymentError))
				{
					formTemplate.SetTag("Authorizenet.PaymentError", paymentError);
				}

				return formTemplate.Output();
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (Exception ex)
			{
				LogError(order, "Unhandled exception with message: {0}", ex.Message);
				return PrintErrorTemplate(order, ex.Message);
			}
		}

		/// <summary>
		/// Redirects user to payment step or to canceled order step
		/// </summary>
		/// <param name="order">Order for processing</param>
		/// <returns>String representation of template output</returns>
		public override string Redirect(Order order)
		{
			try
			{
				LogEvent(order, "Redirected to AuthorizeNet CheckoutHandler");
				var state = Converter.ToString(Context.Current.Request["AuthState"]);

				switch (state)
				{
					case "Payment":
						return StatePayment(order);

					case "Cancel":
						return StateCancel(order);

					default:
						throw new ArgumentException(string.Format("Unhandled payment state: '{0}'", state));
				}
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (Exception ex)
			{
				LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
				return PrintErrorTemplate(order, ex.Message);
			}
		}

		private string StatePayment(Order order)
		{
			LogEvent(order, "State Payment");

			int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
			string amount = Math.Round(order.Price.Price, decimals).ToString("0.00", CultureInfo.CreateSpecificCulture("en-US"));
			string serviceUrl = string.Empty;
			string testMode = string.Empty;
			switch (workMode)
			{
				case WorkModes.Developer:
					serviceUrl = "https://test.authorize.net/gateway/transact.dll";
					testMode = "0";
					break;

				case WorkModes.Test:
					serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
					testMode = "1";
					break;

				case WorkModes.Production:
					serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
					testMode = "0";
					break;
			}

			string customerName = Converter.ToString(order.CustomerName);
			int delimiterPosition = customerName.IndexOf(' ');
			string firstName = delimiterPosition >= 0 ? customerName.Substring(0, delimiterPosition) : customerName;
			string lastName = delimiterPosition >= 0 ? customerName.Substring(delimiterPosition + 1) : string.Empty;

			var formValues = new Dictionary<string, string> {
				{ "x_login", ApiLogin },
				{ "x_tran_key", TransactionKey },
				{ "x_amount", amount },
				{ "x_test_request", testMode },
				{ "x_invoice_num", order.Id },
				{ "x_description", $"Payment to {Context.Current.Request.Url.Host}" },
				{ "x_first_name", firstName },
				{ "x_last_name", lastName },
				{ "x_company", Converter.ToString(order.CustomerCompany) },
				{ "x_address", string.Format("{0} {1}", Converter.ToString(order.CustomerAddress), Converter.ToString(order.CustomerAddress2)) },
				{ "x_city", Converter.ToString(order.CustomerCity) },
				{ "x_state", Converter.ToString(order.CustomerRegion) },
				{ "x_zip", Converter.ToString(order.CustomerZip) },
				{ "x_country", Converter.ToString(order.CustomerCountry) },
				{ "x_email", Converter.ToString(order.CustomerEmail) },
				{ "x_delim_data", "T" },
				{ "x_delim_char", delimeterChar.ToString() },
				{ "x_email_customer", SendReceipt ? "T" : "F" },
				{ "x_merchant_email", MerchantEmail },
				{ "x_header_email_receipt", ReceiptHeader },
				{ "x_footer_email_receipt", ReceiptFooter },
				{ "x_relay_response", "F" },
				{ "x_type", TransactionType }
			}.ToDictionary(x => x.Key, x => x.Value);

			string[] paymentFormValues = { "x_card_num", "x_exp_date", "x_card_code" };

			string parameters = "x_method=CC&x_version=3.1";
			foreach (string key in formValues.Keys)
			{
				parameters += string.Format("&{0}={1}", key, WebUtility.UrlEncode(formValues[key]));
			}

			foreach (string key in paymentFormValues)
			{
				parameters += string.Format("&{0}={1}", key, WebUtility.UrlEncode(Dynamicweb.Context.Current.Request[key]));
			}

			LogEvent(order, "Post to AuthorizeNet");
			string response = SendRequest(serviceUrl, parameters);

			return ProcessResponse(order, response.Split(delimeterChar));
		}

		private string ProcessResponse(Order order, string[] response)
		{
			if (response.Length == 1)
			{
				throw new InvalidDataException(
					string.Format(
						"There was an error returned from AuthorizeNet: {0}; this usually means your data sent along was incorrect. Please recheck that all dates and amounts are formatted correctly",
						response[0]
					)
				);
			}

			ResponseCodes responseCode = (ResponseCodes)Converter.ToInt32(response[(int)AuthorizeNetResponseFields.ResponseCode]);
			order.GatewayPaymentStatus = ConvertAuthorizeNetToDwStatus(responseCode);
			if (responseCode == ResponseCodes.Approved)
			{
				LogEvent(order, "State ok");

				double amount = 0;
                string amountString = string.Empty;
				if (response.Length > (int)AuthorizeNetResponseFields.Amount)
				{
                    amountString = response[(int)AuthorizeNetResponseFields.Amount];
                    amount = double.Parse(amountString, CultureInfo.CreateSpecificCulture("en-US"));
				}

				int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
				if (amount < Math.Round(order.Price.Price, decimals))
				{
					throw new Exception(string.Format("The amount paid for({0}) is less than the amount of the order. Something's fishy...", amount));
				}

                if (!string.IsNullOrEmpty(SignatureKey))
                {
                    var sha2 = response[(int)AuthorizeNetResponseFields.SHA2];
                    if (!CryptoSha.IsMatch(ApiLogin, response[(int)AuthorizeNetResponseFields.TransactionID], amountString, SignatureKey, sha2))
                    {
                        throw new Exception("SHA hash do not match.");
                    }
                }

				try
				{
					SetOrderCompleteSafety(order);

					if (TransactionType == "AUTH_CAPTURE")
					{
						string infoTxt = "Capture succeeded";
						LogEvent(order, infoTxt);
						order.CaptureInfo = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, infoTxt);
					}

					order.GatewayResult = string.Format("<result>{0}</result>", string.Join(delimeterChar.ToString(), response));
					SetTransactionState(order, true, amount, response[(int)AuthorizeNetResponseFields.TransactionID], response[(int)AuthorizeNetResponseFields.CardType]);
				}
				finally
				{
					CheckoutDone(order);
				}

				RedirectToCart(order);

				return null;
			}
			else
			{
				paymentError = string.Format(
					"Payment was unsucceeded with error {0}/{1} - {2}",
					responseCode, response[(int)AuthorizeNetResponseFields.ResponseReasonCode], response[(int)AuthorizeNetResponseFields.ResponseReasonText]
				);
				LogEvent(order, paymentError);

				return StartCheckout(order);
			}
		}

		private string StateCancel(Order order)
		{
			LogEvent(order, "State cancel");

			order.TransactionStatus = "Cancelled";
			order.Save();

			CheckoutDone(order);

			var cancelTemplate = new Template(CancelTemplate);
			var orderRenderer = new Ecommerce.Frontend.Renderer();
			orderRenderer.RenderOrderDetails(cancelTemplate, order, true);

			return cancelTemplate.Output();
		}

		private string SendRequest(string serviceUrl, string postData)
		{
			var result = string.Empty;

			var webRequest = (HttpWebRequest)WebRequest.Create(serviceUrl);
			webRequest.Method = "POST";
			webRequest.ContentLength = postData.Length;
			webRequest.ContentType = "application/x-www-form-urlencoded";

			var myWriter = new StreamWriter(webRequest.GetRequestStream());
			myWriter.Write(postData);
			myWriter.Close();

			var response = (HttpWebResponse)webRequest.GetResponse();
			using (var responseStream = new StreamReader(response.GetResponseStream()))
			{
				result = responseStream.ReadToEnd();
				responseStream.Close();
			}

			return result;
		}

		private string ConvertAuthorizeNetToDwStatus(ResponseCodes gatewayPaymentStatusResult)
		{
			switch (gatewayPaymentStatusResult)
			{
				case ResponseCodes.Approved:
					return GatewayPaymentStatus.Paid;
				case ResponseCodes.Declined:
					return GatewayPaymentStatus.Declined;
				case ResponseCodes.Error:
					return GatewayPaymentStatus.Rejected;
				case ResponseCodes.Held:
					return GatewayPaymentStatus.Hold;
				default:
					return GatewayPaymentStatus.Unknown;
			}
		}

		private string PrintErrorTemplate(Order order, string errMessage)
		{
			LogEvent(order, "Printing error template");

			var errorTemplate = new Template(ErrorTemplate);
			errorTemplate.SetTag("CheckoutHandler:ErrorMessage", errMessage);
			var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
			orderRenderer.RenderOrderDetails(errorTemplate, order, true);

			return errorTemplate.Output();
		}

		/// <summary>
		/// Gets options according to behavior mode
		/// </summary>
		/// <param name="behaviorMode"></param>
		/// <returns>key-value pairs of settings</returns>
		public Hashtable GetOptions(string behaviorMode)
		{
			try
			{
				switch (behaviorMode)
				{
					case "The type of credit card transaction":
						return new Hashtable
                        {
							{ "AUTH_CAPTURE", Translate.Translate("Authorization and Capture") },
							{ "AUTH_ONLY", Translate.Translate("Authorization only") }
						};

					case "Work Mode":
						return new Hashtable
                        {
							{ "Developer", Translate.Translate("Developer") },
							{ "Test", Translate.Translate("Test") },
							{ "Production", Translate.Translate("Production") }
						};

					default:
						throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", behaviorMode));
				}
			}
			catch (Exception ex)
			{
				LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
				return null;
			}
		}

		#region IRemoteCapture
		/// <summary>
		/// Send capture request to transaction service
		/// </summary>
		/// <param name="order">Order to be captured</param>
		/// <returns>Response from transaction service</returns>
		public OrderCaptureInfo Capture(Order order)
		{
			try
			{
				string errorText = string.Empty;

				if (order == null || string.IsNullOrEmpty(order.Id))
				{
					errorText = "No valid Order object set";
				}
				else if (string.IsNullOrEmpty(order.TransactionNumber))
				{
					errorText = "No transaction number set on the order";
				}
				else if (order.TransactionNumber.Trim() == "0")
				{
					errorText = "Order with 0 transaction number can't be captured";
				}

				if (!string.IsNullOrEmpty(errorText))
				{
					LogEvent(order, errorText);
					return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
				}

				string serviceUrl = string.Empty;
				switch (workMode)
				{
					case WorkModes.Developer:
						serviceUrl = "https://test.authorize.net/gateway/transact.dll";
						break;

					case WorkModes.Test:
						serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
						break;

					case WorkModes.Production:
						serviceUrl = "https://secure2.authorize.net/gateway/transact.dll";
						break;
				}

				int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
				string orderPrice = Math.Round(order.Price.Price, decimals).ToString("0.00", CultureInfo.CreateSpecificCulture("en-US"));

				var postDataValues = new StringBuilder();
				postDataValues.Append("x_type=PRIOR_AUTH_CAPTURE");
				postDataValues.AppendFormat("&x_login={0}", ApiLogin);
				postDataValues.AppendFormat("&x_tran_key={0}", TransactionKey);
				postDataValues.AppendFormat("&x_amount={0}", orderPrice);
				postDataValues.AppendFormat("&x_delim_data={0}", "T");
				postDataValues.AppendFormat("&x_delim_char={0}", delimeterChar.ToString());
				postDataValues.AppendFormat("&x_trans_id={0}", order.TransactionNumber);
				postDataValues.Append("&x_relay_response=false");

				var response = SendRequest(serviceUrl, postDataValues.ToString()).Split(delimeterChar);
				LogEvent(order, "Remote capture response from AuthorizeNet");

				return ProcessCaptureResponse(order, response);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
			}
		}

		private OrderCaptureInfo ProcessCaptureResponse(Order order, string[] response)
		{
			if (response.Length == 1)
			{
				throw new InvalidDataException(
					string.Format(
						"There was an error returned from AuthorizeNet: {0}; this usually means your data sent along was incorrect. Please recheck that all dates and amounts are formatted correctly",
						response[0]
					)
				);
			}

			var responseCode = (ResponseCodes)Converter.ToInt32(response[(int)AuthorizeNetResponseFields.ResponseCode]);
			order.GatewayPaymentStatus = ConvertAuthorizeNetToDwStatus(responseCode);
			if (responseCode == ResponseCodes.Approved)
			{
				LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
			}
			else
			{
				string message = string.Format(
					string.Format(
						"Payment was unsucceeded with error {0}/{1} - {2}",
						responseCode, response[(int)AuthorizeNetResponseFields.ResponseReasonCode], response[(int)AuthorizeNetResponseFields.ResponseReasonText]
					)
				);
				LogEvent(order, message);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, message);
			}
		}

        #endregion

        #region private class Crypto

        private class CryptoSha
        {
            private static string EncryptHMACSHA512(string key, string textToHash)
            {
                if (key.Length % 2 != 0 || key.Trim().Length < 2)
                {
                    throw new ArgumentNullException("HMACSHA512: key", "Parameter cannot be odd or less than 2 characters.");
                }

                byte[] k = Enumerable.Range(0, key.Length)
                            .Where(x => x % 2 == 0)
                            .Select(x => Convert.ToByte(key.Substring(x, 2), 16))
                            .ToArray();
                var hmac = new HMACSHA512(k);
                byte[] HashedValue = hmac.ComputeHash((new ASCIIEncoding()).GetBytes(textToHash));
                return BitConverter.ToString(HashedValue).Replace("-", string.Empty);
            }

            public static bool IsMatch( string login, string transactionId, string amount, string signatureKey, string expectedHash)
            {
                string keyString = string.Format("^{0}^{1}^{2}^", new object[] { login, transactionId, amount });
                var computedHash = EncryptHMACSHA512(signatureKey, keyString);
                return computedHash.Equals(expectedHash);
            }
        }


        #endregion
    }
}
