﻿using Dynamicweb.Core;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.BeanstreamHostedForm
{
    /// <summary>
    /// BeanstreamHostedForm Checkout Handler
    /// </summary>
    [
		AddInName("Bambora"),
		AddInDescription("Bambora hosted form")
	]
	public class BeanstreamHostedForm : CheckoutHandler, IDropDownOptions, IRemoteCapture
	{
		private enum PostModes { Auto, Template }

		private PostModes postMode = PostModes.Auto;

		#region Addin parameters

		[AddInParameter("Merchant ID"), AddInParameterEditor(typeof(TextParameterEditor), "")]
		public string MerchantId { get; set; }

		[AddInParameter("Hash key"), AddInParameterEditor(typeof(TextParameterEditor), "")]
		public string HashKey { get; set; }

		[AddInParameter("Transaction type")]
		[AddInParameterEditor(typeof(RadioParameterEditor), "")]
		public string TransactionType { get; set; }

		[AddInParameter("Post mode"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
		public string PostModeSelection
		{
			get
			{
				return postMode.ToString();
			}
			set
			{
				switch (value)
				{
					case "Auto":
						postMode = PostModes.Auto;
						break;

					case "Template":
						postMode = PostModes.Template;
						break;
				}
			}
		}

		[AddInParameter("URL to cart for iframe"), AddInParameterEditor(typeof(PageSelectEditor), "")]
		public string UrlToPageWithCart { get; set; }

		[AddInParameter("Payment URL"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string PaymentFormURL { get; set; }

		[AddInParameter("Transaction URL"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
		public string TransactionURL { get; set; }

		[AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Beanstream/Post")]
		public string PostTemplate { get; set; }

		[AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Beanstream/Cancel")]
		public string CancelTemplate { get; set; }

		[AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Beanstream/Error")]
		public string ErrorTemplate { get; set; }

		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref="BeanstreamHostedForm" /> class.
		/// </summary>
		public BeanstreamHostedForm()
		{
			PaymentFormURL = "https://www.beanstream.com/scripts/payment/payment.asp";
			TransactionURL = "https://www.beanstream.com/scripts/process_transaction.asp";
			TransactionType = "PA";

			PostTemplate = "eCom7/CheckoutHandler/Beanstream/Post/Post_iframe.html";
			CancelTemplate = "eCom7/CheckoutHandler/Beanstream/Cancel/checkouthandler_cancel.html";
			ErrorTemplate = "eCom7/CheckoutHandler/Beanstream/Error/checkouthandler_error.html";
		}

		/// <summary>
		/// Starts order checkout procedure
		/// </summary>
		/// <param name="order">Order to be checked out</param>
		/// <returns>String representation of template output</returns>
		public override string StartCheckout(Order order)
		{
			try
			{
				LogEvent(order, "Checkout started");

				if (order.CurrencyCode != "USD" && order.CurrencyCode != "CAD")
				{
					throw new Exception($"Only USD or CAD currency is allowed. Order currency: {order.CurrencyCode}");
				}

				if (string.IsNullOrEmpty(order.CustomerRegion))
				{
					throw new Exception("Customer region field is empty");
				}

				int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
				var amount = Math.Round(order.Price.Price, decimals).ToString("0.00", System.Globalization.CultureInfo.InvariantCulture);
				bool isDeliveryFieldsFilled = !string.IsNullOrEmpty(order.DeliveryAddress) && !string.IsNullOrEmpty(order.DeliveryZip);
				string customerName = string.IsNullOrEmpty(order.CustomerName) ? order.CustomerFirstName + " " + order.CustomerSurname : order.CustomerName;
				string deliveryName = string.IsNullOrEmpty(order.DeliveryName) ? order.DeliveryFirstName + " " + order.DeliverySurname : order.DeliveryName;

				if (isDeliveryFieldsFilled && string.IsNullOrEmpty(order.DeliveryRegion))
				{
					throw new Exception("Delivery region field is empty");
				}

				var formValues = new Dictionary<string, string> {
					{ "merchant_id", MerchantId },
					{ "trnType", TransactionType },
					{ "trnOrderNumber", order.Id },
					{ "trnAmount", amount },
					{ "trnLanguage", DefaultLanguageCode },
					{ "ordName", Converter.ToString(customerName) },
					{ "ordAddress1", Converter.ToString(order.CustomerAddress) },
					{ "ordAddress2", Converter.ToString(order.CustomerAddress2) },
					{ "ordCity", Converter.ToString(order.CustomerCity) },
					{ "ordProvince", Converter.ToString(order.CustomerRegion) },
					{ "ordCountry", Converter.ToString(order.CustomerCountryCode) },
					{ "ordPostalCode", Converter.ToString(order.CustomerZip) },
					{ "ordEmailAddress", Converter.ToString(order.CustomerEmail) },
					{ "ordPhoneNumber", Converter.ToString(order.CustomerPhone) },
					{ "shipName", Converter.ToString(isDeliveryFieldsFilled ? deliveryName : customerName) },
					{ "shipAddress1", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryAddress : order.CustomerAddress) },
					{ "shipAddress2", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryAddress2 : order.CustomerAddress2) },
					{ "shipCity", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryCity : order.CustomerCity) },
					{ "shipProvince", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryRegion : order.CustomerRegion) },
					{ "shipCountry", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryCountryCode: order.CustomerCountryCode) },
					{ "shipPostalCode", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryZip : order.CustomerZip) },
					{ "shipEmailAddress", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryEmail : order.CustomerEmail) },
					{ "shipPhoneNumber", Converter.ToString(isDeliveryFieldsFilled ? order.DeliveryPhone : order.CustomerPhone) },
					{ "shippingMethod", Converter.ToString(order.ShippingMethodDescription) }
				}.ToDictionary(x => x.Key, x => x.Value);

				string postData = BuildPostData(formValues);
				var post = new StringBuilder();
				post.Append(PaymentFormURL);
				post.Append("?");
				post.Append(postData);
				post.AppendFormat("&hashValue={0}", ComputeHash(postData));

				switch (postMode)
				{
					case PostModes.Auto:
						LogEvent(order, "Auto post");

						post.AppendFormat("&approvedPage={0}", AcceptUrl(order));
						post.AppendFormat("&declinedPage={0}", CancelUrl(order));

						Context.Current.Response.Redirect(post.ToString(), false);
						return string.Empty;

					case PostModes.Template:
						LogEvent(order, "Render template");

						var postTemplate = new Template(PostTemplate);
						postTemplate.SetTag("Beanstream.HostedFormURL", post.ToString());
						postTemplate.SetTag("Beanstream.AcceptUrl", AcceptUrl(order));
						postTemplate.SetTag("Beanstream.CancelUrl", CancelUrl(order));
						return Render(order, postTemplate);

					default:
						LogError(order, "Unhandled post mode: '{0}'", postMode);
						throw new ArgumentException(string.Format("Unhandled post mode: '{0}'", postMode));
				}
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (Exception ex)
			{
				LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
				return OnError(order, ex.Message);
			}
		}

		/// <summary>
		/// Redirects user to payment step or canceled order step
		/// </summary>
		/// <param name="order">Order for processing</param>
		/// <returns>String representation of template output</returns>
		public override string Redirect(Order order)
		{
			try
			{
				LogEvent(null, "Redirected to Beanstream CheckoutHandler");

				switch (Dynamicweb.Context.Current.Request["BeanState"])
				{
					case "Accept":
						return StateOk(order);

					case "Cancel":
						return StateCancel(order);

					default:
						Callback(order);
						Context.Current.Response.End();
						return null;
				}
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (Exception ex)
			{
				LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
				return OnError(order, ex.Message);
			}
		}

		/// <summary>
		/// Send capture request to transaction service
		/// </summary>
		/// <param name="order">Order to be captured</param>
		/// <returns>Response from transaction service</returns>
		public OrderCaptureInfo Capture(Order order)
		{
			try
			{
				var errorText = string.Empty;

				if (order == null || string.IsNullOrEmpty(order.Id))
				{
					errorText = "No valid Order object set";
				}
				else if (string.IsNullOrEmpty(order.TransactionNumber))
				{
					errorText = "No transaction number set on the order";
				}
				else if (order.TransactionNumber.Trim() == "0")
				{
					errorText = "Order with 0 transaction number can't be captured";
				}

				if (!string.IsNullOrEmpty(errorText))
				{
					LogEvent(order, errorText);
					return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
				}

				int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == string.Empty ? 2 : order.Currency.Rounding.Decimals;
				var strAmount = Math.Round(order.Price.Price, decimals).ToString("0.00", CultureInfo.InvariantCulture);

				var content = new StringBuilder();
				content.AppendFormat("merchant_id={0}", MerchantId);
				content.Append("&RequestType=BACKEND");
				content.Append("&trnType=PAC");
				content.AppendFormat("&trnAmount={0}", strAmount);
				content.AppendFormat("&adjId={0}", order.TransactionNumber);
				content.AppendFormat("&hashValue={0}", ComputeHash(content.ToString()));

				LogEvent(order, "Process capture response");

                var completeResponse = SendRequest(TransactionURL, content.ToString());
                var response = ParseResponse(completeResponse);
                var hashData = GetResponseData(response);
                var hash = ComputeHash(hashData);
                var newHashData = completeResponse.Substring(0, completeResponse.IndexOf("&hashValue="));
                var newHash = ComputeHash(newHashData);

                if (hash != response["hashValue"] && newHash != response["hashValue"])
                {
                    var infoTxt = string.Format("The hash checksum returned from callback does not match: Callback: {0}, calculated: {1}, calculated2 {2}", response["hashValue"], hash, newHash);

					LogEvent(order, infoTxt, DebuggingInfoType.CaptureResult);
					return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
				}
				else if (response["trnApproved"] == "1")
				{
					LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
					return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
				}
				else
				{
					string infoTxt = string.Format("Remote Capture failed. messageId: '{0}', messageText: '{1}'", response["messageId"], WebUtility.UrlDecode(response["messageText"]));

					LogEvent(order, infoTxt, DebuggingInfoType.CaptureResult);
					return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
				}

			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);

				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
			}
		}

		/// <summary>
		/// Gets options according to behavior mode
		/// </summary>
		/// <param name="name"></param>
		/// <returns>Key-value pairs of settings</returns>
		public Hashtable GetOptions(string name)
		{
			var result = new Hashtable();
			switch (name)
			{
				case "Transaction type":
					result.Add("P", "Purchase");
					result.Add("PA", "Pre Authorization");

					break;

				case "Post mode":
					return new Hashtable {
						{ "Auto", Dynamicweb.SystemTools.Translate.Translate("Auto post (does not use the template)", 0) },
						{ "Template", Dynamicweb.SystemTools.Translate.Translate("Render template", 0) }
					};

				default:
					break;
			}

			return result;
		}

		private string GetCustomUrlToCart(Order order)
		{
			if (string.IsNullOrEmpty(UrlToPageWithCart)) return GetBaseUrl(order);

			var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
			var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);

			return string.Format(
				"{0}://{1}{2}/{3}&{4}={5}",
				Context.Current.Request.Url.Scheme, Context.Current.Request.Url.Host, portString, UrlToPageWithCart, OrderIdRequestName, order.Id
			);
		}

		private string AcceptUrl(Order order)
		{
			return WebUtility.UrlEncode(string.Format("{0}&BeanState=Accept", GetCustomUrlToCart(order)));
		}

		private string CancelUrl(Order order)
		{
			return WebUtility.UrlEncode(string.Format("{0}&BeanState=Cancel", GetCustomUrlToCart(order)));
		}

		private static string CallbackUrl(Order order)
		{
			var portString = Context.Current.Request.Url.IsDefaultPort ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);
			return string.Format(
				"{1}://{0}{2}/Admin/Public/Gateways/CheckoutHandler/Beanstream/BeanstreamFormCallBack.aspx",
				Context.Current.Request.Url.Host, Context.Current.Request.Url.Scheme, portString
			);
		}

		private static string BuildPostData(Dictionary<string, string> formValues)
		{
			return string.Join("&", formValues.Select(x => string.Format("{0}={1}", x.Key, WebUtility.UrlEncode(x.Value))).ToArray());
		}

		private static string GetResponseData(NameValueCollection formValues)
		{
			var skipFields = new string[] { "ID", OrderIdRequestName, "BeanState", "hashValue" };
			return string.Join("&", formValues.AllKeys.Where(x => !skipFields.Contains(x, StringComparer.InvariantCultureIgnoreCase)).Select(x => string.Format("{0}={1}", x, formValues[x])).ToArray());
		}

		private string ComputeHash(string stringToHash)
		{
			return StringHelper.Md5HashToString(stringToHash + HashKey);
		}

		private static string DefaultLanguageCode
		{
			get
			{
				var currentLanguageCode = Dynamicweb.Environment.ExecutingContext.GetCulture(true).ThreeLetterISOLanguageName;
				var supportedLanguageCodes = new[] { "ENG", "FRE" };
				return supportedLanguageCodes.Contains(currentLanguageCode, StringComparer.InvariantCultureIgnoreCase) ? currentLanguageCode : "ENG";
			}
		}

		private string StateOk(Order order)
		{
			LogEvent(order, "State ok");

			if (!order.Complete)
			{
				ProcessOrder(order);
			}

			RedirectToCart(order);
			return null;
		}

		private string StateCancel(Order order)
		{
			LogEvent(order, "State cancel");

			order.TransactionStatus = "Cancelled";
			order.Save();

			CheckoutDone(order);

			var cancelTemplate = new Template(CancelTemplate);
			cancelTemplate.SetTag("CheckoutHandler:CancelMessage", Dynamicweb.Context.Current.Request["messageText"]);
			return Render(order, cancelTemplate);
		}

		private string OnError(Order order, string msg)
		{
			order.TransactionAmount = 0;
			order.TransactionStatus = "Failed";
			order.Errors.Add(msg);
			order.Save();

			order.DowngradeToCart();
			order.CartV2StepIndex = 0;
			order.TransactionStatus = "";
			Dynamicweb.Ecommerce.Common.Context.SetCart(order);

			if (string.IsNullOrWhiteSpace(ErrorTemplate))
			{
				RedirectToCart(order);
			}

			var errorTemplate = new Template(ErrorTemplate);
			errorTemplate.SetTag("CheckoutHandler:ErrorMessage", msg);

			return Render(order, errorTemplate);
		}

		private void Callback(Order order)
		{
			LogEvent(order, "Callback started");

			if (!order.Complete)
			{
				ProcessOrder(order, true);
			}
		}

		private void ProcessOrder(Order order, bool isCallback = false)
		{
			bool orderWasCompleted = order.Complete;

			try
			{
				bool errorOccured = false;

				var parameters = Context.Current.Request.HttpMethod == "POST" ? ParseRawForm() : ParseResponse(Context.Current.Request.RawUrl);

				string calculatedhash = GetResponseData(parameters);
				if (ComputeHash(calculatedhash) != parameters["hashValue"])
				{
					errorOccured = true;
					LogError(order, "The hash checksum returned from callback does not match: Callback: {0}, calculated: {1}", parameters["hashValue"], ComputeHash(calculatedhash));
				}

				string amount = Dynamicweb.Context.Current.Request["trnAmount"];
				int decimals = order.Currency.Rounding == null || order.Currency.Rounding.Id == "" ? 2 : order.Currency.Rounding.Decimals;
				if (amount != Math.Round(order.Price.Price, decimals).ToString("0.00", CultureInfo.InvariantCulture))
				{
					LogError(order, "The amount returned from callback does not match the amount set on the order: Callback: {0}, order: {1}",
						amount, Math.Round(order.Price.Price, decimals).ToString("0.00", System.Globalization.CultureInfo.InvariantCulture)
					);
					errorOccured = true;
				}

				string transact = Dynamicweb.Context.Current.Request["trnId"];
				if (string.IsNullOrEmpty(transact))
				{
					errorOccured = true;
					LogEvent(order, "No transaction number sent to callback");
				}

				if (Dynamicweb.Context.Current.Request["trnApproved"] != "1")
				{
					errorOccured = true;
					LogEvent(order, "Transaction not approved. Beanstream response: messageId: '{0}', messageText: '{1}'.", Dynamicweb.Context.Current.Request["messageId"], Dynamicweb.Context.Current.Request["messageText"]);
				}

				if (errorOccured)
				{
					LogError(order, "At least one validation error exists - exiting callback routine.");
					order.TransactionStatus = "Failed";
					order.Save();
				}
				else
				{
					LogEvent(order, "Payment succeeded with transaction number {0}", Dynamicweb.Context.Current.Request["trnId"]);

					order.TransactionAmount = Math.Round(order.Price.Price, decimals);
					order.TransactionStatus = "Succeeded";
					order.TransactionCardType = Dynamicweb.Context.Current.Request["cardType"];
					order.TransactionCardNumber = HideCardNumber(Dynamicweb.Context.Current.Request["trnCardNumber"]);

					if (TransactionType == "P")
					{
						string infoTxt = "Capture succeeded";
						LogEvent(order, infoTxt);
						order.CaptureInfo = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, infoTxt);
					}

					SetOrderComplete(order, Dynamicweb.Context.Current.Request["trnId"]);
				}
			}
			finally
			{
				if (!orderWasCompleted && order.Complete)
				{
					CheckoutDone(order);
				}
			}
		}

		private string SendRequest(string serviceUrl, string postData)
		{
			var byteArray = Encoding.UTF8.GetBytes(postData);
			var webRequest = (HttpWebRequest)WebRequest.Create(serviceUrl);
			webRequest.Method = "POST";
			webRequest.ContentLength = byteArray.Length;
			webRequest.ContentType = "application/x-www-form-urlencoded";

			var myWriter = webRequest.GetRequestStream();
			myWriter.Write(byteArray, 0, byteArray.Length);
			myWriter.Close();

			var response = (HttpWebResponse)webRequest.GetResponse();
			using (StreamReader responseStream = new StreamReader(response.GetResponseStream()))
			{
				return responseStream.ReadToEnd();
			}
		}

		private NameValueCollection ParseResponse(string inputString)
		{
			if (inputString.IndexOf("?") > 0)
			{
				inputString = inputString.Substring(inputString.IndexOf("?") + 1);
			}

			var result = new NameValueCollection();
			foreach (var value in inputString.Split(Convert.ToChar("&")))
			{
				var pair = value.Split(Convert.ToChar("="));
				result.Add(
					WebUtility.UrlDecode(pair[0].Trim()),
					pair.Length > 1 ? pair[1].Trim() : WebUtility.UrlDecode(pair[0].Trim())
				);
			}

			return result;
		}

		private NameValueCollection ParseRawForm()
		{
			var bytes = new byte[Context.Current.Request.InputStream.Length];
            Context.Current.Request.InputStream.Read(bytes, 0, bytes.Length);
            Context.Current.Request.InputStream.Position = 0;

			string content = Encoding.UTF8.GetString(bytes);
			return ParseResponse(content);
		}
	}
}
