﻿using Dynamicweb.Configuration;
using Dynamicweb.Core;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Environment;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Frontend;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.Buckaroo
{
    /// <summary>
    /// Buckaroo Checkout Handler
    /// </summary>
    [
        AddInName("Buckaroo"),
        AddInDescription("Buckaroo Checkout Handler")
    ]
    public class Buckaroo : CheckoutHandlerWithStatusPage, IDropDownOptions
    {
        #region Addin parameters

        [AddInParameter("Test Mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        [AddInParameter("Merchant key"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string MerchantKey { get; set; }

        [AddInParameter("Secret key"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string SecretKey { get; set; }

        [AddInParameter("Payment methods"), AddInParameterEditor(typeof(CheckListParameterEditor), "SortBy=Value;")]
        public string PaymentMethods { get; set; }

        [AddInParameter("Language"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string Language { get; set; }

        [AddInParameter("Process Page"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
        public string ProcessPage { get; set; }

        [AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Buckaroo/Post")]
        public string PostTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Buckaroo/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Buckaroo/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Success template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/Buckaroo/Success")]
        public string SuccessTemplate { get; set; }

        #endregion

        #region StartCheckout

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            string paymentMode = TestMode ? "1" : "0";

            try
            {
                LogEvent(order, "Checkout started");

                var formValues = new Dictionary<string, string> {
                    { "BPE_Merchant", MerchantKey },
                    { "BPE_Mode",  paymentMode },
                    { "BPE_Currency", order.Price.Currency.Code },
                    { "BPE_Description", string.Format("{0} - {1}", order.Id, order.CustomerName) },
                    { "BPE_Language", DefaultLanguageCode },
                    { "BPE_Invoice", order.Id },
                    { "BPE_InvoiceVat", "0" },
                    { "BPE_Amount", order.Price.PricePIP.ToString() },
                    { "BPE_PaymentMethodsAllowed", PaymentMethods.Replace(",", "; ") },
                    { "BPE_CustomerId", order.CustomerAccessUserId.ToString() },
                    { "BPE_CustomerFirstName", Converter.ToString(order.CustomerName) },
                    { "BPE_CustomerLastName", string.Empty },
                    { "BPE_CustomerGender", "0" },
                    { "BPE_CustomerAddress1", Converter.ToString(order.CustomerAddress) },
                    { "BPE_CustomerZipCode", Converter.ToString(order.CustomerZip) },
                    { "BPE_CustomerCity",  Converter.ToString(order.CustomerCity) },
                    { "BPE_CustomerCountry", Converter.ToString(order.CustomerCountry) },
                    { "BPE_CustomerMail",  Converter.ToString(order.CustomerEmail) },
                    { "BPE_CustomerPhone", Converter.ToString(order.CustomerPhone) },
                    { "BPE_CustomerState", Converter.ToString(order.CustomerRegion) },
                    { "BPE_CustomerFax", Converter.ToString(order.CustomerFax) },
                    { "BPE_Reference", order.Id },
                    { "BPE_Return_Type", "PAGE" },
                    { "BPE_Signature2", StringHelper.Md5HashToString(string.Format("{0}{1}{2}{3}{4}{5}", MerchantKey, order.Id, order.Price.PricePIP, order.Price.Currency.Code, paymentMode, SecretKey)) },
                    { "BPE_Return_Success", SuccessUrl(order) },
                    { "BPE_Return_Cancel", CancelUrl(order) },
                    { "BPE_Return_Reject", CancelUrl(order) },
                    { "BPE_Return_Error", ErrorUrl(order) },
                    { "BPE_Return_Method", "POST" },
                    { "BPE_Style", "POP" },
                    { "BPE_Autoclose_Popup", "FALSE" },
                }.Where(x => x.Value.Length > 0).ToDictionary(x => x.Key, x => x.Value);

                LogEvent(order, "Render Buckaroo template");

                var formTemplate = new Template(PostTemplate);
                formTemplate.SetTag("Buckaroo.GatewayUrl", ProcessPage);
                foreach (var formValue in formValues)
                {
                    formTemplate.SetTag(formValue.Key.Replace("BPE_", "Buckaroo."), formValue.Value);
                }

                var renderer = new Ecommerce.Frontend.Renderer(PageView.Current());
                renderer.RenderOrder(order, formTemplate);

                return formTemplate.Output();
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        #endregion

        /// <summary>
        /// Redirects user to payment step or canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(null, "Redirected to Buckaroo CheckoutHandler");

                switch (Converter.ToString(Context.Current.Request["BuckarooState"]))
                {
                    case "Success":
                        return StateSuccess(order);

                    case "Error":
                        return StateError(order);

                    case "Cancel":
                        return StateCancel(order);

                    default:
                        Callback(order);
                        return null;
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        #region Form properties

        private static string BaseUrl(Order order)
        {
            var disablePortNumber = SystemConfiguration.Instance.GetBoolean("/Globalsettings/System/http/DisableBaseHrefPort");
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);
            var pageId = Context.Current.Request["ID"] == null ? string.Empty : string.Format("ID={0}&", Context.Current.Request["ID"]);

            return string.Format("{2}://{0}{3}/Default.aspx?{1}{5}={4}", Context.Current.Request.Url.Host, pageId, Context.Current.Request.Url.Scheme, portString, order.Id, OrderIdRequestName);
        }

        private static string SuccessUrl(Order order)
        {
            return string.Format("{0}&BuckarooState=Success", BaseUrl(order));
        }

        private static string ErrorUrl(Order order)
        {
            return string.Format("{0}&BuckarooState=Error", BaseUrl(order));
        }
        private static string CancelUrl(Order order)
        {
            return string.Format("{0}&BuckarooState=Cancel", BaseUrl(order));
        }

        private static string DefaultLanguageCode
        {
            get
            {
                var currentLanguageCode = ExecutingContext.GetCulture(true).TwoLetterISOLanguageName;
                var supportedLanguageCodes = new List<string>(new[] { "en", "nl", "de", "fr" });

                return supportedLanguageCodes.Contains(currentLanguageCode) ? currentLanguageCode : "en";
            }
        }

        private string PrintErrorTemplate(Order order)
        {
            LogEvent(order, "Printing error template");

            var errorTemplate = new Template(ErrorTemplate);
            var orderRenderer = new Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(errorTemplate, order, true);

            return errorTemplate.Output();
        }

        private string PrintSuccessTemplate(Order order)
        {
            LogEvent(order, "Printing success template");

            var successTemplate = new Template(SuccessTemplate);
            var orderRenderer = new Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(successTemplate, order, true);

            return successTemplate.Output();
        }

        private string StateSuccess(Order order)
        {
            LogEvent(order, "State success");

            if (order.Complete)
            {
                RedirectToCart(order);
            }
            else
            {
                return PrintSuccessTemplate(order);
            }

            return null;
        }

        private string StateError(Order order)
        {
            LogEvent(order, "State error");

            order.TransactionStatus = "Error";
            Services.Orders.Save(order);

            CheckoutDone(order);

            return PrintErrorTemplate(order);
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            var orderRenderer = new Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(cancelTemplate, order, true);

            return cancelTemplate.Output();
        }

        #endregion

        #region Callback
        private void Callback(Order order)
        {
            if (order.Complete)
            {
                LogEvent(order, "Callback started,  but order already completed");
                return;
            }

            try
            {
                LogEvent(order, "Callback started");

                var errorOccured = false;
                string md5Sign = StringHelper.Md5HashToString(string.Format("{0}{1}{2}{3}{4}{5}{6}{7}{8}{9}",
                        Context.Current.Request["bpe_trx"],
                        Context.Current.Request["bpe_timestamp"],
                        MerchantKey,
                        order.Id,                       // Base.Request("bpe_invoice"),
                        order.Id,                       // Base.Request("bpe_reference"), 
                        order.Price.Currency.Code,      // Base.Request("bpe_currency"), 
                        order.Price.PricePIP.ToString(),//Base.Request("bpe_amount"), 
                        Context.Current.Request["bpe_result"],
                        Context.Current.Request["bpe_mode"],
                        SecretKey
                    )
                );

                if (!md5Sign.Equals(Context.Current.Request["bpe_signature2"], StringComparison.CurrentCultureIgnoreCase))
                {
                    LogError(
                        order,
                        "The md5 checksum returned from callback does not match: Callback: {0}, calculated: {1}",
                        Context.Current.Request["bpe_signature2"], md5Sign
                    );

                    errorOccured = true;
                }

                if (Context.Current.Request["bpe_amount"] != order.Price.PricePIP.ToString())
                {
                    LogError(
                        order,
                        "The amount returned from callback does not match the amount set on the order: Callback: {0}, order: {1}",
                        Context.Current.Request["bpe_amount"], order.Price.PricePIP
                    );

                    errorOccured = true;
                }

                if (Context.Current.Request["bpe_currency"] != order.Price.Currency.Code)
                {
                    LogError(
                        order,
                        "The currency return from callback does not match the amount set on the order: Callback: {0}, order: {1}",
                        Context.Current.Request["bpe_currency"], order.Price.Currency.Code
                    );

                    errorOccured = true;
                }

                if (errorOccured)
                {
                    LogError(order, "At least one callback validation error exists - exiting callback routine.");
                }
                else
                {
                    order.TransactionNumber = Context.Current.Request["bpe_trx"];
                    order.GatewayPaymentStatus = ConvertBuckarooStatusToDwStatus(Context.Current.Request["bpe_result"]);

                    if (order.GatewayPaymentStatus == GatewayPaymentStatus.Paid)
                    {
                        LogEvent(order, "Payment {0} succeeded with transaction number {1}", order.Id, order.TransactionNumber);

                        order.TransactionAmount = order.Price.Price;
                        order.TransactionStatus = "Succeeded";
                        SetOrderComplete(order);
                    }
                    else
                    {
                        order.TransactionStatus = "Failed";
                        order.GatewayResult = GetInfoFromQueryString(order);
                        Services.Orders.UpdateGatewayResult(order, false);
                        Services.Orders.Save(order);
                    }
                }

            }
            finally
            {
                if (order.Complete)
                {
                    CheckoutDone(order);
                }
            }
        }
        #endregion

        #region Helper function to convert status
        private string ConvertBuckarooStatusToDwStatus(string gatewayPaymentStatusResult)
        {
            switch (gatewayPaymentStatusResult.Trim())
            {
                case "0": // Pending
                case "1":
                    return GatewayPaymentStatus.Hold;
                case "70": // In progress
                case "71":
                    return GatewayPaymentStatus.Started;
                case "72": // Failed
                    return GatewayPaymentStatus.Rejected;
                case "100": // Paid
                    return GatewayPaymentStatus.Paid;
                case "101": // Failed
                case "102":
                case "103":
                    return GatewayPaymentStatus.Rejected;
                case "121": // Paid+
                case "151":
                    return GatewayPaymentStatus.Paid;
                case "171": // Paid
                case "172":
                case "173":
                    return GatewayPaymentStatus.Paid;
                case "181": // Failed
                case "182":
                case "183":
                    return GatewayPaymentStatus.Rejected;
                case "201": // Pending
                    return GatewayPaymentStatus.Hold;
                case "203": // Failed
                case "204":
                case "205":
                case "206":
                case "207":
                    return GatewayPaymentStatus.Rejected;
                case "300": // Pending
                    return GatewayPaymentStatus.Hold;
                case "301": // Paid
                    return GatewayPaymentStatus.Paid;
                case "302": // Failed
                case "303":
                    return GatewayPaymentStatus.Rejected;
                case "304": // In progress
                case "305":
                    return GatewayPaymentStatus.Started;
                case "306": // Paid
                    return GatewayPaymentStatus.Paid;
                case "309": // Failed
                case "371":
                case "372":
                case "373":
                case "381":
                case "382":
                case "383":
                    return GatewayPaymentStatus.Rejected;
                case "390": // Paid
                    return GatewayPaymentStatus.Paid;
                case "400": // In progress
                    return GatewayPaymentStatus.Started;
                case "401": // Paid
                    return GatewayPaymentStatus.Paid;
                case "402": // Failed
                case "409":
                case "410":
                    return GatewayPaymentStatus.Rejected;
                case "411": // Paid
                    return GatewayPaymentStatus.Paid;
                case "414": // Failed
                case "421":
                case "422":
                case "425":
                    return GatewayPaymentStatus.Rejected;
                case "461": // Paid+
                case "462":
                case "463":
                case "464":
                case "471": // Paid
                    return GatewayPaymentStatus.Paid;
                case "472": // Failed
                    return GatewayPaymentStatus.Rejected;
                case "500": // Pending
                    return GatewayPaymentStatus.Hold;
                case "560": // In progress
                    return GatewayPaymentStatus.Started;
                case "600": // Pending
                    return GatewayPaymentStatus.Hold;
                case "601": // Paid
                    return GatewayPaymentStatus.Paid;
                case "602": // Failed
                case "605":
                case "609":
                case "610":
                case "612":
                    return GatewayPaymentStatus.Rejected;
                case "671": // In progress
                    return GatewayPaymentStatus.Started;
                case "672": // Failed
                    return GatewayPaymentStatus.Rejected;
                case "700": // In progress
                    return GatewayPaymentStatus.Started;
                case "701": // Paid
                    return GatewayPaymentStatus.Paid;
                case "702": // Failed
                case "705":
                    return GatewayPaymentStatus.Rejected;
                case "710": // Pending
                    return GatewayPaymentStatus.Hold;
                case "711": // Failed
                    return GatewayPaymentStatus.Rejected;
                case "712": // Paid
                    return GatewayPaymentStatus.Paid;
                case "720": // Failed
                case "721":
                    return GatewayPaymentStatus.Rejected;
                case "800": // Pending
                    return GatewayPaymentStatus.Hold;
                case "801": // Paid
                    return GatewayPaymentStatus.Paid;
                case "802": // Failed
                case "803":
                case "804":
                case "810":
                    return GatewayPaymentStatus.Rejected;
                case "811": // In progress
                    return GatewayPaymentStatus.Started;
                case "812": // Failed
                case "813":
                    return GatewayPaymentStatus.Rejected;
                case "814": // Pending
                    return GatewayPaymentStatus.Hold;
                case "815": // Failed
                case "816":
                    return GatewayPaymentStatus.Rejected;
                case "820": // In progress
                    return GatewayPaymentStatus.Started;
                case "821": // Paid
                    return GatewayPaymentStatus.Paid;
                case "822": // Failed
                case "823":
                case "824":
                case "830":
                    return GatewayPaymentStatus.Rejected;
                case "831": // In progress
                    return GatewayPaymentStatus.Started;
                case "833": // Failed
                case "834":
                case "835":
                case "836":
                case "871":
                case "872":
                case "873":
                case "874":
                case "900":
                case "901":
                case "910":
                case "931":
                case "932":
                case "933":
                case "934":
                case "935":
                case "940":
                case "941":
                case "942":
                case "943":
                case "944":
                case "945":
                case "946":
                case "947":
                case "948":
                case "949":
                case "950":
                case "951":
                case "952":
                case "953":
                case "954":
                case "955":
                case "960":
                case "961":
                case "962":
                case "963":
                case "978":
                case "980":
                case "990":
                case "991":
                case "992":
                case "993":
                case "999":
                    return GatewayPaymentStatus.Rejected;

                default:
                    return GatewayPaymentStatus.Unknown;
            }
        }
        #endregion

        #region GetOptions ( IDropDownOptions )

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="name"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string name)
        {
            try
            {
                switch (name)
                {
                    case "Post Mode":
                        return new Hashtable {
                            { "Auto", Translator.Translate("Auto post (does not use the template)") },
                            { "Template", Translator.Translate("Render template") }
                        };

                    case "Payment methods":
                        var methods = new Dictionary<string, string> {
                            { "creditcard", "Creditcard" },
                            { "ideal", "iDEAL" },
                            { "collect", "Direct Debit" },
                            { "transfer", "Banktransfer" },
                            { "giropay", "Giropay" },
                        }.ToDictionary(
                            x => x.Key,
                            x => Translator.Translate(x.Value)
                        );

                        var returnHashTable = new Hashtable();
                        foreach (var key in methods.Keys)
                        {
                            returnHashTable.Add(key, methods[key]);
                        }

                        return returnHashTable;

                    case "Language":
                        return new Hashtable {
                            { "NL", "NL" },
                            { "DE", "DE" },
                            { "FR", "FR" },
                            { "EN", "EN" }
                        };

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", name));
                }
            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }

        }

        #endregion
    }
}
