using System;
using System.IO;
using System.ServiceModel.Channels;
using System.Xml;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.CyberSource.SimpleOrderApi
{
    /// <summary>
    /// Provides methods for reading and writing encoded data
    /// </summary>
    public class CustomTextMessageEncoder : MessageEncoder
    {
        private CustomTextMessageEncoderFactory factory;
        private XmlWriterSettings writerSettings;

        /// <summary>
        /// Constacts custom text message encoder based on specific facory object
        /// </summary>
        /// <param name="factory">Provides message encoder and version</param>
        public CustomTextMessageEncoder(CustomTextMessageEncoderFactory factory)
        {
            this.factory = factory;
            writerSettings = new XmlWriterSettings();
        }

        /// <summary>
        /// Encoder content type
        /// </summary>
        public override string ContentType
        {
            get { return "text/xml"; }
        }

        /// <summary>
        /// Encoder media type
        /// </summary>
        public override string MediaType
        {
            get { return "text/xml"; }
        }

        /// <summary>
        /// Version of message
        /// </summary>
        public override MessageVersion MessageVersion
        {
            get { return factory.MessageVersion; }
        }

        /// <summary>
        /// Reads message form buffer of bytes array
        /// </summary>
        /// <param name="buffer">buffer with data</param>
        /// <param name="bufferManager"></param>
        /// <param name="contentType">Type of content</param>
        /// <returns>Message object with data</returns>
        public override Message ReadMessage(ArraySegment<byte> buffer, BufferManager bufferManager, string contentType)
        {
            var messageContent = new byte[buffer.Count];
            Array.Copy(buffer.Array, buffer.Offset, messageContent, 0, messageContent.Length);
            bufferManager.ReturnBuffer(buffer.Array);

            var stream = new MemoryStream(messageContent);
            return ReadMessage(stream, int.MaxValue);
        }

        /// <summary>
        /// Reads message from stream
        /// </summary>
        /// <param name="stream">stream with data</param>
        /// <param name="maxSizeOfHeaders"></param>
        /// <param name="contentType">Type of content</param>
        /// <returns>Message object with data</returns>	
        /// <remarks>
        /// We need to get rid of the security header because it is not signed by the web service.
        /// The whole reason for the custom Encoder is to do this. the client rejected the unsigned header.
        /// Our WCF client is set up to allow the absence of a security header but if the header exists then it must be signed.
        /// Hopefully the namespace will not change. Maybe it should be put in a config.
        /// </remarks>
        public override Message ReadMessage(Stream stream, int maxSizeOfHeaders, string contentType)
        {
            var streamReader = new StreamReader(stream);
            var wireResponse = streamReader.ReadToEnd();

            var xmlDocument = new XmlDocument();
            xmlDocument.LoadXml(wireResponse);

            var xmlNavigator = xmlDocument.CreateNavigator();
            if (xmlNavigator.MoveToFollowing("Security", "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"))
            {
                xmlNavigator.DeleteSelf();
            }

            var reader = XmlReader.Create(new StringReader(xmlDocument.InnerXml));
            return Message.CreateMessage(reader, maxSizeOfHeaders, MessageVersion.Soap11);
        }

        /// <summary>
        /// Writes message data to stream using offset value and message length restriction
        /// </summary>
        /// <param name="message">Message with text</param>
        /// <param name="maxMessageSize">maximum size of message text</param>
        /// <param name="bufferManager">Buffer manager</param>
        /// <param name="messageOffset">offset value</param>
        /// <returns>ArraySegment of bytes</returns>
        public override ArraySegment<byte> WriteMessage(Message message, int maxMessageSize, BufferManager bufferManager, int messageOffset)
        {
            var stream = new MemoryStream();
            var writer = XmlWriter.Create(stream, this.writerSettings);
            message.WriteMessage(writer);
            writer.Close();

            byte[] messageBytes = stream.GetBuffer();
            int messageLength = (int)stream.Position;
            stream.Close();

            int totalLength = messageLength + messageOffset;
            byte[] totalBytes = bufferManager.TakeBuffer(totalLength);
            Array.Copy(messageBytes, 0, totalBytes, messageOffset, messageLength);
            ArraySegment<byte> byteArray = new ArraySegment<byte>(totalBytes, messageOffset, messageLength);

            return byteArray;
        }

        /// <summary>
        /// Writes message data to stream
        /// </summary>
        /// <param name="message">Message instance with data</param>
        /// <param name="stream">Data stream object</param>
        public override void WriteMessage(Message message, Stream stream)
        {
            var writer = XmlWriter.Create(stream, writerSettings);
            message.WriteMessage(writer);
            writer.Close();
        }
    }
}
