﻿using Dynamicweb.Core;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.DIBS
{
    /// <summary>
    /// DIBS Checkout Handler
    /// </summary>
    [
        AddInName("DIBS Flexwin"),
        AddInDescription("DIBS Flexwin Checkout Handler")
    ]
    public class Dibs : CheckoutHandler, IDropDownOptions, IRemotePartialCapture, ISavedCard, IRecurring
    {
        private enum PostModes { Auto, Template }
        private enum ErrorType { Undefined, SavedCard }
        private enum MitType { RECURRING, UCOF, INCREMENTAL }

        private PostModes postMode = PostModes.Auto;

        #region Editors

        [AddInParameter("Merchant"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Merchant { get; set; }

        [AddInParameter("Key1"), AddInParameterEditor(typeof(TextParameterEditor), "size=50")]
        public string Key1 { get; set; }

        [AddInParameter("Key2"), AddInParameterEditor(typeof(TextParameterEditor), "size=50")]
        public string Key2 { get; set; }

        [AddInParameter("Card type"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=true; nonetext=All; SortBy=Value;")]
        public string PayType { get; set; }

        [AddInParameter("Post mode"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PostModeSelection
        {
            get
            {
                return postMode.ToString();
            }
            set
            {
                switch (value)
                {
                    case "Auto":
                        postMode = PostModes.Auto;
                        break;

                    case "Template":
                        postMode = PostModes.Template;
                        break;
                }
            }
        }

        [AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Post")]
        public string PostTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Color in DIBS window"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value")]
        public string DibsColor { get; set; }

        [AddInParameter("Add card fee by DIBS"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UseAddFee { get; set; }

        [AddInParameter("Use MD5 checksum"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UseMD5 { get; set; }

        [AddInParameter("Test Mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        #endregion

        #region Constructors

        public Dibs()
        {
            ServicePointManager.SecurityProtocol = ServicePointManager.SecurityProtocol | SecurityProtocolType.Tls12;
        }

        #endregion

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");

                var paymentType = PayType;
                if (TestMode)
                {
                    switch (PayType)
                    {
                        case "FSB":
                            paymentType = "swd_atest";
                            break;

                        case "SEB":
                            paymentType = "seb_atest";
                            break;

                        case "SHB":
                            paymentType = "shbtest";
                            break;
                    }
                }

                var formValues = new Dictionary<string, string> {
                    { "merchant", Merchant},
                    { "amount", order.Price.PricePIP.ToString()},
                    { "currency", order.Price.Currency.PayGatewayCode.ToString()},
                    { "orderid", order.Id},
                    { "accepturl", AcceptUrl(order)},
                    { "cancelurl", CancelUrl(order)},
                    { "callbackurl", CallbackUrl(order)},
                    { "paytype", paymentType},
                    { "uniqueoid", "true"},
                    { "calcfee", UseAddFee ? "yes" : ""},
                    { "test", TestMode ? "true" : ""},
                    { "lang", LanguageCode},
                    { "decorator", "responsive"},
                    { "color", DibsColor},
                    { "md5key", UseMD5 ? StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(Key1 + string.Format("merchant={0}&orderid={1}&currency={2}&amount={3}", Merchant, order.Id, order.Price.Currency.PayGatewayCode, order.Price.PricePIP))) : string.Empty },
                }.Where(x => x.Value.Length > 0).ToDictionary(x => x.Key, x => x.Value);

                if (order.IsRecurringOrderTemplate || order.DoSaveCardToken || !string.IsNullOrWhiteSpace(order.SavedCardDraftName))
                {
                    //formValues["maketicket"] = "true"; // Doesn't work on 12.12.2016. And as it doesn't support md5 check - it shouldn't be used at all 
                    formValues["preauth"] = "true";
                    formValues["orderid"] = $"{order.Id}_preauth";
                    formValues["md5key"] = UseMD5 ? StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(Key1 + string.Format("merchant={0}&orderid={1}_preauth&currency={2}&amount={3}", Merchant, order.Id, order.Price.Currency.PayGatewayCode, order.Price.PricePIP))) : string.Empty;
                    formValues["callbackurl"] = CallbackUrl(order, order.SavedCardDraftName);
                    if (order.IsRecurringOrderTemplate)
                    {
                        formValues["mitType"] = MitType.UCOF.ToString();
                    }
                    else
                    {
                        formValues["COF"] = "Yes";
                    }
                    formValues.Remove("captureNow");
                }

                switch (postMode)
                {
                    case PostModes.Auto:
                        LogEvent(order, "Auto post to DIBS");

                        SubmitForm("https://payment.architrade.com/paymentweb/start.action", formValues);
                        return string.Empty;

                    case PostModes.Template:
                        LogEvent(order, "Render template");

                        var formTemplate = new Template(PostTemplate);

                        foreach (var formValue in formValues)
                        {
                            formTemplate.SetTag(string.Format("Dibs.{0}", formValue.Key), formValue.Value);
                        }

                        var renderer = new Ecommerce.Frontend.Renderer(Dynamicweb.Frontend.PageView.Current());
                        renderer.RenderOrder(order, formTemplate);

                        return formTemplate.Output();

                    default:
                        LogError(order, "Unhandled post mode: '{0}'", postMode);
                        throw new ArgumentException(string.Format("Unhandled post mode: '{0}'", postMode));
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        /// <summary>
        /// Redirects user to payment step or canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(null, "Redirected to Dibs CheckoutHandler");

                var dibsStateValue = Context.Current.Request["DibsState"];

                if (dibsStateValue == null)
                {
                    LogError(null, "The request variable 'DibsState' is not present");
                    return string.Empty;
                }

                switch (Converter.ToString(dibsStateValue))
                {
                    case "Accept":
                        return StateOk(order);

                    case "Cancel":
                        return StateCancel(order);

                    case "Callback":
                        Callback(order);
                        Context.Current.Response.End();
                        return null;

                    default:
                        LogError(order, "Unknown DibsState: '{0}'", Converter.ToString(dibsStateValue));
                        return string.Empty;
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        #region Form properties

        private static string BaseUrl(Order order)
        {
            var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);
            var pageId = Context.Current.Request["ID"] == null ? string.Empty : string.Format("ID={0}&", Context.Current.Request["ID"]);

            return string.Format("{2}://{0}{3}/Default.aspx?{1}{5}={4}", Context.Current.Request.Url.Host, pageId, Context.Current.Request.Url.Scheme, portString, order.Id, OrderIdRequestName);
        }
        private static string AcceptUrl(Order order)
        {
            return string.Format("{0}&DibsState=Accept", BaseUrl(order));
        }
        private static string CancelUrl(Order order)
        {
            return string.Format("{0}&DibsState=Cancel", BaseUrl(order));
        }
        private static string CallbackUrl(Order order, string cardName = "")
        {
            if (string.IsNullOrEmpty(cardName))
            {
                return string.Format("{0}&DibsState=Callback", BaseUrl(order));
            }
            else
            {
                return string.Format("{0}&DibsState=Callback&SavedCardName={1}", BaseUrl(order), cardName);
            }
        }

        private static string LanguageCode
        {
            get
            {
                var currentLanguageCode = Dynamicweb.Environment.ExecutingContext.GetCulture(true).TwoLetterISOLanguageName;
                currentLanguageCode = currentLanguageCode == "nb" || currentLanguageCode == "nn" ? "no" : currentLanguageCode;
                var supportedLanguageCodes = new List<string>(new[] { "da", "sv", "no", "en", "nl", "de", "fr", "fi", "es", "it", "fo", "pl" });
                return supportedLanguageCodes.Contains(currentLanguageCode) ? currentLanguageCode : "en";
            }
        }

        #endregion

        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");

            if (order.Complete)
            {
                RedirectToCart(order);
                return null;
            }

            if (Common.Context.Cart == null)
            {
                CheckoutDone(order);
            }

            LogError(order, "Called State ok, but order is not set complete - this should have happened in the callback.");

            return PrintErrorTemplate(order);
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(cancelTemplate, order, true);

            return cancelTemplate.Output();
        }

        private static void SetOrderSucceeded(Order order, bool success)
        {
            if (success)
            {
                order.TransactionAmount = order.Price.PricePIP / 100d;
                order.TransactionStatus = "Succeeded";
            }
            else
            {
                order.TransactionStatus = "Failed";
            }
            Services.Orders.Save(order);
        }

        private void Callback(Order order)
        {
            var parameters = Context.Current.Request.Params;
            var statusCode = parameters["statuscode"];
            if (string.IsNullOrEmpty(statusCode))
            {
                LogEvent(order, "Callback started");
            }
            else
            {
                LogEvent(order, "Callback started. Status code: {0}", statusCode);
            }

            try
            {
                ProcessOrder(order, parameters);
            }
            finally
            {
                CheckoutDone(order);
            }
        }

        private void ProcessOrder(Order order, NameValueCollection parameters)
        {
            var errorOccured = false;
            var transactionID = Converter.ToString(parameters["transact"]);

            if (string.IsNullOrEmpty(transactionID))
            {
                errorOccured = true;
                LogEvent(order, "No transaction number sent to callback");
            }

            long feeAmount = 0;
            var fee = parameters["fee"];
            long.TryParse(fee, out feeAmount);

            var statusCode = parameters["statuscode"];
            if (UseMD5)
            {
                var calculatedMD5 = "";
                if (string.Equals(statusCode, "13", StringComparison.OrdinalIgnoreCase))
                {
                    calculatedMD5 = StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(
                        $"{Key1}transact={transactionID}&preauth=true&currency={order.Price.Currency.PayGatewayCode}"));
                }
                else
                {
                    calculatedMD5 = StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(
                        $"{Key1}transact={transactionID}&amount={order.Price.PricePIP + feeAmount}&currency={order.Price.Currency.PayGatewayCode}"));
                }

                var authKey = Converter.ToString(parameters["authkey"]);
                if (!calculatedMD5.Equals(authKey, StringComparison.InvariantCultureIgnoreCase))
                {
                    errorOccured = true;
                    LogEvent(order, "MD5 sums do not match: Callback: {0}, calculated: {1}. Status code: {2}", authKey, calculatedMD5, statusCode);
                }
            }

            // Check statusCode {https://tech.dibspayment.com/nodeaddpage/toolboxstatuscodes}
            // Don't forget to configure the return parameters in the DIBS administration
            if (!errorOccured && !string.IsNullOrEmpty(statusCode))
            {
                if (string.Equals(statusCode, "2", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "5", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "12", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "13", StringComparison.OrdinalIgnoreCase))
                {
                    // Success
                }
                else if (string.Equals(statusCode, "0", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "1", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "4", StringComparison.OrdinalIgnoreCase) ||
                    string.Equals(statusCode, "17", StringComparison.OrdinalIgnoreCase))
                {
                    errorOccured = true;
                    LogEvent(order, "Order failed: Error status code: {0}", statusCode);
                }
                else
                {
                    errorOccured = true;
                    LogEvent(order, "Order failed: Unknown status code: {0}", statusCode);
                }
            }

            if (errorOccured)
            {
                LogError(order, "At least one callback validation error exists - exiting callback routine. Transaction number: {0}", transactionID);

                SetOrderSucceeded(order, false);
            }
            else
            {
                LogEvent(order, "Payment succeeded with transaction number {0}", transactionID);

                // Adding fee from DIBS
                if (feeAmount != 0)
                {
                    LogEvent(order, "Saving card fee '{0}'", feeAmount);
                    order.ExternalPaymentFee = feeAmount / 100d;
                }

                if (order.SavedCardId == 0)
                {
                    order.TransactionCardType = Converter.ToString(parameters["paytype"]);
                    order.TransactionCardNumber = HideCardNumber(Converter.ToString(parameters["cardnomask"]));
                }

                // Status code 13 - The transaction is a "ticket" transaction.
                if (string.Equals(statusCode, "13", StringComparison.OrdinalIgnoreCase))
                {
                    var savedCardName = Context.Current.Request["SavedCardName"];
                    if (string.IsNullOrEmpty(savedCardName))
                    {
                        savedCardName = order.Id;
                    }

                    var tiket = transactionID;

                    if (order.CustomerAccessUserId > 0)
                    {
                        var savedCard = Services.PaymentCard.CreatePaymentCard(order.CustomerAccessUserId, order.PaymentMethodId, order.LanguageId, savedCardName, order.TransactionCardType, order.TransactionCardNumber, transactionID);
                        order.SavedCardId = savedCard.ID;

                        LogEvent(order, "Saved card created: {0}", savedCardName);
                    }
                    else
                    {
                        order.TransactionToken = tiket;
                    }

                    transactionID = ""; // This is ticket token, when Status code 13
                    if (!order.IsRecurringOrderTemplate)
                    {
                        UseSavedCardInternal(order, tiket);
                    }
                }

                SetOrderComplete(order, transactionID);
                SetOrderSucceeded(order, true);
            }
        }

        private string PrintErrorTemplate(Order order, string msg = "", ErrorType errorType = ErrorType.Undefined)
        {
            LogEvent(order, "Printing error template");

            var errorTemplate = new Template(ErrorTemplate);
            errorTemplate.SetTag("CheckoutHandler:ErrorType", errorType.ToString());
            errorTemplate.SetTag("CheckoutHandler:ErrorMessage", msg);

            return Render(order, errorTemplate);
        }

        #region IDropDownOptions

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="behaviorMode"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string behaviorMode)
        {
            try
            {
                switch (behaviorMode)
                {
                    case "Post mode":
                        return new Hashtable {
                            { "Auto", Translate.Translate("Auto post (does not use the template)") },
                            { "Template", Translate.Translate("Render template") }
                    };

                    case "Card type":
                        var cardTypes = new Dictionary<string, string> {
                            { "ACCEPT", "Accept card" },
                            { "AMEX", "American Express" },
                            { "AMEX(DK)", "American Express (Danish card)" },
                            { "DIN", "Diners Club" },
                            { "DIN(DK)", "Diners Club (Danish card)" },
                            { "DK", "Dankort" },
                            { "MC", "Mastercard" },
                            { "MC(DK)", "Mastercard (Danish card)" },
                            { "MC(SE)", "Mastercard (Swedish card)" },
                            { "MTRO", "Maestro" },
                            { "MTRO(DK)", "Maestro (Danish card)" },
                            { "MTRO(UK)", "Maestro (UK)" },
                            { "NSBK", "Nordea Bank card" },
                            { "SBSBK", "Skandiabanken Bank card" },
                            { "SEBSBK", "Swedish bank card (SEB)" },
                            { "V-DK", "Visa/Dankort" },
                            { "VISA", "Visa" },
                            { "VISA(DK)", "Visa (Danish card)" },
                            { "VISA(SE)", "Visa (Swedish card)" },
                            { "ELEC", "Visa Electron" },
                            { "DNB", "Danske Netbetaling (Danske Bank)" },
                            { "FSB", "Swedbank Direktbetaling" },
                            { "NDB", "Nordea Solo-E betaling (Sweden)" },
                            { "RBS", "Royal Bank of Scotland iDeal Payment" },
                            { "SEB", "SEB Direktbetalning" },
                            { "SHB", "SHB Direktbetalning" },
                            { "SOLO", "Nordea Solo-E betaling" },
                            { "BAX", "BankAxess" },
                            { "MPO_Nets", "MobilePay Online" },
                        }.ToDictionary(
                            x => x.Key,
                            x => Translate.Translate(x.Value)
                        );

                        var returnHashTable = new Hashtable();
                        foreach (var key in cardTypes.Keys)
                        {
                            returnHashTable.Add(key, cardTypes[key]);
                        }

                        return returnHashTable;

                    case "Color in DIBS window":
                        return new Hashtable {
                            { "sand", Translate.Translate("Sand") },
                            { "grey", Translate.Translate("Grey") },
                            { "blue", Translate.Translate("Blue") }
                        };

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", behaviorMode));
                }
            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }
        }

        #endregion

        #region IRemoteCapture

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        /// <remarks>http://tech.dibspayment.com/capturecgi</remarks>
        public OrderCaptureInfo Capture(Order order)
        {
            return Capture(order, order.Price.PricePIP + PriceHelper.ConvertToPIP(order.Currency, order.ExternalPaymentFee), false);
        }

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <param name="amount">Amount to be transferred. The amount must be given in smallest possible unit, e.g. cents 100 = USD 1.</param>
        /// <param name="final">Shows that this step is final</param>
        /// <returns>Response from transaction service</returns>
        /// <remarks>
        ///     Create byte array with POST data
        ///		merchant       The merchant parameter is the shops DIBS identification number. The merchant number is available in the order confirmation email and is a 7-digit integer.
        ///		amount         Amount to be transferred. The amount must be given in smallest possible unit, e.g. cents 100 = USD 1.
        ///		transact       The unique DIBS identification number for the transaction you want to capture. The transact is returned from a successful authorisation, and is a as minimum 6-digit integer , e.g. transact=123456.
        ///		orderid        The shops (unique) order identification number as given in the authorisation process.
        ///		[textreply]    If this variable is set (e.g. textreply=true) the DIBS server returns its answer  in simple text format. If you are not using the standard DIBS payment window, e.g. using server-to-server requests, this significantly simplifies parsing the answer and is recommended.
        ///		[force]        If force=true is posted, the DIBS server will attempt to capture otherwise expired authorisations (more than 7 days old). This is achieved through an initial re-authorisation of the original transaction, followed by a capture process.
        ///		[account]      If several departments in a company use the same acquirer contract it may be beneficial to keep track of the transactions by department. Including the account parameter enables separation of the transaction by department.
        ///		[md5key]       This variable enables a MD5 key control of the values received by DIBS. This control  confirms that the values sent to DIBS has not been tampered with during the transfer. The MD5 key is calculated as: MD5(Key2 + MD5(Key1 + "merchant=&lt;merchant&gt;&orderid=&lt;orderid&gt; &transact=&lt;transact&gt;&amount=&lt;amount&gt;")) Where key1 and key2 are shop specific keys available through the DIBS administration interface, and + is the concatenation operator. NB! MD5 key check must also be enabled through the DIBS administration interface in order to work. Further details on MD5-key control .
        ///</remarks>
        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            try
            {
                if (order == null || string.IsNullOrEmpty(order.Id))
                {
                    const string errorText = "No valid Order object set";
                    LogError(order, errorText);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
                }

                if (string.IsNullOrEmpty(order.TransactionNumber))
                {
                    const string errorText = "No transaction number set on the order";
                    LogEvent(order, errorText);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
                }

                var rqstCapture = new Dictionary<string, string>()
                {
                    {"merchant", Merchant},
                    {"amount", amount.ToString()},
                    {"transact", order.TransactionNumber},
                    {"orderid", order.Id}
                };

                var authorizedAmount = order.Price.PricePIP + PriceHelper.ConvertToPIP(order.Currency, order.ExternalPaymentFee);
                if (authorizedAmount != amount)
                {
                    if (final)
                    {
                        rqstCapture.Add("close", "true");
                    }
                    rqstCapture.Add("split", "true");
                }
                else
                {
                    rqstCapture.Add("force", "true");
                }

                if (UseMD5)
                {
                    rqstCapture.Add("md5key",
                        StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(Key1 +
                            string.Format(
                                "merchant={0}&orderid={1}&transact={2}&amount={3}",
                                Merchant, order.Id, order.TransactionNumber, amount
                            )
                        ))
                    );
                }

                var responseText = ExecutePostRequest(order, "cgi-bin/capture.cgi", rqstCapture);

                var statusCodes = new Dictionary<string, string> {
                    { "0", "Accepted" },
                    { "1", "No response from acquirer." },
                    { "2", "Error in the parameters sent to the DIBS server. An additional parameter called 'message' is returned, with a value that may help identifying the error." },
                    { "3", "Credit card expired." },
                    { "4", "Rejected by acquirer." },
                    { "5", "Authorisation older than7 days." },
                    { "6", "Transaction status on the DIBS server does not allow capture." },
                    { "7", "Amount too high." },
                    { "8", "Amount is zero." },
                    { "9", "Order number (orderid) does not correspond to the authorisation order number." },
                    { "10", "Re-authorisation of the transaction was rejected." },
                    { "11", "Not able to communicate with the acquier." },
                    { "15", "Capture was blocked by DIBS." }
                };

                var statusMatch = Regex.Match(responseText, "status=(?<status>[^&]+)");
                if (statusMatch.Success)
                {
                    var status = statusMatch.Groups["status"].Value;

                    if (status == "ACCEPTED")
                    {
                        if (authorizedAmount == amount)
                        {
                            LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
                        }
                        else if (final)
                        {
                            LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture(final)", amount / 100f), DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Split capture successful");
                        }
                        else
                        {
                            LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture", amount / 100f), DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Split, "Split capture successful");
                        }
                    }
                    else
                    {
                        var resultMatch = Regex.Match(responseText, "result=(?<result>[^&]+)");
                        string result = resultMatch.Success ? resultMatch.Groups["result"].Value : "-1";

                        string message;
                        if (statusCodes.ContainsKey(result))
                        {
                            message = string.Format(
                                "Remote Capture failed. Statuscode: '{0}', description: '{1}', responseText: '{2}'",
                                result, statusCodes[result], responseText
                            );
                        }
                        else
                        {
                            message = string.Format(
                                "Remote Capture failed with unknown statuscode '{0}', responseText: '{1}'",
                                result, responseText
                            );
                        }

                        LogEvent(order, message, DebuggingInfoType.CaptureResult);
                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, message);
                    }
                }

                string eventTxt = string.Format(
                    "The text returned from DIBS did not contain the result state. ResponseText: {0}",
                    responseText
                );

                LogEvent(order, eventTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, eventTxt);
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        /// <summary>
        /// Shows if capture supported
        /// </summary>
        /// <param name="order">This object is not used in current implementation</param>
        /// <returns>This method always return 'true' value</returns>
        public bool CaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Shows if partial capture of the order supported
        /// </summary>
        /// <param name="order">instance of order</param>
        /// <returns>True, if partial capture of the order is supported</returns>
        public bool SplitCaptureSupported(Order order)
        {
            var supportedCards = new ArrayList { "DK", "MC", "MC(DK)", "MC(SE)", "V-DK", "VISA", "VISA(DK)", "VISA(SE)", "ELEC" };
            return supportedCards.Contains(order.TransactionCardType);
        }

        #endregion

        #region ISavedCard interface

        public void DeleteSavedCard(int savedCardId)
        {
            // Not implemented
            // https://login:password@payment.architrade.com/cgi-adm/delticket.cgi">
        }

        public string UseSavedCard(Order order)
        {
            try
            {
                UseSavedCardInternal(order);
                RedirectToCart(order);

                return string.Empty;
            }
            catch (System.Threading.ThreadAbortException)
            {
                // Do Nothing, this is expected
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogEvent(order, ex.Message, DebuggingInfoType.UseSavedCard);
                return PrintErrorTemplate(order, ex.Message, ErrorType.SavedCard);
            }
            finally
            {
                CheckoutDone(order);
            }
        }

        public bool SavedCardSupported(Order order)
        {
            return true;
        }

        private void UseSavedCardInternal(Order order, string savedCardToken = null)
        {
            PaymentCardToken savedCard = null;
            if (string.IsNullOrEmpty(savedCardToken))
            {
                savedCard = Services.PaymentCard.GetById(order.SavedCardId);
                if (savedCard == null || order.CustomerAccessUserId != savedCard.UserID)
                {
                    throw new PaymentCardTokenException("Token is incorrect.");
                }
                else
                {
                    LogEvent(order, "Using saved card({0}) with id: {1}", savedCard.Identifier, savedCard.ID);
                    savedCardToken = savedCard.Token;
                }

                if (order.IsRecurringOrderTemplate)
                {
                    order.TransactionCardType = savedCard.CardType;
                    order.TransactionCardNumber = savedCard.Identifier;
                    SetOrderComplete(order);
                    return;
                }
            }

            var rqstCharge = new Dictionary<string, string>()
                {
                    {"merchant", Merchant},
                    {"amount", order.Price.PricePIP.ToString()},
                    {"currency", order.CurrencyCode},
                    {"ticket", savedCardToken},
                    {"orderid", order.Id},
                    {"textreply", "true"},
                    {"uniqueoid", "true"},
                    {"test", TestMode ? "yes" : ""},
                };

            if (UseAddFee)
            {
                rqstCharge.Add("calcfee", "yes");
            }
            if (UseMD5)
            {
                rqstCharge.Add("md5key", StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(Key1 + $"merchant={Merchant}&orderid={order.Id}&ticket={savedCardToken}currency={order.Price.Currency.PayGatewayCode}&amount={order.Price.PricePIP}")));
            }

            LogEvent(order, "Starting payment process");
            var response = HttpUtility.ParseQueryString(ExecutePostRequest(order, "cgi-ssl/ticket_auth.cgi", rqstCharge));

            if (response != null && string.Equals(response["status"], "ACCEPTED", StringComparison.InvariantCultureIgnoreCase))
            {
                if (savedCard != null)
                {
                    order.TransactionCardType = savedCard.CardType;
                    order.TransactionCardNumber = savedCard.Identifier;
                }
                ProcessOrder(order, response);
            }
            else
            {
                order.TransactionStatus = "Failed";

                // Print error message
                var errMessage = "Authorize ticket failed";
                if (response != null)
                {
                    errMessage += string.Format(". Rejection reason: {0}", response["reason"]);
                }
                throw new Exception(errMessage);
            }
        }

        #endregion

        #region IRecurring interface

        public void Recurring(Order order, Order originalOrder)
        {
            if (order != null)
            {
                try
                {
                    UseSavedCardInternal(order);
                    LogEvent(order, "Recurring succeeded");
                }
                catch (Exception ex)
                {
                    LogEvent(order, "Recurring order failed for {0} (based on {1}). The payment failed with the message: {2}",
                        DebuggingInfoType.RecurringError, order.Id, originalOrder.Id, ex.Message);
                }
                finally
                {
                    CheckoutDone(order);
                }
            }
        }

        public bool RecurringSupported(Order order)
        {
            return true;
        }

        #endregion

        #region DIBS API

        private string ExecutePostRequest(Order order, string function, Dictionary<string, string> parameters)
        {
            var content = string.Join("&", parameters.Select(pair => $"{pair.Key}={HttpUtility.UrlEncode(pair.Value)}"));

            //Set correct POST URL corresponding to the payment function requested
            string postUrl = $"https://payment.architrade.com/{function}";

            var values = Encoding.UTF8.GetBytes(content);

            //Using HttpWebRequest for posting and receiving response
            var request = (HttpWebRequest)WebRequest.Create(postUrl);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.ContentLength = values.Length;
            request.Timeout = 15000; //15 seconds timeout

            //Send the POST request
            using (Stream stream = request.GetRequestStream())
            {
                stream.Write(values, 0, values.Length);
            }

            //Receive response
            var response = (HttpWebResponse)request.GetResponse();
            string responseString = new StreamReader(response.GetResponseStream()).ReadToEnd();

            LogEvent(order, "Response HttpStatusCode: {0}", response.StatusCode);

            return responseString;
        }

        #endregion
    }
}
