﻿using Dynamicweb.Configuration;
using Dynamicweb.Core;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Environment;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Frontend;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using FrontendRenderer = Dynamicweb.Ecommerce.Frontend.Renderer;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.DibsMobile
{
    /// <summary>
    /// DIBS Mobile Checkout Handler (Mopay)
    /// </summary>
    [
        AddInName("DIBS Mobile"),
        AddInDescription("DIBS Mobile Checkout Handler (Mopay)")
    ]
    public class DibsMobile : CheckoutHandler, IDropDownOptions, IRemotePartialCapture
    {
        private enum PostModes { Auto, Template }
        private PostModes postMode = PostModes.Auto;

        #region Editors

        [AddInParameter("Merchant"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Merchant { get; set; }

        [AddInParameter("HMAC key"), AddInParameterEditor(typeof(TextParameterEditor), "size=50")]
        public string HmacKey { get; set; }

        [AddInParameter("Key1"), AddInParameterEditor(typeof(TextParameterEditor), "size=50")]
        public string Key1 { get; set; }

        [AddInParameter("Key2"), AddInParameterEditor(typeof(TextParameterEditor), "size=50")]
        public string Key2 { get; set; }

        [AddInParameter("Card type"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PayType { get; set; }

        [AddInParameter("Post mode"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PostModeSelection
        {
            get
            {
                return postMode.ToString();
            }
            set
            {
                switch (value)
                {
                    case "Auto":
                        postMode = PostModes.Auto;
                        break;

                    case "Template":
                        postMode = PostModes.Template;
                        break;
                }
            }
        }

        [AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Post")]
        public string PostTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/DIBS/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Use HMAC checksum"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UseHMAC { get; set; }

        [AddInParameter("Use MD5 checksum for capture"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UseMD5 { get; set; }

        [AddInParameter("Test Mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        #endregion

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");
                var formValues = GetFormValues(order);

                switch (postMode)
                {
                    case PostModes.Auto:
                        LogEvent(order, "Auto post to DIBS Mobile");
                        SubmitForm("https://mopay.dibspayment.com", formValues, false);
                        return string.Empty;

                    case PostModes.Template:
                        LogEvent(order, "Render template");

                        var formTemplate = new Template(PostTemplate);

                        foreach (var formValue in formValues)
                        {
                            formTemplate.SetTag(string.Format("Dibs.{0}", formValue.Key), formValue.Value);
                        }

                        var renderer = new FrontendRenderer(PageView.Current());
                        renderer.RenderOrder(order, formTemplate);

                        return formTemplate.Output();

                    default:
                        LogError(order, "Unhandled post mode: '{0}'", postMode);
                        throw new ArgumentException(string.Format("Unhandled post mode: '{0}'", postMode));
                }
            }
            catch (ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        /// <summary>
        /// Redirects user to payment step or canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(null, "Redirected to DIBS Mobile CheckoutHandler");

                if (Context.Current.Request.QueryString["DibsState"] == null)
                {
                    LogError(null, "The request variable 'DibsState' is not present");
                    return string.Empty;
                }

                switch (Context.Current.Request.QueryString["DibsState"])
                {
                    case "Accept":
                        return StateOk(order);

                    case "Cancel":
                        return StateCancel(order);

                    case "Callback":
                        Callback(order);
                        Context.Current.Response.End();
                        return null;

                    default:
                        LogError(order, "Unknown DibsState: '{0}'", Converter.ToString(Context.Current.Request["DibsState"]));
                        return string.Empty;
                }
            }
            catch (ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        #region IDropDownOptions

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="behaviorMode"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string behaviorMode)
        {
            try
            {
                switch (behaviorMode)
                {
                    case "Post mode":
                        return new Hashtable
                        {
                            { "Auto", Translator.Translate("Auto post (does not use the template)") },
                            { "Template", Translator.Translate("Render template") }
                        };

                    case "Card type":
                        var cardTypes = new Dictionary<string, string>
                        {
                            { "AMEX", "American Express" },
                            { "AMEX(DK)", "American Express (Danish card)" },
                            { "DIN", "Diners Club" },
                            { "DIN(DK)", "Diners Club (Danish card)" },
                            { "DK", "Dankort" },
                            { "MC", "Mastercard" },
                            { "MC(DK)", "Mastercard (Danish card)" },
                            { "MC(SE)", "Mastercard (Swedish card)" },
                            { "MTRO", "Maestro" },
                            { "MTRO(DK)", "Maestro (Danish card)" },
                            { "MTRO(UK)", "Maestro (UK)" },
                            { "V-DK", "Visa/Dankort" },
                            { "VISA", "Visa" },
                            { "VISA(DK)", "Visa (Danish card)" },
                            { "VISA(SE)", "Visa (Swedish card)" },
                            { "ELEC", "Visa Electron (Danish card)" }
                        }.ToDictionary(
                            x => x.Key,
                            x => Translator.Translate(x.Value)
                        );

                        var danishCardsIncludeList = new List<string> { "AMEX(DK)", "DIN(DK)", "DK", "MC(DK)", "MTRO(DK)", "V-DK", "VISA(DK)", "ELEC" };
                        var allCards = new StringBuilder();
                        var danishCards = new StringBuilder();
                        foreach (var key in cardTypes.Keys)
                        {
                            if (danishCardsIncludeList.Contains(key))
                            {
                                if (danishCards.Length > 0)
                                {
                                    danishCards.Append(",");
                                }

                                danishCards.Append(key);
                            }

                            if (allCards.Length > 0)
                            {
                                allCards.Append(",");
                            }

                            allCards.Append(key);
                        }

                        var returnHashTable = new Hashtable
                        {
                            { allCards.ToString(), Translator.Translate("All cards") },
                            { danishCards.ToString(), Translator.Translate("Danish cards") }
                        };

                        foreach (var key in cardTypes.Keys)
                        {
                            returnHashTable.Add(key, cardTypes[key]);
                        }

                        return returnHashTable;

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", behaviorMode));
                }
            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }
        }

        #endregion

        #region IRemoteCapture

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        /// <remarks>http://tech.dibspayment.com/capturecgi</remarks>
        public OrderCaptureInfo Capture(Order order) => Capture(order, order.Price.PricePIP, false);

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <param name="amount">Amount to be transferred. The amount must be given in smallest possible unit, e.g. cents 100 = USD 1.</param>
        /// <param name="final">Shows that this step is final</param>
        /// <returns>Response from transaction service</returns>
        /// <remarks>
        ///		Create byte array with POST data
        ///		merchant       The merchant parameter is the shops DIBS identification number. The merchant number is available in the order confirmation email and is a 7-digit integer.
        ///		amount         Amount to be transferred. The amount must be given in smallest possible unit, e.g. cents 100 = USD 1.
        ///		transact       The unique DIBS identification number for the transaction you want to capture. The transact is returned from a successful authorisation, and is a as minimum 6-digit integer , e.g. transact=123456.
        ///		orderid        The shops (unique) order identification number as given in the authorisation process.
        ///		[textreply]    If this variable is set (e.g. textreply=true) the DIBS server returns its answer  in simple text format. If you are not using the standard DIBS payment window, e.g. using server-to-server requests, this significantly simplifies parsing the answer and is recommended.
        ///		[force]        If force=true is posted, the DIBS server will attempt to capture otherwise expired authorisations (more than 7 days old). This is achieved through an initial re-authorisation of the original transaction, followed by a capture process.
        ///		[account]      If several departments in a company use the same acquirer contract it may be beneficial to keep track of the transactions by department. Including the account parameter enables separation of the transaction by department.
        ///		[md5key]       This variable enables a MD5 key control of the values received by DIBS. This control  confirms that the values sent to DIBS has not been tampered with during the transfer. The MD5 key is calculated as: MD5(Key2 + MD5(Key1 + "merchant=<merchant>&orderid=<orderid> &transact=<transact>&amount=<amount>")) Where key1 and key2 are shop specific keys available through the DIBS administration interface, and + is the concatenation operator. NB! MD5 key check must also be enabled through the DIBS administration interface in order to work. Further details on MD5-key control . 
        /// </remarks>
        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            try
            {
                if (order == null || string.IsNullOrEmpty(order.Id))
                {
                    const string errorText = "No valid Order object set";
                    LogError(order, errorText);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
                }
                if (string.IsNullOrEmpty(order.TransactionNumber))
                {
                    const string errorText = "No transaction number set on the order";
                    LogEvent(order, errorText);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
                }

                var postRequestData = new StringBuilder();
                postRequestData.AppendFormat("merchant={0}", Merchant);
                postRequestData.AppendFormat("&amount={0}", amount);
                postRequestData.AppendFormat("&transact={0}", order.TransactionNumber);
                postRequestData.AppendFormat("&orderid={0}", order.Id);

                if (order.Price.PricePIP != amount)
                {
                    if (final)
                    {
                        postRequestData.Append("&close=true");
                    }
                    postRequestData.Append("&splitpay=true");
                }
                else
                {
                    postRequestData.Append("&force=true");
                }

                if (UseHMAC)
                {
                    postRequestData.AppendFormat(
                        "&md5key={0}",
                        StringHelper.Md5HashToString(Key2 + StringHelper.Md5HashToString(Key1 +
                            string.Format(
                                "merchant={0}&orderid={1}&transact={2}&amount={3}",
                                Merchant, order.Id, order.TransactionNumber, order.Price.PricePIP)
                            )
                        )
                    );
                }

                var values = Encoding.UTF8.GetBytes(postRequestData.ToString());

                var request = WebRequest.Create("https://payment.architrade.com/cgi-bin/capture.cgi");
                request.Method = "POST";
                request.ContentType = "application/x-www-form-urlencoded";
                request.ContentLength = values.Length;

                using (var stream = request.GetRequestStream())
                {
                    stream.Write(values, 0, values.Length);
                }

                var response = (HttpWebResponse)request.GetResponse();
                LogEvent(
                    order,
                    "Remote capture response from DIBS: HttpStatusCode = {0}, HttpStatusDescription = {1}",
                    response.StatusCode,
                    response.StatusDescription
                );

                var reader = new StreamReader(response.GetResponseStream());
                var responseText = reader.ReadToEnd();

                var statusCodes = new Dictionary<string, string>
                {
                    { "0", "Accepted" },
                    { "1", "No response from acquirer." },
                    { "2", "Error in the parameters sent to the DIBS server. An additional parameter called 'message' is returned, with a value that may help identifying the error." },
                    { "3", "Credit card expired." },
                    { "4", "Rejected by acquirer." },
                    { "5", "Authorisation older than7 days." },
                    { "6", "Transaction status on the DIBS server does not allow capture." },
                    { "7", "Amount too high." },
                    { "8", "Amount is zero." },
                    { "9", "Order number (orderid) does not correspond to the authorisation order number." },
                    { "10", "Re-authorisation of the transaction was rejected." },
                    { "11", "Not able to communicate with the acquier." },
                    { "15", "Capture was blocked by DIBS." }
                };

                var statusMatch = Regex.Match(responseText, "status=(?<status>[^&]+)");
                if (statusMatch.Success)
                {
                    var status = statusMatch.Groups["status"].Value;

                    if (status == "ACCEPTED")
                    {
                        if (order.Price.PricePIP == amount)
                        {
                            LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
                        }
                        else if (final)
                        {
                            LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture(final)", amount / 100f), DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Split capture successful");
                        }
                        else
                        {
                            LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture", amount / 100f), DebuggingInfoType.CaptureResult);
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Split, "Split capture successful");
                        }
                    }
                    else
                    {
                        string infoText;
                        var resultMatch = Regex.Match(responseText, "result=(?<result>[^&]+)");
                        string result = resultMatch.Success ? resultMatch.Groups["result"].Value : "-1";

                        if (statusCodes.ContainsKey(result))
                        {
                            infoText = string.Format(
                                "Remote Capture failed. Statuscode: '{0}', description: '{1}', responseText: '{2}'",
                                result, statusCodes[result], responseText
                            );
                        }
                        else
                        {
                            infoText = string.Format(
                                "Remote Capture failed with unknown statuscode '{0}', responseText: '{1}'",
                                result, responseText
                            );
                        }

                        LogEvent(order, infoText, DebuggingInfoType.CaptureResult);
                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoText);
                    }
                }

                string eventTxt = string.Format("The text returned from DIBS did not contain the result state. ResponseText: {0}", responseText);
                LogEvent(order, eventTxt);

                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, eventTxt);
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        /// <summary>
        /// Shows if capture supported
        /// </summary>
        /// <param name="order">This object is not used in current implementation</param>
        /// <returns>This method always return 'true' value</returns>
        public bool CaptureSupported(Order order) => true;

        /// <summary>
        /// Shows if partial capture of the order supported
        /// </summary>
        /// <param name="order">Instance of order</param>
        /// <returns>True, if partial capture of the order is supported</returns>
        public bool SplitCaptureSupported(Order order)
        {
            var supportedCards = new ArrayList { "Dankort", "MasterCard", "VISA", "VISA-Dankort" };
            return supportedCards.Contains(order.TransactionCardType);
        }

        #endregion

        #region Form properties

        private static string BaseUrl(Order order)
        {
            var disablePortNumber = SystemConfiguration.Instance.GetBoolean("/Globalsettings/System/http/DisableBaseHrefPort");
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? "" : string.Format(":{0}", Context.Current.Request.Url.Port);
            var pageId = Context.Current.Request["ID"] == null ? string.Empty : string.Format("ID={0}", Context.Current.Request["ID"]);

            return string.Format("{2}://{0}{3}/Default.aspx?{1}{5}={4}", Context.Current.Request.Url.Host, pageId, Context.Current.Request.Url.Scheme, portString, order.Id, OrderIdRequestName);
        }

        private static string AcceptUrl(Order order)
        {
            return string.Format("{0}&DibsState=Accept", BaseUrl(order));
        }

        private static string CancelUrl(Order order)
        {
            return string.Format("{0}&DibsState=Cancel", BaseUrl(order));
        }

        private static string CallbackUrl(Order order)
        {
            return string.Format("{0}&DibsState=Callback", BaseUrl(order));
        }

        private static string LanguageCode
        {
            get
            {
                var supportedLanguageCodes = new List<string>(new[] { "da_DK", "sv_SE", "nb_NO", "en_US", "en_GB", "fi_FI" });

                var cultureInfo = ExecutingContext.GetCulture(true);
                var currentRegionCode = new RegionInfo(cultureInfo.LCID);
                var currentLanguageCode = cultureInfo.TwoLetterISOLanguageName;
                currentLanguageCode = currentLanguageCode == "nn" ? "nb" : currentLanguageCode;
                var currentIetfLanguageTag = currentLanguageCode + "_" + currentRegionCode.TwoLetterISORegionName;
                return supportedLanguageCodes.Contains(currentIetfLanguageTag) ? currentIetfLanguageTag : "en_GB";
            }
        }

        #endregion

        private Dictionary<string, string> GetFormValues(Order order)
        {
            var formValues = new Dictionary<string, string>
            {
                { "acceptreturnurl", AcceptUrl(order) },
                { "amount", order.Price.PricePIP.ToString() },
                { "callbackurl", CallbackUrl(order) },
                { "cancelreturnurl", CancelUrl(order) },
                { "currency", order.Price.Currency.PayGatewayCode.ToString() },
                { "language", LanguageCode },
                { "merchant", Merchant },
                { "orderid", order.Id },
                { "paytype", PayType },
                { "s_" + OrderIdRequestName, order.Id },
                { "test", TestMode ? "1" : string.Empty },
                { "uniqueid", order.Id }
            }.Where(x => x.Value.Length > 0).ToDictionary(x => x.Key, x => x.Value);

            if (UseHMAC)
            {
                formValues.Add("MAC", ComputeHash(HmacKey, GetMacString(formValues, false)));
            }

            return formValues;
        }

        private string GetMacString(IDictionary<string, string> formValues, bool urlEncode)
        {
            var excludeList = new List<string> { "MAC" };

            var keysSorted = formValues.Keys.ToArray();
            Array.Sort(keysSorted, StringComparer.Ordinal);

            var message = new StringBuilder();
            foreach (string key in keysSorted)
            {
                if (excludeList.Contains(key))
                    continue;

                if (message.Length > 0)
                    message.Append("&");

                var value = formValues[key];
                message.AppendFormat("{0}={1}", key, (urlEncode ? Context.Current.Server.UrlEncode(value) : value));
            }

            return message.ToString();
        }

        private byte[] HexToByteArray(string hexString)
        {
            if (hexString.Length % 2 != 0)
            {
                throw new ArgumentException("HexToString argument 'hexString'. Length must be divisible by 2. Length: '" + hexString.Length + "'");
            }

            var bytes = new List<byte>();
            for (var i = 0; i < hexString.Length; i += 2)
            {
                bytes.Add(Convert.ToByte(hexString.Substring(i, 2), 16));
            }

            return bytes.ToArray();
        }

        private string ByteArrayToHexString(byte[] bytes)
        {
            var result = new StringBuilder();
            foreach (byte b in bytes)
            {
                result.Append(b.ToString("x2"));
            }

            return result.ToString();
        }

        private string ComputeHash(string key, string message)
        {
            var byteKey = HexToByteArray(key);
            using (HMACSHA256 hmac = new HMACSHA256(byteKey))
            {
                var encoding = new UTF8Encoding();
                var messageBytes = encoding.GetBytes(message);
                var hashedBytes = hmac.ComputeHash(messageBytes);

                return ByteArrayToHexString(hashedBytes);
            }
        }

        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");
            if (order.Complete)
            {
                RedirectToCart(order);
                return null;
            }

            if (Common.Context.Cart == null)
            {
                CheckoutDone(order);
            }

            LogError(order, "Called State ok, but order is not set complete - this should have happened in the callback.");

            return PrintErrorTemplate(order);
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            var orderRenderer = new FrontendRenderer();
            orderRenderer.RenderOrderDetails(cancelTemplate, order, true);
            return cancelTemplate.Output();
        }

        private static void SetOrderSucceeded(Order order, bool success)
        {
            if (success)
            {
                order.TransactionAmount = order.Price.PricePIP / 100d;
                order.TransactionStatus = "Succeeded";
            }
            else
            {
                order.TransactionStatus = "Failed";
            }

            Services.Orders.Save(order);
        }

        private void Callback(Order order)
        {
            try
            {
                LogEvent(order, "Callback started");

                var errorOccured = false;
                var transactionID = Converter.ToString(Context.Current.Request["transaction"]);

                if (string.IsNullOrEmpty(transactionID))
                {
                    errorOccured = true;
                    LogEvent(order, "No transaction number sent to callback");
                }

                if (UseHMAC)
                {
                    var post = new Dictionary<string, string>();
                    foreach (string key in Context.Current.Request.Form.AllKeys)
                    {
                        post.Add(key, Context.Current.Request.Form[key]);
                    }

                    var calculatedHMAC = ComputeHash(HmacKey, GetMacString(post, false));

                    var authKey = Converter.ToString(Context.Current.Request["MAC"]);
                    if (!calculatedHMAC.Equals(authKey, StringComparison.InvariantCultureIgnoreCase))
                    {
                        errorOccured = true;
                        LogEvent(order, "HMAC checksums do not match");
                    }
                }

                if (errorOccured)
                {
                    LogError(order, "At least one callback validation error exists - exiting callback routine.");

                    SetOrderSucceeded(order, false);
                }
                else
                {
                    LogEvent(order, "Payment succeeded with transaction number {0}", transactionID);

                    order.TransactionCardType = Converter.ToString(Context.Current.Request["cardTypeName"]);
                    order.TransactionCardNumber = HideCardNumber(Converter.ToString(Context.Current.Request["cardNumberMasked"]));
                    SetOrderComplete(order, transactionID);

                    SetOrderSucceeded(order, true);
                }
            }
            finally
            {
                CheckoutDone(order);
            }
        }

        private string PrintErrorTemplate(Order order)
        {
            LogEvent(order, "Printing error template");

            var errorTemplate = new Template(ErrorTemplate);
            var orderRenderer = new FrontendRenderer();
            orderRenderer.RenderOrderDetails(errorTemplate, order, true);

            return errorTemplate.Output();
        }
    }
}
