﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.CheckoutHandlers.EBizCharge.EBizApi;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.Security.UserManagement;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Web;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.EBizCharge
{
    /// <summary>
    /// EBizCharge Payment Window Checkout Handler
    /// </summary>
    [AddInName("EBizCharge"), AddInDescription("EBizCharge Checkout handler")]
    public class EBizCharge : CheckoutHandler, IDropDownOptions, ISavedCard, IRemotePartialCapture
    {
        private const string ClientSoftwareName = "Dynamicweb CMS";

        private enum TransactionType   //Processing command. Possible values: Sale, Credit, Void, CreditVoid, AuthOnly, Capture, PostAuth, Check, and CheckCredit. Default is Sale.(optional)
        {
            Sale,
            Credit,
            Void,
            CreditVoid,
            AuthOnly,
            Capture,
            PostAuth,
            Check,
            CheckCredit
        }

        private enum WindowModes { Redirect, Embedded }

        private WindowModes windowMode = WindowModes.Redirect;
        private enum PaymentMethodTypes { CreditCard, ACH, Both }
        private PaymentMethodTypes PaymentMethodType;

        #region Addin parameters

        /// <summary>
        /// Gets or sets the EBiz security id
        /// </summary>
        [AddInLabel("API SecurityId"), AddInParameter("SecurityId"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string SecurityId { get; set; }

        /// <summary>
        /// Gets or sets the EBiz username
        /// </summary>
        [AddInLabel("API user name"), AddInParameter("UserName"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string UserName { get; set; }

        /// <summary>
        /// Gets or sets the EBiz password
        /// </summary>
        [AddInLabel("API password"), AddInParameter("Password"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Password { get; set; }

        /// <summary>
        /// Gets or sets value whether to use customer number instead userid when creating/retrieving EBiz customer..
        /// </summary>
        [AddInLabel("Use user customer number to create/retrieve EBiz customer instead of user id"), AddInParameter("UseCustomerNumber"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UseCustomerNumber { get; set; }

        /// <summary>
        /// Gets or sets value whether to update user saved cards up-to-date on every order.
        /// </summary>
        [AddInLabel("Update user saved cards up-to-date on every order."), AddInParameter("UpdateSavedCards"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UpdateSavedCards { get; set; }

        /// <summary>
        /// Gets or sets payment method type: Credit Card, ACH, or both
        /// </summary>
        [AddInLabel("Payment Method Type"), AddInParameter("PaymentMethodType"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PaymentMethodTypeSelection
        {
            get { return PaymentMethodType.ToString(); }
            set { Enum.TryParse<PaymentMethodTypes>(value, out PaymentMethodType); }
        }

        /// <summary>
        /// Gets or sets payment transaction type - sale or auth only
        /// </summary>
        [AddInLabel("Payment transaction type"), AddInParameter("PaymentTransactionType"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PaymentTransactionType { get; set; } = "Sale";

        [AddInLabel("Window Mode"), AddInParameter("WindowMode"), AddInParameterEditor(typeof(RadioParameterEditor), "")]
        public string WindowMode
        {
            get { return windowMode.ToString(); }
            set { Enum.TryParse<WindowModes>(value, out windowMode); }
        }

        /// <summary>
        /// Gets or sets template for output when any arror during checkout flow happened
        /// </summary>
        [AddInLabel("Embedded template"), AddInParameter("EmbeddedTemplate"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/EbizCharge/Payment")]
        public string EmbeddedPaymentTemplate { get; set; } = "eCom7/CheckoutHandler/EbizCharge/Payment/Payment.cshtml";

        /// <summary>
        /// Gets or sets template for output when any arror during checkout flow happened
        /// </summary>
        [AddInLabel("Error template"), AddInParameter("ErrorTemplate"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/ErrorTemplate")]
        public string FailedOrderTemplate { get; set; }

        #endregion

        /// <summary>
        /// Starts the checkout for order via requesting payment url from EBiz and redirecting the customer to it
        /// </summary>
        /// <param name="order">The order</param>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");
                using (var client = new IeBizServiceClient())
                {
                    SecurityToken token = GetToken();
                    EbizWebForm ePaymentForm = GetForm(order);
                    var response = client.GetEbizWebFormURL(token, ePaymentForm);
                    if (windowMode == WindowModes.Embedded)
                    {
                        return RenderPaymentFrame(order, response);
                    }
                    else
                    {
                        Context.Current.Response.Redirect(response);
                    }
                }
                return string.Empty;
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintFailedOrderTemplate(order, ex.Message);
            }
        }

        private string RenderPaymentFrame(Orders.Order order, string gatewayUrl)
        {
            if (string.IsNullOrWhiteSpace(EmbeddedPaymentTemplate))
            {
                LogError(order, "Embedded payment template not set");
                return PrintFailedOrderTemplate(order, "Embedded payment template not set");
            }

            // Get template
            var formTemplate = new Template(EmbeddedPaymentTemplate);

            // Render tags
            formTemplate.SetTag("EBizCharge.EbizWebFormURL", gatewayUrl);

            // Render and return
            return this.Render(order, formTemplate);
        }

        private string PrintFailedOrderTemplate(Order order, string msg)
        {
            order.TransactionAmount = 0;
            order.TransactionStatus = "Failed";
            order.Errors.Add(msg);
            Services.Orders.Save(order);
            Services.Orders.DowngradeToCart(order);
            order.CartV2StepIndex = 0;
            order.TransactionStatus = "";
            Common.Context.SetCart(order);

            if (string.IsNullOrWhiteSpace(FailedOrderTemplate))
            {
                RedirectToCart(order);
            }

            var template = new Template(FailedOrderTemplate);
            template.SetTag("CheckoutHandler:ErrorMessage", msg);

            return Render(order, template);
        }

        #region EBiz api

        private EbizWebForm GetForm(Order order)
        {
            var savedCardName = GetSavedCardName(order);
            var baseUrl = GetBaseUrl(order);
            if (windowMode == WindowModes.Embedded)
            {
                baseUrl = $"{baseUrl}&Embedded=True";
            }
            var customerName = order.CustomerName;
            if (string.IsNullOrWhiteSpace(customerName))
            {
                customerName = $"{order.CustomerFirstName} {order.CustomerMiddleName} {order.CustomerSurname}";
            }
            if (string.IsNullOrWhiteSpace(customerName))
            {
                customerName = savedCardName;
            }
            var paymentForm = new EbizWebForm
            {
                FormType = "WebForm",
                FromEmail = order.CustomerEmail,
                FromName = order.CustomerName,
                EmailAddress = order.CustomerEmail,
                CustomerId = GetCustomer(order.CustomerAccessUserId, order)?.CustomerId,
                CustFullName = customerName,
                TransId = order.Id,
                TransDetail = order.Comment,
                InvoiceNumber = order.Id,
                PoNum = order.Id,
                SoNum = order.Id,
                OrderId = order.Id,
                Date = order.Date,
                DueDate = order.Date,
                TotalAmount = (decimal)order.Price.Price,
                AmountDue = (decimal)order.Price.Price,
                TaxAmount = (decimal)order.Price.VAT,
                BillingAddress = GetAddress(order.CustomerAddress, order.CustomerAddress2, order.CustomerCity, order.CustomerZip, order.CustomerRegion),
                ShippingAddress = GetAddress(order.DeliveryAddress, order.DeliveryAddress2, order.DeliveryCity, order.DeliveryZip, order.DeliveryRegion),
                ApprovedURL = $"{baseUrl}&Action=Approved&SavedCardName={savedCardName}",
                DeclinedURL = $"{baseUrl}&Action=Declined",
                ErrorURL = $"{baseUrl}&Action=Error",
                DisplayDefaultResultPage = 0,
                PayByType = getPayByType(PaymentMethodType),
                SavePaymentMethod = order.CustomerAccessUserId > 0 && !string.IsNullOrWhiteSpace(savedCardName),
                ShowSavedPaymentMethods = order.CustomerAccessUserId > 0,
                CountryCode = order.CustomerCountryCode,
                CurrencyCode = order.CurrencyCode,
                SoftwareId = ClientSoftwareName,
                TransactionLookupKey = order.Id,
                ProcessingCommand = PaymentTransactionType,
                LineItems = GetTransactionLineItems(order)
            };

            return paymentForm;
        }

        private string getPayByType(PaymentMethodTypes paymentType)
        {
            switch (paymentType)
            {
                case PaymentMethodTypes.CreditCard:
                    return "";
                case PaymentMethodTypes.ACH:
                    return "ACH";
                case PaymentMethodTypes.Both:
                    return "CC,ACH";
                default:
                    return "";
            }
        }

        private string GetSavedCardName(Order order)
        {
            return !string.IsNullOrWhiteSpace(order.SavedCardDraftName) ? order.SavedCardDraftName : (order.DoSaveCardToken || order.IsRecurringOrderTemplate ? order.Id : "");
        }

        private TransactionLineItem[] GetTransactionLineItems(Order order)
        {
            var lineItems = new TransactionLineItem[order.OrderLines.Count];
            for (var i = 0; i < order.OrderLines.Count; i++)
            {
                lineItems[i] = CreateTransactionLineItem(order.OrderLines[i]);
            }
            return lineItems;
        }

        private static TransactionLineItem CreateTransactionLineItem(OrderLine orderLine)
        {
            return new TransactionLineItem
            {
                ProductRefNum = orderLine.ProductId,
                SKU = orderLine.ProductId,
                ProductName = orderLine.ProductName,
                UnitPrice = (decimal)orderLine.UnitPrice.Price,
                Qty = (decimal)orderLine.Quantity,
                TaxAmount = (decimal)orderLine.Price.VAT,
                Taxable = orderLine.HasType(OrderLineType.Product)
            };
        }

        private LineItem[] GetLineItems(Order order)
        {
            var lineItems = new LineItem[order.OrderLines.Count];
            for (var i = 0; i < order.OrderLines.Count; i++)
            {
                lineItems[i] = CreateLineItem(order.OrderLines[i]);
            }
            return lineItems;
        }

        private static LineItem CreateLineItem(OrderLine orderLine)
        {
            return new LineItem
            {
                ProductRefNum = orderLine.ProductId,
                SKU = orderLine.ProductId,
                ProductName = orderLine.ProductName,
                UnitPrice = orderLine.UnitPrice.Price.ToString(),
                Qty = orderLine.Quantity.ToString(),
                TaxAmount = orderLine.Price.VAT.ToString(),
                Taxable = orderLine.HasType(OrderLineType.Product)
            };
        }

        private static Address GetAddress(string street, string street2, string city, string zip, string state)
        {
            return new Address
            {
                Address1 = street,
                Address2 = street2,
                City = city,
                ZipCode = zip,
                State = state
            };
        }

        private Customer GetCustomer(int userId, Order order)
        {
            var user = User.GetUserByID(userId);
            var ebizCustomerId = UseCustomerNumber ? user?.CustomerNumber : userId.ToString();
            Customer customer = null;
            using (var client = new IeBizServiceClient())
            {
                try
                {
                    customer = client.GetCustomer(GetToken(), ebizCustomerId, "");
                }
                catch (Exception e)
                {
                    LogError(order, $"Error on getting customer, {e.Message}. Will create new EBizCustomer for user {userId}");
                }
                if (string.IsNullOrWhiteSpace(customer?.CustomerToken) && order != null)
                {
                    if (userId == 0 || user == null)
                    {
                        customer = new Customer
                        {
                            CustomerId = "0",
                            FirstName = "Anonymous"
                        };
                    }
                    else
                    {
                        customer = new Customer
                        {
                            CustomerId = ebizCustomerId,
                            FirstName = order.CustomerFirstName,
                            LastName = order.CustomerSurname,
                            CompanyName = order.CustomerCompany,
                            Phone = order.CustomerPhone,
                            Email = order.CustomerEmail,
                            CellPhone = order.CustomerCell,
                            ShippingAddress = GetAddress(order.DeliveryAddress, order.DeliveryAddress2, order.DeliveryCity, order.DeliveryZip, order.DeliveryRegion),
                            BillingAddress = GetAddress(order.CustomerAddress, order.CustomerAddress2, order.CustomerCity, order.CustomerZip, order.CustomerRegion)
                        };
                    }
                    var response = client.AddCustomer(GetToken(), customer);
                    if (!"Success".Equals(response.Status, StringComparison.OrdinalIgnoreCase))
                    {
                        LogError(order, $"Error on creating customer, {response.ErrorCode}:{response.Error}.");
                    }
                }
            }
            return customer;
        }

        private SecurityToken GetToken()
        {
            return new SecurityToken { UserId = UserName, SecurityId = SecurityId, Password = Password };
        }

        #endregion

        #region Redirect from payment gateway

        /// <summary>
        /// Handles redirects from gateway back to dynamicweb
        /// </summary>
        /// <param name="order">The order</param>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(order, "Redirected to EbizCharge CheckoutHandler");
                if (Converter.ToBoolean(Context.Current.Request["Embedded"]))
                {
                    var newQueryString = HttpUtility.ParseQueryString(Context.Current.Request.Url.Query);
                    newQueryString.Remove("Embedded");
                    Context.Current.Response.Write(string.Format("<script>parent.location.href = '{0}';</script>", $"{Context.Current.Request.Url.GetLeftPart(UriPartial.Path)}?{newQueryString}"));
                    Context.Current.Response.End();
                }
                switch (Context.Current.Request["Action"])
                {
                    case "Approved":
                        return StateOk(order);

                    case "Declined":
                        return StateCancel(order);

                    default:
                        return PrintFailedOrderTemplate(order, "Error happened during payment");
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintFailedOrderTemplate(order, ex.Message);
            }
        }
        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");
            try
            {
                var user = User.GetCurrentUser(PagePermissionLevels.Frontend);
                if ("Approved".Equals(Context.Current.Request["TranResult"], StringComparison.OrdinalIgnoreCase) && "A".Equals(Context.Current.Request["TranResultCode"]))
                {
                    order.GatewayPaymentStatus = GatewayPaymentStatus.Accepted;
                    order.TransactionCardNumber = HideCardNumber(Converter.ToString(Context.Current.Request["MaskedCC"]));
                    order.TransactionCardType = GetCardType(Converter.ToString(Context.Current.Request["CCType"]));
                    order.TransactionAmount = Converter.ToDouble(Context.Current.Request["AuthAmount"]);
                    order.TransactionStatus = "Succeeded";
                    var cardName = Converter.ToString(Context.Current.Request["SavedCardName"]);
                    if (!string.IsNullOrWhiteSpace(cardName))
                    {
                        if (user != null)
                        {
                            var savedCard = Services.PaymentCard.CreatePaymentCard(user.ID, order.PaymentMethodId, order.LanguageId, cardName, order.TransactionCardType, order.TransactionCardNumber, Context.Current.Request["PmToken"]);
                            order.SavedCardId = savedCard.ID;
                            Services.Orders.Save(order);
                        }
                        else
                        {
                            order.TransactionToken = Context.Current.Request["PmToken"];
                            Services.Orders.Save(order);
                        }
                    }

                    if (TransactionType.Sale.ToString().Equals(PaymentTransactionType, StringComparison.OrdinalIgnoreCase))
                    {
                        LogEvent(order, "Autocapturing order", DebuggingInfoType.CaptureResult);
                        order.CaptureInfo = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Autocapture successful");
                        order.CaptureAmount = order.TransactionAmount;
                    }

                    SetOrderComplete(order, Context.Current.Request["TranRefNum"]);
                }
                if (UpdateSavedCards)
                {
                    UpdateUserSavedCards(user, order);
                }
            }
            finally
            {
                CheckoutDone(order);
            }

            RedirectToCart(order);
            return null;
        }

        private static string GetCardType(string requestCCType)
        {
            switch (requestCCType)
            {
                case "V":
                    return "Visa";
                case "M":
                    return "MaterCard";
                case "A":
                    return "American Express";
                case "DS":
                    return "Discover";
                case "J":
                    return "JCB";
                case "DN":
                    return "Diners Club";

                default:
                    return requestCCType;
            }
        }

        private void UpdateUserSavedCards(User user, Order order)
        {
            if (user != null)
            {
                var savedPayments = new Dictionary<string, PaymentMethodProfile>();
                using (var client = new IeBizServiceClient())
                {
                    foreach (var savedPayment in client.GetCustomerPaymentMethodProfiles(GetToken(), GetCustomer(user.ID, null)?.CustomerToken))
                    {
                        savedPayments.Add(savedPayment.MethodID, savedPayment);
                    };
                }
                foreach (var savedCard in Services.PaymentCard.GetByUserId(user.ID))
                {
                    if (savedCard.PaymentID != order.PaymentMethodId)
                    {
                        continue;
                    }
                    if (!savedPayments.ContainsKey(savedCard.Token))
                    {
                        Services.PaymentCard.Delete(savedCard);
                    }
                    else
                    {
                        var savedPayment = savedPayments[savedCard.Token];
                        if (savedPayment.MethodID != savedCard.Token || HideCardNumber(savedPayment.CardNumber) != savedCard.Identifier || GetCardType(savedPayment.CardType) != savedCard.CardType)
                        {
                            Services.PaymentCard.Delete(savedCard);
                            Services.PaymentCard.CreatePaymentCard(user.ID, order.PaymentMethodId, order.LanguageId, savedPayment.MethodName, GetCardType(savedPayment.CardType), HideCardNumber(savedPayment.CardNumber), savedPayment.MethodID);
                        }
                        else if (!savedCard.Name.Equals(savedPayment.MethodName))
                        {
                            savedCard.Name = savedPayments[savedCard.Token].MethodName;
                            Services.PaymentCard.Save(savedCard);
                        }
                    }

                    savedPayments.Remove(savedCard.Token);
                };
                foreach (var newPaymentMethod in savedPayments.Keys)
                {
                    var savedPayment = savedPayments[newPaymentMethod];
                    Services.PaymentCard.CreatePaymentCard(user.ID, order.PaymentMethodId, order.LanguageId, savedPayment.MethodName, GetCardType(savedPayment.CardType), HideCardNumber(savedPayment.CardNumber), newPaymentMethod);
                }
            }
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");
            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);
            CheckoutDone(order);
            var cancelTemplate = new Template(FailedOrderTemplate);
            return Render(order, cancelTemplate);
        }

        #endregion

        #region IDropDownOptions

        /// <summary>
        /// Gets options for provided dropdown add-in editor by its name
        /// </summary>
        /// <param name="dropdownName"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string dropdownName)
        {
            try
            {
                switch (dropdownName)
                {
                    case "PaymentTransactionType":
                        return new Hashtable() { { TransactionType.AuthOnly.ToString(), TransactionType.AuthOnly.ToString() }, { TransactionType.Sale.ToString(), TransactionType.Sale.ToString() } };

                    case "WindowMode":
                        return new Hashtable
                                   {
                                       {WindowModes.Redirect.ToString(), Translate.Translate("Redirect")},
                                       {WindowModes.Embedded.ToString(), Translate.Translate("Embedded")}
                                   };
                    case "PaymentMethodType":
                        return new Hashtable
                        {
                            { PaymentMethodTypes.CreditCard.ToString(), Translate.Translate("Credit Card") },
                            { PaymentMethodTypes.ACH.ToString(), Translate.Translate("ACH") },
                            { PaymentMethodTypes.Both.ToString(), Translate.Translate("Both Credit Card and ACH") }
                        };

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", dropdownName));
                }

            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }
        }

        #endregion

        #region ISavedCard

        /// <summary>
        /// Gets value indicating whether saving card is supported
        /// </summary>
        /// <param name="order">The order</param>
        /// <remarks>Support any order</remarks>
        public bool SavedCardSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Use saved card on order to make a new payment request to EBiz gateway.
        /// </summary>
        /// <param name="order">The order.</param>
        public string UseSavedCard(Order order)
        {
            var token = Services.PaymentCard.GetById(order.SavedCardId);
            if (token == null || order.CustomerAccessUserId != token.UserID)
            {
                throw new Exception("Token is incorrect.");
            }

            var customer = GetCustomer(token.UserID, null);
            CustomerTransactionRequest tran = new CustomerTransactionRequest
            {
                Command = PaymentTransactionType,
                Details = new TransactionDetail
                {
                    Invoice = order.Id,
                    Description = "Payment from saved card.",
                    Amount = order.Price.Price,
                    Tax = order.Price.VAT
                },
                LineItems = GetLineItems(order)
            };
            try
            {
                using (var client = new IeBizServiceClient())
                {
                    var response = client.runCustomerTransaction(GetToken(), customer.CustomerToken, token.Token, tran);
                    LogEvent(order, response.Result);
                    if ("Approved".Equals(response.Result, StringComparison.OrdinalIgnoreCase) && "A".Equals(response.ResultCode, StringComparison.OrdinalIgnoreCase))
                    {
                        var transaction = client.GetTransactionDetails(GetToken(), response.RefNum);
                        order.GatewayPaymentStatus = GatewayPaymentStatus.Accepted;
                        order.TransactionCardNumber = HideCardNumber(transaction.CreditCardData.CardNumber);
                        order.TransactionCardType = GetCardType(transaction.CreditCardData.CardType);
                        order.TransactionAmount = transaction.Details.Amount;
                        order.TransactionStatus = "Succeeded";

                        if (TransactionType.Sale.ToString().Equals(PaymentTransactionType, StringComparison.OrdinalIgnoreCase))
                        {
                            LogEvent(order, "Autocapturing order", DebuggingInfoType.CaptureResult);
                            order.CaptureInfo = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Autocapture successful");
                            order.CaptureAmount = order.TransactionAmount;
                        }
                        SetOrderComplete(order, response.RefNum);
                    }
                    else
                    {
                        LogError(order, "Error during payment with saved card: {0}", response.Error);
                        return PrintFailedOrderTemplate(order, "Error during payment with saved card.");
                    }
                }

                var user = User.GetCurrentUser(PagePermissionLevels.Frontend);
                if (UpdateSavedCards)
                {
                    UpdateUserSavedCards(user, order);
                }
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintFailedOrderTemplate(order, ex.Message);
            }
            finally
            {
                CheckoutDone(order);
            }
            RedirectToCart(order);
            return string.Empty;
        }

        /// <summary>
        /// Deletes the saved card from gateway.
        /// </summary>
        /// <param name="savedCardId"></param>
        public void DeleteSavedCard(int savedCardId)
        {
            try
            {
                var token = Services.PaymentCard.GetById(savedCardId);
                if (token == null)
                {
                    throw new Exception("Token is incorrect.");
                }
                var customer = GetCustomer(token.UserID, null);
                using (var client = new IeBizServiceClient())
                {
                    client.DeleteCustomerPaymentMethodProfile(GetToken(), customer.CustomerToken, token.Token);
                }
            }
            catch (Exception e)
            {
                if ("Token is incorrect.".Equals(e.Message, StringComparison.OrdinalIgnoreCase))
                {
                    throw e;
                }
            }
        }

        #endregion

        #region IRemoteCapture

        /// <summary> 
        /// Gets value indicating whether capturing supported
        /// </summary>
        /// <param name="order">The order</param>
        /// <remarks>Support any order</remarks>
        public bool CaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Gets value indicating whether split capturing supported
        /// </summary>
        /// <param name="order">The order</param>
        /// <remarks>Support any order</remarks>
        public bool SplitCaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Captures the order for it's whole price.
        /// </summary>
        /// <param name="order">The order.</param>
        public OrderCaptureInfo Capture(Order order)
        {
            return Capture(order, order.Price.PricePIP, true);
        }

        /// <summary>
        /// Captures the order for specified amount.
        /// </summary>
        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            try
            {
                // Check order
                if (order == null)
                {
                    LogError(null, "Order not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Order not set");
                }
                else if (string.IsNullOrEmpty(order.Id))
                {
                    LogError(null, "Order id not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Order id not set");
                }
                else if (string.IsNullOrEmpty(order.TransactionNumber))
                {
                    LogError(null, "Transaction number not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Transaction number not set");
                }
                else if (order.Price.PricePIP < amount)
                {
                    LogError(null, "Amount to capture should be less then order total");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Amount to capture should be less then order total");
                }

                using (var client = new IeBizServiceClient())
                {
                    var customer = GetCustomer(order.CustomerAccessUserId, null);
                    TransactionRequestObject tran = new TransactionRequestObject()
                    {
                        Command = TransactionType.Capture.ToString(),
                        RefNum = order.TransactionNumber,
                        Details = new TransactionDetail()
                        {
                            OrderID = order.Id,
                            Description = "Capture from backend",
                            Amount = amount / 100.0,
                            Tax = 0.0
                        }
                    };
                    var response = client.runTransaction(GetToken(), tran);
                    LogEvent(order, response.Result);

                    switch (response.Result)
                    {
                        case "Approved":
                            {
                                if (amount < order.Price.PricePIP)
                                {

                                    if (final)
                                    {
                                        LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture(final)", amount / 100f), DebuggingInfoType.CaptureResult);
                                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Split capture successful");
                                    }
                                    else
                                    {
                                        LogEvent(order, string.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture", amount / 100f), DebuggingInfoType.CaptureResult);
                                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Split, "Split capture successful");
                                    }
                                }
                                else
                                {
                                    LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
                                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
                                }
                            }
                        default:
                            LogError(order, $"Incorrect response received({response.Result})");
                            return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, $"Incorrect response received({response.Result})");
                    }
                }

            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unexpected error during capture: {0}", ex.Message);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Unexpected error during capture");
            }
        }

        #endregion
    }
}
