﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Frontend.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Klarna.Api;
using Klarna.Checkout;
using System;
using System.Collections.Generic;
using System.Linq;
using KlarnaOrder = Klarna.Checkout.Order;
using Order = Dynamicweb.Ecommerce.Orders.Order;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.KlarnaCheckout
{
    /// <summary>
    /// Klarna Checkout Handler
    /// </summary>
    [
		AddInName("Klarna Checkout v2"),
		AddInDescription("Klarna Checkout Handler kco.v2")
	]
	public class KlarnaCheckout : CheckoutHandler, IRemoteCapture, ICheckOrderStatus, ICancelOrder, IChangeOrder, IAmountRefund, IPartialReturn, IFullReturn
	{
		const string ContentType = "application/vnd.klarna.checkout.aggregated-order-v2+json";

		#region Properties

		[AddInParameter("Merchant ID"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
		public string MerchantId { get; set; }

		[AddInParameter("Shared secret"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
		public string SharedSecret { get; set; }

		[AddInParameter("URL to terms and conditions"), AddInParameterEditor(typeof(PageSelectEditor), "")]
		public string UrlToTermsAndConditions { get; set; }

		[AddInParameter("Payment template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Payment")]
		public string PaymentTemplate { get; set; }

		[AddInParameter("Confirmation template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Confirmation")]
		public string ConfirmTemplate { get; set; }

		[AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Cancel")]
		public string CancelTemplate { get; set; }

		[AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Error")]
		public string ErrorTemplate { get; set; }

		[AddInParameter("Send invoice via e­-mail"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool SendInvoice { get; set; }

		[AddInParameter("Making the phone field mandatory")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool PhoneMandatory { get; set; }

		[AddInParameter("Allow separate shipping addresses")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool AllowSeparateShippingAddresses { get; set; }

		[AddInParameter("Enable packstation")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool EnablePackstation { get; set; }

		[AddInParameter("Making the date of birth mandatory")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool BirthDateMandatory { get; set; }

		[AddInParameter("Update billing info")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool UpdateBillingInfo { get; set; }

		[AddInParameter("Update shipping info")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool UpdateShippingInfo { get; set; }

		[AddInParameter("Exclude shipping address")]
		[AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool ExcludeShippingInfo { get; set; }

		[AddInParameter("Test mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
		public bool TestMode { get; set; }

		[AddInParameter("Klarna’s online order management tool"), AddInParameterEditor(typeof(LabelParameterEditor), "")]
		public string OnlineManagementTool { get; set; }

		#endregion

		#region Form properties

		private string FullUrlToTermsAndConditions()
		{
			var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
			var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);

			return string.Format("{0}://{1}{2}/{3}", Context.Current.Request.Url.Scheme, Context.Current.Request.Url.Host, portString, UrlToTermsAndConditions);
		}

		private static string CallbackUrl(Order order)
		{
			var portString = Context.Current.Request.Url.IsDefaultPort ? "" : string.Format(":{0}", Context.Current.Request.Url.Port);
			return string.Format(
				"{0}://{1}{2}/Admin/Public/Gateways/CheckoutHandler/Klarna/KlarnaCheckoutCallBack.aspx?OrderId={3}&klarna_order={{checkout.order.uri}}",
				Context.Current.Request.Url.Scheme, Context.Current.Request.Url.Host, portString, order.Id
			);
		}

		private string ConfirmUrl(Order order)
		{
			return string.Format("{0}&KlarnaState=Confirm&klarna_order={{checkout.order.uri}}", this.GetBaseUrl(order));
		}

		private string CancelUrl(Order order)
		{
			return string.Format("{0}&KlarnaState=Cancel", this.GetBaseUrl(order));
		}

		#endregion

		/// <summary>
		/// Default constructor
		/// </summary>
		public KlarnaCheckout()
		{
			OnlineManagementTool = @"<a href=""https://online.klarna.com"" target=""_blank"">https://online.klarna.com</a>";
		}

		/// <summary>
		/// Starts order checkout procedure
		/// </summary>
		/// <param name="order">Order to be checked out</param>
		/// <returns>String representation of template output</returns>
		public override string StartCheckout(Order order)
		{
			try
			{
				LogEvent(order, "Checkout started");

				if (!((order.CustomerCountryCode == "SE" && order.CurrencyCode == "SEK") || (order.CustomerCountryCode == "FI" && order.CurrencyCode == "EUR") || (order.CustomerCountryCode == "NO" && order.CurrencyCode == "NOK") || (order.CustomerCountryCode == "DE" && order.CurrencyCode == "EUR") || (order.CustomerCountryCode == "AT" && order.CurrencyCode == "EUR")))
				{
					throw new Exception(string.Format("Incorrect combination of customer country and currency: country: {0}, currency: {1}", order.CustomerCountryCode, order.CurrencyCode));
				}

				var klarnaOrder = CreateKlarnaOrder(order);
				klarnaOrder.Fetch();

				var temp = klarnaOrder.GetValue("gui");
				var gui = temp != null ? (dynamic)temp : null;
				var snippet = gui["snippet"];

				var paymentTemplate = new Template(PaymentTemplate);
				paymentTemplate.SetTag("KlarnaCheckout.CheckoutSnippet", snippet.ToString());
				paymentTemplate.SetTag("KlarnaCheckout.CancelURL", CancelUrl(order));

				return Render(order, paymentTemplate);
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (ConnectorException ex)
			{
				var msg = ex.Message;
				if (ex.Data.Contains("internal_message"))
				{
					msg = (string)ex.Data["internal_message"];
				}
				LogError(order, ex, "Unhandled exception with message: {0}", msg);

				return OnError(order, msg);
			}
			catch (Exception ex)
			{
				LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
				return OnError(order, ex.Message);
			}
		}

		/// <summary>
		/// Redirects user to payment step or canceled order step
		/// </summary>
		/// <param name="order">Order for processing</param>
		/// <returns>String representation of template output</returns>
		public override string Redirect(Order order)
		{
			try
			{
				LogEvent(order, "Redirected to Klarna Checkout");

				switch (Converter.ToString(Context.Current.Request["KlarnaState"]))
				{
					case "Confirm":
						return StateConfirm(order);

					case "Cancel":
						return StateCancel(order);

					default:
						Callback(order);
						return null;
				}
			}
			catch (System.Threading.ThreadAbortException)
			{
				return string.Empty;
			}
			catch (Exception ex)
			{
				LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
				return OnError(order, ex.Message);
			}
		}

		private string StateConfirm(Order order)
		{
			LogEvent(order, "State confirm");

            LogEvent(order, "Removing cart from session");
            CartCatch.ClearCart();

            var klarnaOrder = FetchKlarnaOrder();

			var temp = klarnaOrder.GetValue("gui");
			var gui = temp != null ? (dynamic)temp : null;
			var snippet = gui["snippet"];

			var confirmTemplate = new Template(ConfirmTemplate);

			confirmTemplate.SetTag("KlarnaCheckout.ConfirmationSnippet", snippet.ToString());

			return Render(order, confirmTemplate);
		}

		private string StateCancel(Order order)
		{
			LogEvent(order, "State cancel");

			order.TransactionStatus = "Cancelled";
			order.Save();

			CheckoutDone(order);

			var cancelTemplate = new Template(CancelTemplate);
			return Render(order, cancelTemplate);
		}

		private void Callback(Order order)
		{
			try
			{
				var klarnaOrder = FetchKlarnaOrder();

				if ((string)klarnaOrder.GetValue("status") == "checkout_complete")
				{
					var data = new Dictionary<string, object> {
						{ "status", "created" }
					};

					klarnaOrder.Update(data);
					var transact = klarnaOrder.GetValue("reservation").ToString();

					LogEvent(order, "Payment succeeded with reservation number: {0}; id: {1}, refference: {2}", transact, klarnaOrder.GetValue("id"), klarnaOrder.GetValue("reference"));

					var temp = klarnaOrder.GetValue("cart");
					var cart = temp != null ? (dynamic)temp : null;					
					order.TransactionAmount = double.Parse(cart["total_price_including_tax"].ToString());
					order.TransactionStatus = "Succeeded";
					order.GatewayPaymentStatus = OrderStatus.Accepted.ToString();

					if (UpdateBillingInfo || UpdateShippingInfo)
					{
						if (UpdateBillingInfo)
						{
							var temp2 = klarnaOrder.GetValue("billing_address");
							var addr = temp != null ? (dynamic)temp2 : null;							
							order.CustomerName = string.Format("{0} {1}", addr["given_name"], addr["family_name"]);
							order.CustomerFirstName = Converter.ToString(addr["given_name"]);
							order.CustomerSurname = Converter.ToString(addr["family_name"]);
							order.CustomerCountryCode = Converter.ToString(addr["country"]).ToUpperInvariant();
							order.CustomerCity = Converter.ToString(addr["city"]);
							order.CustomerZip = Converter.ToString(addr["postal_code"]);
							order.CustomerAddress = Converter.ToString(addr["street_address"]);
						}

						if (UpdateShippingInfo)
						{
							var temp2 = klarnaOrder.GetValue("shipping_address");
							var addr = temp != null ? (dynamic)temp2 : null;
							order.DeliveryName = string.Format("{0} {1}", addr["given_name"], addr["family_name"]);
							order.DeliveryFirstName = Converter.ToString(addr["given_name"]);
							order.DeliverySurname = Converter.ToString(addr["family_name"]);
							order.DeliveryCountryCode = Converter.ToString(addr["country"]).ToUpperInvariant();
							order.DeliveryCity = Converter.ToString(addr["city"]);
							order.DeliveryZip = Converter.ToString(addr["postal_code"]);
							order.DeliveryAddress = Converter.ToString(addr["street_address"]);
						}

						order.Save();
					}

					SetOrderComplete(order, transact);
				}
				else
				{
					LogEvent(order, "Klarna order status is incorrect: [0}", (string)klarnaOrder.GetValue("status"));
					LogError(order, "At least one callback validation error exists - exiting callback routine.");

					order.TransactionStatus = "Failed";
					order.Save();
				}
			}
			finally
			{
				if (order.Complete)
				{
					CheckoutDone(order);
				}
			}
		}

		private string OnError(Order order, string message)
		{
			order.TransactionAmount = 0;
			order.TransactionStatus = "Failed";
			order.Errors.Add(message);
			order.Save();

			order.DowngradeToCart();
			order.CartV2StepIndex = 0;
			order.TransactionStatus = string.Empty;
			Dynamicweb.Ecommerce.Common.Context.SetCart(order);

			if (string.IsNullOrWhiteSpace(ErrorTemplate))
			{
				RedirectToCart(order);
			}

			var errorTemplate = new Template(ErrorTemplate);
			errorTemplate.SetTag("CheckoutHandler:ErrorMessage", message);

			return Render(order, errorTemplate);
		}

		#region IRemoteCapture

		/// <summary>
		/// Send capture request to transaction service
		/// </summary>
		/// <param name="order">Order to be captured</param>
		/// <returns>Response from transaction service</returns>
		/// <remarks>https://developers.klarna.com/en/manage-orders/activate-an-order</remarks>
		public OrderCaptureInfo Capture(Order order)
		{
			if (order == null || string.IsNullOrEmpty(order.Id))
			{
				const string errorText = "No valid Order object set";
				LogError(order, errorText);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
			}
			if (string.IsNullOrEmpty(order.TransactionNumber))
			{
				const string errorText = "No transaction number set on the order";
				LogEvent(order, errorText);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
			}

			try
			{
				var api = GetApi(order);

				var response = api.Activate(order.TransactionNumber);
				if (SendInvoice)
				{
					api.EmailInvoice(response.InvoiceNumber);
				}
				order.TransactionNumber = response.InvoiceNumber;
				order.Save();

				var infoTxt = string.Format("Capture successful. Url: {0}, New invoice number: {1}", response.RiskStatus, response.InvoiceNumber);
				LogEvent(order, infoTxt, DebuggingInfoType.CaptureResult);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, infoTxt);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Remote capture during the call to Klarna. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
				return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
			}
		}

		#endregion

		#region Order management

		/// <summary>
		/// Gets order gateway payment status
		/// </summary>
		/// <param name="order">Order that needs to update payment status</param>
		public void CheckOrderStatus(Order order)
		{
			try
			{
				var api = GetApi(order);
				var orderStatus = api.CheckOrderStatus(order.TransactionNumber, OrderNumberType.InvoiceOrReservation);

				order.GatewayPaymentStatus = orderStatus.ToString();
				order.Save();

				var message = string.Format("Check order status succeed. Order status: {0}", orderStatus.ToString());
				LogEvent(order, message);
			}
			catch (KlarnaException ex)
			{
				var message = string.Format("Check order status failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, message);
			}
			catch (Exception ex)
			{
				var message = string.Format("Check order status failed with the message: {0}", ex.Message);
				LogError(order, ex, message);
			}
		}

		/// <summary>
		/// Cancels order reservation
		/// </summary>
		/// <param name="order">Order that needs to be canceled</param>
		/// <returns></returns>
		public bool CancelOrder(Order order)
		{
			var result = false;

			try
			{
				var api = GetApi(order);
				result = api.CancelReservation(order.TransactionNumber);

				LogEvent(order, "Cancel order succeed. Result: {0}", result);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Cancel order failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Cancel order failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
			}

			return result;
		}

		/// <summary>
		/// Updates modified order on Klarna server
		/// </summary>
		/// <param name="order">Order to be updated</param>
		/// <returns>True if operation succeeded</returns>
		public bool ChangeOrder(Order order)
		{
			var result = false;

			try
			{
				var api = GetApi(order);
				KlarnaInvoice.InitializeOrder(api, order);

				result = api.Update(order.TransactionNumber);

				LogEvent(order, "Change order succeed. Result: {0}", result);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Change order failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Change order failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
			}

			return result;
		}

		/// <summary>
		/// Refunds order
		/// </summary>
		/// <param name="order">Modified order</param>
		/// <param name="originalOrder">Original order</param>
		public void AmountRefund(Order order, Order originalOrder)
		{
			try
			{
				var api = GetApi(order);

				var invoiceNumber = api.ReturnAmount(originalOrder.TransactionNumber, PriceInfo.ConvertToPIP(order.Price.PriceWithVAT), order.Price.VATPercent, GoodsFlags.IncVAT, "Amount refund");
				order.TransactionNumber = invoiceNumber;
				order.Save();

				LogEvent(order, "Amount refund succeed. New invoice number: {0}", invoiceNumber);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Amount refund failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Amount refund failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
			}
		}

		/// <summary>
		/// Performs partial return of order
		/// </summary>
		/// <param name="order">Modified order</param>
		/// <param name="originalOrder">Original order</param>
		public void PartialReturn(Order order, Order originalOrder)
		{
			try
			{
				var api = GetApi(order);

				foreach (var item in CartItem.GetCartItems(order))
				{
					api.AddArticle(item.Quantity, item.Number, item.Title, item.PriceWithVAT, item.TaxRate, 0, item.Flags);
				}

				var invoiceNumber = api.CreditPart(originalOrder.TransactionNumber);

				order.TransactionNumber = invoiceNumber;
				order.Save();

				LogEvent(order, "Partial return succeed. New invoice number: {0}", invoiceNumber);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Partial return failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Partial return failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
			}
		}

		/// <summary>
		/// Performs full return of order
		/// </summary>
		/// <param name="order">Order that needs to be return</param>
		public void FullReturn(Order order)
		{
			try
			{
				var api = GetApi(order);

				var invoiceNumber = api.CreditInvoice(order.TransactionNumber);
				LogEvent(order, "Full return succeed. New invoice number: {0}", invoiceNumber);
			}
			catch (KlarnaException ex)
			{
				var infoTxt = string.Format("Full return failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
				LogError(order, ex, infoTxt);
			}
			catch (Exception ex)
			{
				var infoTxt = string.Format("Full return failed with the message: {0}", ex.Message);
				LogError(order, ex, infoTxt);
			}
		}

		#endregion

		#region Private members

		private KlarnaOrder FetchKlarnaOrder()
		{
			var checkoutId = new Uri(Context.Current.Request.QueryString["klarna_order"]);

			var connector = Connector.Create(SharedSecret);
			var klarnaOrder = new KlarnaOrder(connector, checkoutId)
			{
				ContentType = ContentType
			};

			klarnaOrder.Fetch();

			return klarnaOrder;
		}

		private KlarnaOrder CreateKlarnaOrder(Order order)
		{
			var cartItems = new List<Dictionary<string, object>>();
			foreach (var item in CartItem.GetCartItems(order))
			{
				var cartItem = new Dictionary<string, object> {
					{ "reference", item.Number},
					{ "name", item.Title },
					{ "quantity", item.Quantity },
					{ "unit_price", PriceInfo.ConvertToPIP(item.PriceWithVAT) },
					{ "tax_rate", PriceInfo.ConvertToPIP(item.TaxRate) }
				};

				if (!string.IsNullOrEmpty(item.ItemType))
				{
					cartItem["type"] = item.ItemType;
				}

				cartItems.Add(cartItem);
			}

			var cart = new Dictionary<string, object> { { "items", cartItems } };

			var merchant = new Dictionary<string, object> {
				{ "id", MerchantId },
				{ "terms_uri", this.FullUrlToTermsAndConditions() },
				{ "checkout_uri", this.GetBaseUrl(order) },
				{ "confirmation_uri", this.ConfirmUrl(order) },
				{ "push_uri", CallbackUrl(order)}
			};

			var options = new Dictionary<string, object>();
			if (!string.IsNullOrEmpty(order.ShippingMethodDescription)) options["shipping_details"] = order.ShippingMethodDescription.Length > 70 ? order.ShippingMethodDescription.Substring(0, 70) : order.ShippingMethodDescription;
			if (PhoneMandatory) options["phone_mandatory"] = PhoneMandatory;
			if (AllowSeparateShippingAddresses) options["allow_separate_shipping_address"] = AllowSeparateShippingAddresses;
			if (EnablePackstation) options["packstation_enabled"] = EnablePackstation;
			if (BirthDateMandatory) options["date_of_birth_mandatory"] = BirthDateMandatory;

			var data = new Dictionary<string, object>
					{
						{ "merchant_reference",
							new Dictionary<string, object> {
								{"orderid1", order.Id }
							}
						},
						{ "purchase_country", order.CustomerCountryCode },
						{ "purchase_currency", order.CurrencyCode},
						{ "locale", this.GetLocale(order.CustomerCountryCode) },
						{ "merchant", merchant},
						{ "cart", cart },
						{ "shipping_address", getShippingAddress(order) },
						{ "options", options }
					};
           

			var connector = Connector.Create(SharedSecret);
			var klarnaOrder = new KlarnaOrder(connector)
			{
				BaseUri = new Uri(TestMode ? "https://checkout.testdrive.klarna.com/checkout/orders" : "https://checkout.klarna.com/checkout/orders"),
				ContentType = ContentType
			};

			klarnaOrder.Create(data);

			return klarnaOrder;
		}

		private string GetLocale(string countryCode)
		{
			string locale = null;
			if (countryCode == "SE") locale = "sv-se";
			else if (countryCode == "NO") locale = "nb-no";
			else if (countryCode == "DE") locale = "de-de";
			else if (countryCode == "AT") locale = "de-at";
			else if (countryCode == "FI")
			{
				locale = Dynamicweb.Environment.ExecutingContext.GetCulture(true).Name.ToLowerInvariant() == "fi-fi" ? "fi-fi" : "sv-fi";
			}

			return locale;
		}

		private Dictionary<string, object> getShippingAddress(Order order)
		{
			var shippingAddress = new Dictionary<string, object>();

			bool isDeliveryFieldsFilled = !string.IsNullOrEmpty(order.DeliveryAddress) && !string.IsNullOrEmpty(order.DeliveryZip);
			if (isDeliveryFieldsFilled)
			{
				if (!string.IsNullOrEmpty(order.DeliveryFirstName) && !string.IsNullOrEmpty(order.DeliverySurname))
				{
					shippingAddress["given_name"] = order.DeliveryFirstName;
					shippingAddress["family_name"] = order.DeliverySurname;
				}
				else if (!string.IsNullOrEmpty(order.DeliveryName))
				{
					if (order.DeliveryName.Trim().Contains(' '))
					{
						string[] name = order.DeliveryName.Trim().Split(new char[] { ' ' }, 2, StringSplitOptions.RemoveEmptyEntries);
						shippingAddress["given_name"] = name[0];
						shippingAddress["family_name"] = name[1];
					}
					else
					{
						shippingAddress["given_name"] = order.DeliveryName;
					}
				}

                if (!ExcludeShippingInfo)
                {
                    if (order.DeliveryCountryCode == "DE" || order.DeliveryCountryCode == "AT")
                    {
                        if (!string.IsNullOrEmpty(order.DeliveryAddress)) shippingAddress["street_name"] = order.DeliveryAddress;
                        if (!string.IsNullOrEmpty(order.DeliveryHouseNumber)) shippingAddress["street_number"] = order.DeliveryHouseNumber;
                        if (order.DeliveryTitle == "Herr" || order.DeliveryTitle == "Frau") shippingAddress["title"] = order.DeliveryTitle;
                    }
                    else
                    {
                        shippingAddress["street_address"] = string.Format("{0} {1}", order.DeliveryAddress, order.DeliveryHouseNumber);
                    }
                    if (!string.IsNullOrEmpty(order.DeliveryAddress2)) shippingAddress["care_of"] = order.DeliveryAddress2;
                    if (!string.IsNullOrEmpty(order.DeliveryZip)) shippingAddress["postal_code"] = order.DeliveryZip;
                    if (!string.IsNullOrEmpty(order.DeliveryCity)) shippingAddress["city"] = order.DeliveryCity;
                    if (!string.IsNullOrEmpty(order.DeliveryCountryCode)) shippingAddress["country"] = order.DeliveryCountryCode;
                }
				if (!string.IsNullOrEmpty(order.DeliveryEmail)) shippingAddress["email"] = order.DeliveryEmail;
				if (!string.IsNullOrEmpty(order.DeliveryPhone)) shippingAddress["phone"] = order.DeliveryPhone;
			}
			else
			{
				if (!string.IsNullOrEmpty(order.CustomerFirstName) && !string.IsNullOrEmpty(order.CustomerSurname))
				{
					shippingAddress["given_name"] = order.CustomerFirstName;
					shippingAddress["family_name"] = order.CustomerSurname;
				}
				else if (!string.IsNullOrEmpty(order.CustomerName))
				{
					if (order.CustomerName.Trim().Contains(' '))
					{
						string[] name = order.CustomerName.Trim().Split(new char[] { ' ' }, 2, StringSplitOptions.RemoveEmptyEntries);
						shippingAddress["given_name"] = name[0];
						shippingAddress["family_name"] = name[1];
					}
					else
					{
						shippingAddress["given_name"] = order.CustomerName;
					}
				}

                if (!ExcludeShippingInfo)
                {
                    if (order.CustomerCountryCode == "DE" || order.CustomerCountryCode == "AT")
                    {
                        if (!string.IsNullOrEmpty(order.CustomerAddress)) shippingAddress["street_name"] = order.CustomerAddress;
                        if (!string.IsNullOrEmpty(order.CustomerHouseNumber)) shippingAddress["street_number"] = order.CustomerHouseNumber;
                        if (order.CustomerTitle == "Herr" || order.CustomerTitle == "Frau") shippingAddress["title"] = order.CustomerTitle;
                    }
                    else
                    {
                        shippingAddress["street_address"] = string.Format("{0} {1}", order.CustomerAddress, order.CustomerHouseNumber);
                    }
                    if (!string.IsNullOrEmpty(order.CustomerAddress2)) shippingAddress["care_of"] = order.CustomerAddress2;
                    if (!string.IsNullOrEmpty(order.CustomerZip)) shippingAddress["postal_code"] = order.CustomerZip;
                    if (!string.IsNullOrEmpty(order.CustomerCity)) shippingAddress["city"] = order.CustomerCity;
                    if (!string.IsNullOrEmpty(order.CustomerCountryCode)) shippingAddress["country"] = order.CustomerCountryCode;
                }
				if (!string.IsNullOrEmpty(order.CustomerEmail)) shippingAddress["email"] = order.CustomerEmail;
				if (!string.IsNullOrEmpty(order.CustomerPhone)) shippingAddress["phone"] = order.CustomerPhone;
			}

			return shippingAddress;
		}

		private Api GetApi(Order order)
		{
			Klarna.Api.Configuration configuration;
			var countryCode = order.CustomerCountryCode.ToUpperInvariant();
			var currencyCode = order.CurrencyCode.ToUpperInvariant();

			if (countryCode == "SE" && currencyCode == "SEK")
				configuration = new Klarna.Api.Configuration(Country.Code.SE, Language.Code.SV, Currency.Code.SEK, Encoding.Sweden);
			else if (countryCode == "DK" && currencyCode == "DKK")
				configuration = new Klarna.Api.Configuration(Country.Code.DK, Language.Code.DA, Currency.Code.DKK, Encoding.Denmark);
			else if (countryCode == "FI" && currencyCode == "EUR")
				configuration = new Klarna.Api.Configuration(Country.Code.FI, Language.Code.FI, Currency.Code.EUR, Encoding.Finland);
			else if (countryCode == "NO" && currencyCode == "NOK")
				configuration = new Klarna.Api.Configuration(Country.Code.NO, Language.Code.DE, Currency.Code.NOK, Encoding.Norway);
			else if (countryCode == "DE" && currencyCode == "EUR")
				configuration = new Klarna.Api.Configuration(Country.Code.DE, Language.Code.DE, Currency.Code.EUR, Encoding.Germany);
			else if (countryCode == "AT" && currencyCode == "EUR")
				configuration = new Klarna.Api.Configuration(Country.Code.AT, Language.Code.DE, Currency.Code.EUR, Encoding.Austria);
			else if (countryCode == "NL" && currencyCode == "EUR")
				configuration = new Klarna.Api.Configuration(Country.Code.NL, Language.Code.DE, Currency.Code.EUR, Encoding.Netherlands);
			else
			{
				throw new Exception(string.Format("Incorrect combination of customer country and currency: country: {0}, currency: {1}", countryCode, currencyCode));
			}

			configuration.Eid = Converter.ToInt32(MerchantId);
			configuration.Secret = SharedSecret;
			configuration.IsLiveMode = !TestMode;

			var api = new Api(configuration);
			var clientIP = (Context.Current.Request.ServerVariables["HTTP_X_FORWARDED_FOR"] == null) ? Context.Current.Request.UserHostAddress : Context.Current.Request.ServerVariables["HTTP_X_FORWARDED_FOR"];
			api.ClientIp = clientIP;

			return api;
		}

		/// <summary>
		/// Represents helper class for getting cart items information
		/// </summary>
		internal class CartItem
		{
			public string ItemType;
			public string Number;
			public string Title;
			public int Quantity;
			public double PriceWithVAT;
			public double TaxRate;
			public GoodsFlags Flags;

			/// <summary>
			/// Gets cart items based on order data
			/// </summary>
			/// <param name="order">Order instance</param>
			/// <returns>Set of CartItems</returns>
			public static IEnumerable<CartItem> GetCartItems(Order order)
			{
				var result = new List<CartItem>();

				foreach (var orderLine in order.OrderLines)
				{
					if (orderLine.HasType(OrderLineType.Product) || orderLine.HasType(OrderLineType.PointProduct))
					{
						result.Add(new CartItem()
						{
							Number = orderLine.Product.Number,
							Title = orderLine.Product.Name,
							Quantity = orderLine.AllowOverridePrices ? 1 : (int)orderLine.Quantity,
							PriceWithVAT = orderLine.AllowOverridePrices ? orderLine.Price.PriceWithVAT : orderLine.UnitPrice.PriceWithVAT,
							TaxRate = orderLine.Price.VATPercent,
							Flags = GoodsFlags.IncVAT
						});
					}
					else
					{
						var isDiscount = orderLine.HasType(OrderLineType.Discount) || orderLine.HasType(OrderLineType.ProductDiscount);
						result.Add(new CartItem()
						{
							ItemType = isDiscount ? "discount" : "physical",
							Number = isDiscount ? orderLine.DiscountId : orderLine.Id,
							Title = orderLine.ProductName,
							Quantity = 1,
							PriceWithVAT = orderLine.Price.PriceWithVAT,
							TaxRate = 0,
							Flags = GoodsFlags.IncVAT
						});
					}
				}

				var shipping = order.ShippingFee;
				if (shipping.Price > 0)
				{
					result.Add(new CartItem()
					{
						ItemType = "shipping_fee",
						Number = "SHIPPINGFEE",
						Title = "Shipping Fee",
						Quantity = 1,
						PriceWithVAT = shipping.PriceWithVAT,
						TaxRate = shipping.VATPercent,
						Flags = GoodsFlags.Shipping | GoodsFlags.IncVAT
					});
				}

				var payment = order.PaymentFee;
				if (payment.Price > 0)
				{
					result.Add(new CartItem()
					{
						Number = "PAYMENTFEE",
						Title = "Payment Fee",
						Quantity = 1,
						PriceWithVAT = payment.PriceWithVAT,
						TaxRate = payment.VATPercent,
						Flags = GoodsFlags.IncVAT
					});
				}

				double orderDetailsAmount = 0, roundingError = 0;
				foreach (CartItem item in result)
				{
					item.PriceWithVAT = Math.Round(item.PriceWithVAT, 2, MidpointRounding.AwayFromZero);
					orderDetailsAmount += item.PriceWithVAT * item.Quantity;
				}
				roundingError = Math.Round(order.Price.PriceWithVAT, 2, MidpointRounding.AwayFromZero) - orderDetailsAmount;
				if (roundingError != 0 && Math.Abs(roundingError) <= 0.1)
				{
					result.Add(new CartItem()
					{
						Number = "ROUNDINGERROR",
						Title = "ROUNDING ERROR CORRECTION",
						Quantity = 1,
						PriceWithVAT = roundingError,
						TaxRate = 0,
						Flags = GoodsFlags.IncVAT
					});
				}

				return result;
			}
		}

		#endregion
	}
}
