﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Klarna.Api;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.KlarnaCheckout
{
    /// <summary>
    /// Klarna invoice and part payment Handler
    /// </summary>
    [
        AddInName("Klarna invoice and part payment"),
        AddInDescription("Klarna invoice and part payment")
    ]
    public class KlarnaInvoice : CheckoutHandler, IRemoteCapture, IActionController
    {

        #region Properties

        [AddInParameter("Merchant ID"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
        public string MerchantId { get; set; }

        [AddInParameter("Shared secret"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
        public string SharedSecret { get; set; }

        [AddInParameter("Select payment method template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaInvoice/SelectPaymentMethod")]
        public string SelectPaymentMethodTemplate { get; set; }

        [AddInParameter("Payment template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaInvoice/Payment")]
        public string PaymentTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Send invoice via e­-mail"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool SendInvoice { get; set; }

        [AddInParameter("Lock payment method to invoice"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool LockPaymentMethodToInvoice { get; set; }

        [AddInParameter("Update billing info")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UpdateBillingInfo { get; set; }

        [AddInParameter("Update shipping info")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UpdateShippingInfo { get; set; }

        [AddInParameter("Test mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        [AddInParameter("Klarna’s online order management tool"), AddInParameterEditor(typeof(LabelParameterEditor), "")]
        public string OnlineManagementTool { get; set; }

        #endregion

        #region Form properties

        private string PClassStorageUri
        {
            get { return Context.Current.Server.MapPath(string.Format(@"\Files\System\Klarna\Merchant_{0}\pclasses.xml", MerchantId)); }
        }

        private string GetAddressUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=GetAddress", this.GetBaseUrl(order));
        }

        private string BuyUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Buy", this.GetBaseUrl(order));
        }

        private string CancelUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Cancel", this.GetBaseUrl(order));
        }

        private string PaymentUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Payment", this.GetBaseUrl(order));
        }

        #endregion

        public KlarnaInvoice()
        {
            OnlineManagementTool = @"<a href=""https://online.klarna.com"" target=""_blank"">https://online.klarna.com</a>";
        }

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");

                if (!(
                    (order.CustomerCountryCode == "SE" && order.CurrencyCode == "SEK") ||
                    (order.CustomerCountryCode == "DK" && order.CurrencyCode == "DKK") ||
                    (order.CustomerCountryCode == "FI" && order.CurrencyCode == "EUR") ||
                    (order.CustomerCountryCode == "NO" && order.CurrencyCode == "NOK") ||
                    (order.CustomerCountryCode == "DE" && order.CurrencyCode == "EUR") ||
                    (order.CustomerCountryCode == "AT" && order.CurrencyCode == "EUR") ||
                    (order.CustomerCountryCode == "NL" && order.CurrencyCode == "EUR")))
                {
                    throw new Exception(string.Format("Incorrect combination of customer country and currency: country: {0}, currency: {1}", order.CustomerCountryCode, order.CurrencyCode));
                }

                if (LockPaymentMethodToInvoice)
                {
                    return RenderPaymentForm(order, PClass.Invoice);
                }
                else
                {
                    return RenderPaymentMethods(order);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (KlarnaException ex)
            {
                var infoTxt = string.Format("Operation failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
                LogError(order, ex, infoTxt);
                return OnError(order, infoTxt);
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return OnError(order, ex.Message);
            }
        }

        #region RenderPaymentMethods

        private string RenderPaymentMethods(Order order)
        {
            var selectTemplate = new Template(SelectPaymentMethodTemplate);

            selectTemplate.SetTag("Klarna.EID", MerchantId);
            selectTemplate.SetTag("Klarna.Country", order.CustomerCountryCode);
            selectTemplate.SetTag("Klarna.Locale", getLocale(order.CustomerCountryCode));
            selectTemplate.SetTag("Klarna.CancelURL", CancelUrl(order));
            selectTemplate.SetTag("Klarna.PaymentURL", PaymentUrl(order));

            var file = new FileInfo(PClassStorageUri);
            if (file.Exists)
            {
                var api = GetApi(order.CustomerCountryCode, order.CurrencyCode);

                api.PClassStorageUri = file.FullName;
                var classes = api.GetPClasses(PClass.TypeCode.Special);
                if (classes.Any())
                {
                    RenderPaymentMethodInvoiceSpecial(selectTemplate, order, classes.First());
                }

                classes = api.GetPClasses(PClass.TypeCode.Account);
                if (classes.Any())
                {
                    RenderPaymentMethodPPFlexible(selectTemplate, order, classes.First(), api);
                    selectTemplate.SetTag("Klarna.PartPaymentsExists", true);
                }

                classes = api.GetPClasses(PClass.TypeCode.Campaign);
                if (classes.Any())
                {
                    selectTemplate.SetTag("Klarna.Fixed", true);
                    if (selectTemplate.LoopExists("Campaigns"))
                    {
                        selectTemplate.SetTag("Klarna.PartPaymentsExists", true);
                        var campaignTemplate = selectTemplate.GetLoop("Campaigns");

                        foreach (var pclass in classes.Where(c1 => c1.MinimumPurchaseAmount <= order.Price.PriceWithVAT).OrderBy(c2 => c2.Months).Take(3))
                        {
                            RenderPaymentMethodPPFixed(campaignTemplate, order, pclass);
                            campaignTemplate.CommitLoop();
                        }
                    }
                }
            }            

            return Render(order, selectTemplate);
        }

        private void RenderPaymentMethodInvoiceSpecial(Template paymentTemplate, Order order, PClass pclass)
        {
            var calc = new Calculation();
            paymentTemplate.SetTag("Klarna.InvoiceSpecial", true);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.Id", pclass.Id);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.Description", pclass.Description);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.InterestRate", pclass.InterestRate);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.AnnualRate", calc.CalculateAPR(order.Price.PriceWithVAT, pclass, Calculation.CalculateOn.CheckoutPage, 0));
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.StartFee", pclass.StartFee);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.InvoiceFee", pclass.InvoiceFee);

            var totalCreditPurchaseCost = calc.TotalCreditPurchaseCost(order.Price.PriceWithVAT, pclass, Calculation.CalculateOn.CheckoutPage);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.CreditCost", totalCreditPurchaseCost - order.Price.PriceWithVAT);
            paymentTemplate.SetTag("Klarna.InvoiceSpecial.TotalAmountPayable", totalCreditPurchaseCost);
        }

        private void RenderPaymentMethodPPFlexible(Template paymentTemplate, Order order, PClass pclass, Api api)
        {
            var calc = new Calculation();
            paymentTemplate.SetTag("Klarna.Flexible", true);
            paymentTemplate.SetTag("Klarna.Flexible.Id", pclass.Id);
            paymentTemplate.SetTag("Klarna.Flexible.Description", pclass.Description);
            paymentTemplate.SetTag("Klarna.Flexible.InterestRate", pclass.InterestRate);
            paymentTemplate.SetTag("Klarna.Flexible.StartFee", pclass.StartFee);
            paymentTemplate.SetTag("Klarna.Flexible.InvoiceFee", pclass.InvoiceFee);
            paymentTemplate.SetTag("Klarna.Flexible.MonthlyPay", calc.CalculateMonthlyCost(order.Price.PriceWithVAT, api.GetCheapestPClass(order.Price.PriceWithVAT, Calculation.CalculateOn.CheckoutPage), Calculation.CalculateOn.CheckoutPage));
        }

        private void RenderPaymentMethodPPFixed(Template paymentTemplate, Order order, PClass pclass)
        {
            var calc = new Calculation();
            paymentTemplate.SetTag("Klarna.Fixed.Id", pclass.Id);
            paymentTemplate.SetTag("Klarna.Fixed.Description", pclass.Description);
            paymentTemplate.SetTag("Klarna.Fixed.InterestRate", pclass.InterestRate);
            paymentTemplate.SetTag("Klarna.Fixed.AnnualRate", calc.CalculateAPR(order.Price.PriceWithVAT, pclass, Calculation.CalculateOn.CheckoutPage, 0));
            paymentTemplate.SetTag("Klarna.Fixed.StartFee", pclass.StartFee);
            paymentTemplate.SetTag("Klarna.Fixed.InvoiceFee", pclass.InvoiceFee);
            paymentTemplate.SetTag("Klarna.Fixed.Months", pclass.Months);
            paymentTemplate.SetTag("Klarna.Fixed.MonthlyPay", calc.CalculateMonthlyCost(order.Price.PriceWithVAT, pclass, Calculation.CalculateOn.CheckoutPage));

            var totalCreditPurchaseCost = calc.TotalCreditPurchaseCost(order.Price.PriceWithVAT, pclass, Calculation.CalculateOn.CheckoutPage);
            paymentTemplate.SetTag("Klarna.Fixed.CreditCost", totalCreditPurchaseCost - order.Price.PriceWithVAT);
            paymentTemplate.SetTag("Klarna.Fixed.TotalAmountPayable", totalCreditPurchaseCost);
        }

        #endregion

        #region RenderPayment

        private string RenderPaymentForm(Order order, int PClassID = 0)
        {
            if (PClassID == 0)
            {
                PClassID = Converter.ToInt32(Dynamicweb.Context.Current.Request["PClassId"]);
                if (PClassID == 0)
                {
                    throw new Exception("Payment method is not defined!");
                }
            }

            var paymentTemplate = new Template(PaymentTemplate);

            paymentTemplate.SetTag("Klarna.EID", MerchantId);
            paymentTemplate.SetTag("Klarna.Country", order.CustomerCountryCode);
            paymentTemplate.SetTag("Klarna.Locale", getLocale(order.CustomerCountryCode));
            paymentTemplate.SetTag("Klarna.PClassId", PClassID);

            paymentTemplate.SetTag("Klarna.BaseURL", GetBaseUrl(order));
            paymentTemplate.SetTag("Klarna.CancelURL", CancelUrl(order));
            paymentTemplate.SetTag("Klarna.BuyURL", BuyUrl(order));

            paymentTemplate.SetTag("Klarna.UpdateBillingInfo", UpdateBillingInfo);
            paymentTemplate.SetTag("Klarna.UpdateShippingInfo", UpdateShippingInfo);

            return Render(order, paymentTemplate);
        }

        #endregion

        /// <summary>
        /// Redirects user to buy, payment, cancel or get address step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(order, "Redirected to Klarna Invoice");

                switch (Converter.ToString(Context.Current.Request["KlarnaState"]))
                {
                    case "Buy":
                        return StateBuy(order);

                    case "Payment":
                        return RenderPaymentForm(order);

                    case "Cancel":
                        return StateCancel(order);

                    case "GetAddress":
                        return StateGetAddress(order);

                    default:
                        break;
                }
                return null;
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (KlarnaException ex)
            {
                var infoTxt = string.Format("Operation failed. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
                LogError(order, ex, infoTxt);
                return OnError(order, infoTxt);
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return OnError(order, ex.Message);
            }
        }

        private string StateGetAddress(Order order)
        {
            LogEvent(order, "State get address");

            if (!UpdateBillingInfo || !UpdateShippingInfo)
            {
                Context.Current.Response.ContentType = "application/json";
                Context.Current.Response.Write(Converter.SerializeCompact(new { Error = "Changing of billing/shipping info is not allowed." }));
            }
            else
            {
                var api = GetApi(order.CustomerCountryCode, order.CurrencyCode);

                var customerPNO = Converter.ToString(Context.Current.Request["CustomerPNO"]);
                if (!string.IsNullOrEmpty(customerPNO))
                {
                    var addresses = new List<Address>();
                    try
                    {
                        addresses = api.GetAddresses(customerPNO, Encoding.Sweden, GetAddressFlags.Given);
                    }
                    catch (KlarnaException e)
                    {
                        Context.Current.Response.ContentType = "application/json";
                        Context.Current.Response.Write(Converter.SerializeCompact(new { Error = e.Message }));
                    }
                    if (addresses.Any())
                    {
                        Context.Current.Response.ContentType = "application/json";
                        Context.Current.Response.Write(Converter.SerializeCompact(addresses.First()));
                    }
                }
            }

            Context.Current.Response.End();
            return null;
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            order.Save();

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            return Render(order, cancelTemplate);
        }

        private string StateBuy(Order order)
        {
            LogEvent(order, "State buy");

            if (UpdateBillingInfo && Converter.ToBoolean(Dynamicweb.Context.Current.Request["UpdateBillingInfo"]))
            {
                order.CustomerFirstName = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerFirstName"]);
                order.CustomerSurname = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerLastName"]);
                order.CustomerCity = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerCity"]);
                order.CustomerAddress = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerAddress"]);
                order.CustomerHouseNumber = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerHouseNumber"]);
                order.CustomerZip = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerZip"]);
                order.CustomerPhone = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerPhone"]);
                order.CustomerCell = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerCell"]);
                order.CustomerEmail = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerEmail"]);
            }

            if (UpdateShippingInfo && Converter.ToBoolean(Dynamicweb.Context.Current.Request["UpdateShippingInfo"]))
            {
                order.DeliveryFirstName = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryFirstName"]);
                order.DeliverySurname = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryLastName"]);
                order.DeliveryCity = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryCity"]);
                order.DeliveryAddress = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryAddress"]);
                order.DeliveryHouseNumber = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryHouseNumber"]);
                order.DeliveryZip = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryZip"]);
                order.DeliveryPhone = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryPhone"]);
                order.DeliveryCell = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryCell"]);
                order.DeliveryEmail = Converter.ToString(Dynamicweb.Context.Current.Request["DeliveryEmail"]);
            }

            var api = GetApi(order.CustomerCountryCode, order.CurrencyCode);

            var invoice = PClass.Invoice;
            if (!LockPaymentMethodToInvoice)
            {
                invoice = Converter.ToInt32(Dynamicweb.Context.Current.Request["PClassId"]);
                if (invoice != -1)
                {
                    api.PClassStorageUri = PClassStorageUri;
                    var pclass = api.GetPClass(invoice);

                    if (pclass.Type == PClass.TypeCode.Campaign && pclass.MinimumPurchaseAmount > order.Price.PriceWithVAT)
                    {
                        throw new Exception(string.Format("Selected payment method isn't allowed for order amount less {0}", pclass.MinimumPurchaseAmount));
                    }
                }
            }

            var customerPno = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerPNO"]);
            var customerDateOfBirth = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerDateOfBirth"]);
            Gender? gender = null;

            bool isGermany = order.CustomerCountryCode == "AT" || order.CustomerCountryCode == "DE" || order.CustomerCountryCode == "NL";
            if (isGermany)
            {
                if (string.IsNullOrEmpty(customerDateOfBirth)) throw new Exception("Custromer date of birth is empty!");
                customerDateOfBirth = customerDateOfBirth.Replace("/", "").Replace("-", "");

                var customerGender = Converter.ToString(Dynamicweb.Context.Current.Request["CustomerGender"]);
                if (!string.IsNullOrEmpty(customerGender))
                    gender = customerGender.ToLowerInvariant() == "male" ? Gender.Male : Gender.Female;
                else
                    gender = order.CustomerTitle.ToLowerInvariant() == "herr" ? Gender.Male : Gender.Female;
            }
            else
            {
                if (string.IsNullOrEmpty(customerPno)) throw new Exception("Custromer PNO is empty!");
            }

            KlarnaInvoice.InitializeOrder(api, order);

            var response = api.ReserveAmount(
                isGermany ? customerDateOfBirth : customerPno,
                gender,
                null,
                ReservationFlags.NoFlag,
                invoice,
                api.PnoEncoding,
                false
            );

            if (response.OrderStatus != OrderStatus.Denied)
            {
                LogEvent(order, "Order was registered in Klarna with reservation number: {0}; order status is: {1}", response.ReservationNumber, response.OrderStatus.ToString());
                order.TransactionAmount = order.Price.PriceWithVAT;
                order.TransactionStatus = "Succeeded";
                order.GatewayPaymentStatus = response.OrderStatus.ToString();

                SetOrderComplete(order, response.ReservationNumber);
                CheckoutDone(order);
                RedirectToCart(order);
            }
            else
            {
                LogEvent(order, "Order status is: Denied");
                return OnError(order, "Order status is: Denied");
            }

            return null;
        }

        private string OnError(Order order, string msg)
        {
            order.TransactionAmount = 0;
            order.TransactionStatus = "Failed";
            order.Errors.Add(msg);
            order.Save();

            order.DowngradeToCart();
            order.CartV2StepIndex = 0;
            order.TransactionStatus = string.Empty;
            Common.Context.SetCart(order);

            if (string.IsNullOrWhiteSpace(ErrorTemplate))
            {
                RedirectToCart(order);
            }

            var errorTemplate = new Template(ErrorTemplate);
            errorTemplate.SetTag("CheckoutHandler:ErrorMessage", msg);

            return Render(order, errorTemplate);
        }

        #region ActionController

        private void ClearPClasses(string countryCode, string currencyCode)
        {
            var file = new FileInfo(PClassStorageUri);
            if (file.Exists)
            {
                var api = GetApi(countryCode, currencyCode);
                api.PClassStorageUri = file.FullName;
                api.ClearPClasses();
            }
        }

        private void FetchPClasses(string countryCode, string currencyCode)
        {
            var folder = new DirectoryInfo(Context.Current.Server.MapPath(string.Format(@"\Files\System\Klarna\Merchant_{0}", MerchantId)));
            if (!folder.Exists) folder.Create();

            var api = GetApi(countryCode, currencyCode);
            api.PClassStorageUri = PClassStorageUri;

            var countries = new Country.Code[] { Country.Code.AT, Country.Code.DE, Country.Code.DK, Country.Code.FI, Country.Code.NL, Country.Code.NO, Country.Code.SE };
            foreach (Country.Code country in countries)
            {
                try
                {
                    api.FetchPClasses(country);
                    LogEvent(null, string.Format("PClasses for country: {0} were fetched successful", country));
                }
                catch (Exception e)
                {
                    LogEvent(null, string.Format("Exception when fetched PClasses for country: {0}. Exception: {1}", country, e.Message));
                }
            }
        }

        /// <summary>
        /// Fires action
        /// </summary>
        /// <param name="action">Fetch or clear action</param>
        /// <param name="args">not used</param>
        public void FireAction(string action, params string[] args)
        {
            LogEvent(null, string.Format("Fire action: {0}", action));

            if (action == "FetchPClasses")
            {
                FetchPClasses("SE", string.Empty);
            }
            else if (action == "ClearPClasses")
            {
                ClearPClasses("SE", "");
            }
        }

        /// <summary>
        /// Retrieves key-value pairs of supported actions data
        /// </summary>
        /// <returns>Key-value pairs of actions and their translated names</returns>
        public IDictionary<string, string> SupportedActions()
        {
            var actions = new Dictionary<string, string> {
                { "ClearPClasses", "Clear PClasses" },
                { "FetchPClasses", "Fetch PClasses" }
            };

            return actions;
        }

        #endregion

        #region IRemoteCapture

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        /// <remarks>https://developers.klarna.com/en/manage-orders/activate-an-order</remarks>
        public OrderCaptureInfo Capture(Order order)
        {
            if (order == null || string.IsNullOrEmpty(order.Id))
            {
                const string errorText = "No valid Order object set";
                LogError(order, errorText);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
            }
            else if (string.IsNullOrEmpty(order.TransactionNumber))
            {
                const string errorText = "No transaction number set on the order";
                LogEvent(order, errorText);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
            }
            else if (order.GatewayPaymentStatus != OrderStatus.Accepted.ToString())
            {
                string errorText = string.Format("Order should have 'Accepted' payment state for activation. Current state: {0}", order.GatewayPaymentStatus);
                LogEvent(order, errorText);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
            }

            try
            {
                var api = GetApi(order.CustomerCountryCode, order.CurrencyCode);

                var response = api.Activate(order.TransactionNumber);
                if (SendInvoice)
                {
                    api.EmailInvoice(response.InvoiceNumber);
                }
                order.TransactionNumber = response.InvoiceNumber;
                order.Save();

                var infoTxt = string.Format("Capture successful. Url: {0}, New invoice number: {1}", response.RiskStatus, response.InvoiceNumber);
                LogEvent(order, infoTxt, DebuggingInfoType.CaptureResult);

                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, infoTxt);
            }
            catch (KlarnaException ex)
            {
                var infoTxt = string.Format("Remote capture during the call to Klarna. Error code: {0}, message: {1}", ex.ErrorCode, ex.Message);
                LogError(order, ex, infoTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        #endregion

        #region Private members

        private Api GetApi(string countryCode, string currencyCode)
        {
            Klarna.Api.Configuration configuration;

            if (countryCode == "SE" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "SEK"))
                configuration = new Klarna.Api.Configuration(Country.Code.SE, Language.Code.SV, Currency.Code.SEK, Encoding.Sweden);
            else if (countryCode == "DK" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "DKK"))
                configuration = new Klarna.Api.Configuration(Country.Code.DK, Language.Code.DA, Currency.Code.DKK, Encoding.Denmark);
            else if (countryCode == "FI" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "EUR"))
                configuration = new Klarna.Api.Configuration(Country.Code.FI, Language.Code.FI, Currency.Code.EUR, Encoding.Finland);
            else if (countryCode == "NO" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "NOK"))
                configuration = new Klarna.Api.Configuration(Country.Code.NO, Language.Code.NB, Currency.Code.NOK, Encoding.Norway);
            else if (countryCode == "DE" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "EUR"))
                configuration = new Klarna.Api.Configuration(Country.Code.DE, Language.Code.DE, Currency.Code.EUR, Encoding.Germany);
            else if (countryCode == "AT" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "EUR"))
                configuration = new Klarna.Api.Configuration(Country.Code.AT, Language.Code.DE, Currency.Code.EUR, Encoding.Austria);
            else if (countryCode == "NL" && (string.IsNullOrEmpty(currencyCode) || currencyCode == "EUR"))
                configuration = new Klarna.Api.Configuration(Country.Code.NL, Language.Code.NL, Currency.Code.EUR, Encoding.Netherlands);
            else
            {
                throw new Exception(string.Format("Incorrect combination of customer country and currency: country: {0}, currency: {1}", countryCode, currencyCode));
            }
            configuration.Eid = Converter.ToInt32(MerchantId);
            configuration.Secret = SharedSecret;
            configuration.IsLiveMode = !TestMode;

            var api = new Api(configuration);
            var clientIP = (Context.Current.Request.ServerVariables["HTTP_X_FORWARDED_FOR"] == null) ? Context.Current.Request.UserHostAddress : Context.Current.Request.ServerVariables["HTTP_X_FORWARDED_FOR"];
            api.ClientIp = clientIP;

            return api;
        }

        private string getLocale(string countryCode)
        {
            string locale = null;
            if (countryCode == "SE") locale = "sv_se";
            else if (countryCode == "DK") locale = "da_dk";
            else if (countryCode == "NO") locale = "nb_no";
            else if (countryCode == "NL") locale = "nl_nl";
            else if (countryCode == "DE") locale = "de_de";
            else if (countryCode == "AT") locale = "de_at";
            else if (countryCode == "FI")
            {
                locale = Dynamicweb.Environment.ExecutingContext.GetCulture(true).Name.ToLowerInvariant() == "fi-fi" ? "fi_fi" : "sv_fi";
            }

            return locale;
        }

        /// <summary>
        /// Initialize order object with values
        /// </summary>
        /// <param name="api">klarna api instance, that provides chechout methods</param>
        /// <param name="order">Oredr instance</param>
        internal static void InitializeOrder(Api api, Order order)
        {
            api.OrderId1 = order.Id;

            foreach (var item in KlarnaCheckout.CartItem.GetCartItems(order))
            {
                api.AddArticle(item.Quantity, item.Number, item.Title, item.PriceWithVAT, item.TaxRate, 0, item.Flags);
            }

            var billingAddress = new Address()
            {
                Email = order.CustomerEmail,
                PhoneNumber = order.CustomerPhone,
                CellPhoneNumber = order.CustomerCell,
                FirstName = order.CustomerFirstName,
                LastName = order.CustomerSurname,
                Street = order.CustomerAddress,
                CareOf = order.CustomerAddress2,
                ZipCode = order.CustomerZip,
                City = order.CustomerCity,
                Country = GetCountryCode(order.CustomerCountryCode),
                HouseNumber = order.CustomerHouseNumber,
                HouseExtension = string.Empty,
                CompanyName = order.CustomerCompany
            };
            api.BillingAddress = billingAddress;

            bool isDeliveryFieldsFilled = !string.IsNullOrEmpty(order.DeliveryAddress) && !string.IsNullOrEmpty(order.DeliveryZip);
            if (isDeliveryFieldsFilled)
            {
                var shippingAddress = new Address()
                {
                    Email = order.DeliveryEmail,
                    PhoneNumber = order.DeliveryPhone,
                    CellPhoneNumber = order.DeliveryCell,
                    FirstName = order.DeliveryFirstName,
                    LastName = order.DeliverySurname,
                    Street = order.DeliveryAddress,
                    CareOf = order.DeliveryAddress2,
                    ZipCode = order.DeliveryZip,
                    City = order.DeliveryCity,
                    Country = GetCountryCode(order.DeliveryCountryCode),
                    HouseNumber = order.DeliveryHouseNumber,
                    HouseExtension = string.Empty,
                    CompanyName = order.DeliveryCompany
                };

                api.ShippingAddress = shippingAddress;
            }
            else
            {
                api.ShippingAddress = billingAddress;
            }
        }

        private static Country.Code GetCountryCode(string countryCode)
        {
            var rslt = Country.Code.Undefined;

            switch (countryCode.ToUpperInvariant())
            {
                case "SE":
                    rslt = Country.Code.SE;
                    break;
                case "DK":
                    rslt = Country.Code.DK;
                    break;
                case "FI":
                    rslt = Country.Code.FI;
                    break;
                case "NO":
                    rslt = Country.Code.NO;
                    break;
                case "DE":
                    rslt = Country.Code.DE;
                    break;
                case "AT":
                    rslt = Country.Code.AT;
                    break;
                case "NL":
                    rslt = Country.Code.NL;
                    break;
                default:
                    throw new Exception(string.Format("Incorrect country code: {0}", countryCode));
            }

            return rslt;
        }

        #endregion
    }
}
