﻿using Dynamicweb.Core;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.CheckoutHandlers.KlarnaCheckout_v3.Model;
using Dynamicweb.Ecommerce.Frontend.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Text.Json;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.KlarnaCheckout_v3
{
    /// <summary>
    /// Klarna Checkout Handler
    /// </summary>
    [
        AddInName("Klarna Checkout v3 - BETA"),
        AddInDescription("Klarna Checkout Handler v3")
    ]
    public class KlarnaCheckout : CheckoutHandler, IRemoteCapture, IRemotePartialCapture, ICancelOrder, IAmountRefund, IPartialReturn, IFullReturn
    {
        #region Properties

        [AddInParameter("Api username(UID)"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
        public string ApiUID { get; set; }

        [AddInParameter("Api password"), AddInParameterEditor(typeof(TextParameterEditor), "NewGUI=true;")]
        public string ApiPassword { get; set; }

        [AddInParameter("URL to terms and conditions"), AddInParameterEditor(typeof(PageSelectEditor), "")]
        public string UrlToTermsAndConditions { get; set; }

        [AddInParameter("Payment template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Payment")]
        public string PaymentTemplate { get; set; }

        [AddInParameter("Confirmation template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Confirmation")]
        public string ConfirmTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/KlarnaCheckout/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Allow separate shipping addresses")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool AllowSeparateShippingAddresses { get; set; }

        [AddInParameter("Update billing info")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UpdateBillingInfo { get; set; }

        [AddInParameter("Update shipping info")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UpdateShippingInfo { get; set; }

        [AddInParameter("Exclude shipping address")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool ExcludeShippingInfo { get; set; }

        [AddInParameter("Allow to choose between the B2C or B2B flows"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool AllowSwitchCustomerTypes { get; set; }

        [AddInParameter("Set defaut flow to B2B instead of B2C"), AddInParameterEditor(typeof(YesNoParameterEditor), "info=\"affects only when alowed to choose betwin B2C and B2B flows\"")]
        public bool DefaultFlowIsB2B { get; set; }

        [AddInParameter("Use Klarna receipt instead of cart module receipt"), AddInParameterEditor(typeof(YesNoParameterEditor), "info=\"affects only when confirm template is selected\"")]
        public bool UseKlarnaOrderReceipt { get; set; }

        [AddInParameter("Test mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        [AddInParameter("Klarnas online order management tool"), AddInParameterEditor(typeof(LabelParameterEditor), "")]
        public string OnlineManagementTool { get; set; }

        #endregion

        #region Form properties

        private string FullUrlToTermsAndConditions()
        {
            var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);

            return string.Format("{0}://{1}{2}/{3}", Context.Current.Request.Url.Scheme, Context.Current.Request.Url.Host, portString, UrlToTermsAndConditions);
        }

        private string CallbackUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Completed", GetBaseUrl(order));
        }

        private string ConfirmUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Confirm&klarna_order={{checkout.order.uri}}", GetBaseUrl(order));
        }

        private string CancelUrl(Order order)
        {
            return string.Format("{0}&KlarnaState=Cancel", GetBaseUrl(order));
        }

        #endregion

        /// <summary>
        /// Default constructor
        /// </summary>
        public KlarnaCheckout()
        {
            OnlineManagementTool = @"<a href=""https://playground.eu.portal.klarna.com"" target=""_blank"">EU portal</a> <a href=""https://orders.playground.us.portal.klarna.com"" target=""_blank"">US portal</a>";
        }

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");
               
                var klarnaOrder = CreateKlarnaOrder(order);
                if (klarnaOrder.HasError)
                {
                    string errorMessage = klarnaOrder.GetErrorMessage();
                    LogError(order, $"Create a new order request returned the error with code: '{klarnaOrder.error_code}' and message: '{errorMessage}'");
                    return OnError(order, errorMessage);
                }

                var paymentTemplate = new Template(PaymentTemplate);
                paymentTemplate.SetTag("KlarnaCheckout.CheckoutSnippet", klarnaOrder.html_snippet);
                paymentTemplate.SetTag("KlarnaCheckout.CancelURL", CancelUrl(order));

                return Render(order, paymentTemplate);
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                var msg = ex.Message;
                if (ex.Data.Contains("internal_message"))
                {
                    msg = (string)ex.Data["internal_message"];
                }
                LogError(order, ex, "Unhandled exception with message: {0}", msg);

                return OnError(order, msg);
            }
        }

        /// <summary>
        /// Redirects user to payment step or canceled order step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(order, "Redirected to Klarna Checkout");

                switch (Converter.ToString(Context.Current.Request["KlarnaState"]))
                {
                    case "Confirm":
                        return StateConfirm(order);

                    case "Cancel":
                        return StateCancel(order);

                    case "Completed":
                        return StateCancel(order);

                    default:
                        return null;
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return OnError(order, ex.Message);
            }
        }

        private string StateConfirm(Order order)
        {
            LogEvent(order, "State confirm");

            LogEvent(order, "Removing cart from session");
            CartCatch.ClearCart();

            var klarnaOrder = FetchKlarnaOrder();
            if (klarnaOrder.HasError)
            {
                string errorMessage = klarnaOrder.GetErrorMessage();
                LogError(order, $"Order request returned the error with code: '{klarnaOrder.error_code}' and message: '{errorMessage}'");
                return OnError(order, errorMessage);
            }

            SetOrderCompleted(order, klarnaOrder);

            if (UseKlarnaOrderReceipt && !string.IsNullOrWhiteSpace(ConfirmTemplate))
            {
                var confirmTemplate = new Template(ConfirmTemplate);

                confirmTemplate.SetTag("KlarnaCheckout.ConfirmationSnippet", klarnaOrder.html_snippet);

                return Render(order, confirmTemplate);
            }
            else
            {
                RedirectToCart(order);
            }
            return null;
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            var cancelTemplate = new Template(CancelTemplate);
            return Render(order, cancelTemplate);
        }

        private void SetOrderCompleted(Order order, KlarnaOrder klarnaOrder)
        {
            try
            {
                if ((string)klarnaOrder.status == "checkout_complete")
                {
                    var data = new Dictionary<string, object> {
                        { "status", "created" }
                    };

                    var transact = klarnaOrder.order_id;

                    LogEvent(order, "Payment succeeded with external id: {0}", transact);

                    order.TransactionAmount = double.Parse(klarnaOrder.order_amount.ToString()) / 100;
                    order.TransactionStatus = "Succeeded";

                    if (UpdateBillingInfo || UpdateShippingInfo)
                    {
                        if (UpdateBillingInfo)
                        {
                            var billingAddress = klarnaOrder.billing_address;
                            order.CustomerName = string.Format("{0} {1}", billingAddress.given_name, billingAddress.family_name);
                            order.CustomerEmail = billingAddress.email;
                            order.CustomerFirstName = billingAddress.given_name;
                            order.CustomerSurname = billingAddress.family_name;
                            order.CustomerCountryCode = billingAddress.country.ToUpperInvariant();
                            order.CustomerPhone = billingAddress.phone;
                            order.CustomerCity = billingAddress.city;
                            order.CustomerZip = billingAddress.postal_code;
                            order.CustomerAddress = billingAddress.street_address;
                        }

                        if (UpdateShippingInfo)
                        {
                            var shippingAddress = klarnaOrder.shipping_address;
                            order.DeliveryName = string.Format("{0} {1}", shippingAddress.given_name, shippingAddress.family_name);
                            order.DeliveryEmail = shippingAddress.email;
                            order.DeliveryFirstName = shippingAddress.given_name;
                            order.DeliverySurname = shippingAddress.family_name;
                            order.DeliveryCountryCode = shippingAddress.country.ToUpperInvariant();
                            order.DeliveryPhone = shippingAddress.phone;
                            order.DeliveryCity = shippingAddress.city;
                            order.DeliveryZip = shippingAddress.postal_code;
                            order.DeliveryAddress = shippingAddress.street_address;
                        }
                        Services.Orders.Save(order);
                    }

                    SetOrderComplete(order, transact);
                }
                else
                {
                    LogEvent(order, "Klarna order status is incorrect: [0}", (string)klarnaOrder.status);
                    LogError(order, "At least one callback validation error exists - exiting callback routine.");

                    order.TransactionStatus = "Failed";
                    Services.Orders.Save(order);
                }
            }
            finally
            {
                if (order.Complete)
                {
                    CheckoutDone(order);
                }
            }
        }

        private string OnError(Order order, string message)
        {
            order.TransactionAmount = 0;
            order.TransactionStatus = "Failed";
            order.Errors.Add(message);
            Services.Orders.Save(order);

            Services.Orders.DowngradeToCart(order);
            order.CartV2StepIndex = 0;
            order.TransactionStatus = string.Empty;
            Common.Context.SetCart(order);

            if (string.IsNullOrWhiteSpace(ErrorTemplate))
            {
                RedirectToCart(order);
            }

            var errorTemplate = new Template(ErrorTemplate);
            errorTemplate.SetTag("CheckoutHandler:ErrorMessage", message);

            return Render(order, errorTemplate);
        }

        #region IRemoteCapture

        /// <summary>
        /// Shows if capture supported
        /// </summary>
        /// <param name="order">This order object</param>
        /// <returns>'true' if order transaction card type is defined</returns>
        public bool CaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Shows if partial capture of the order supported
        /// </summary>
        /// <param name="order">Instance of order</param>
        /// <returns>True, if partial capture of the order is supported</returns>
        public bool SplitCaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order)
        {
            return Capture(order, order.Price.PricePIP, true);
        }

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <param name="amount">Amount to be captured</param>
        /// <param name="final">Value indicating whether further capturing is prohibited.</param>
        /// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            if (order == null || string.IsNullOrEmpty(order.Id))
            {
                const string errorText = "No valid Order object set";
                LogError(order, errorText);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
            }
            if (string.IsNullOrEmpty(order.TransactionNumber))
            {
                const string errorText = "No transaction number set on the order";
                LogEvent(order, errorText);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, errorText);
            }

            try
            {
                KlarnaResponse response = PostToKlarna(GetAPIURL(),
                    $"ordermanagement/v1/orders/{order.TransactionNumber}/captures", 
                    JsonSerializer.Serialize(new Captures { captured_amount = amount }));

                if (response.HasError)
                {
                    string error = $"Capture request returned the error with code: '{response.error_code}' and message: '{response.GetErrorMessage()}'";
                    LogError(order, error);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, error);
                }

                LogEvent(order, "Capture successful.", DebuggingInfoType.CaptureResult);
                var captureState = !final && (amount + PriceHelper.ConvertToPIP(order.Currency, order.CaptureAmount)) < order.Price.PricePIP ? OrderCaptureInfo.OrderCaptureState.Split : OrderCaptureInfo.OrderCaptureState.Success;
                return new OrderCaptureInfo(captureState, "Capture successful.");
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Remote capture failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, infoTxt);
            }
        }

        #endregion

        #region Order management

        /// <summary>
        /// Cancels order reservation
        /// </summary>
        /// <param name="order">Order that needs to be canceled</param>
        /// <returns></returns>
        public bool CancelOrder(Order order)
        {
            try
            {
                KlarnaResponse response = PostToKlarna(GetAPIURL(), $"ordermanagement/v1/orders/{order.TransactionNumber}/cancel", "");
                if (response.HasError)
                {
                    LogError(order, $"Cancel order request returned the error with code: '{response.error_code}' and message: '{response.GetErrorMessage()}'");
                    return false;
                }

                LogEvent(order, "Cancel order succeed.");
                return true;
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Cancel order failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
            }

            return false;
        }

        /// <summary>
        /// Refunds order
        /// </summary>
        /// <param name="order">Modified order</param>
        /// <param name="originalOrder">Original order</param>
        public void AmountRefund(Order order, Order originalOrder)
        {
            try
            {
                KlarnaResponse response = PostToKlarna(GetAPIURL(), 
                    $"ordermanagement/v1/orders/{originalOrder.TransactionNumber}/refunds", 
                    JsonSerializer.Serialize(new Refunds { refunded_amount = order.Price.PricePIP }));
                
                if (!response.HasError)
                    LogEvent(originalOrder, "Amount refund succeed");
                else
                    LogError(originalOrder, $"Refund request returned the error with code: '{response.error_code}' and message: '{response.GetErrorMessage()}'");
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Amount refund failed with the message: {0}", ex.Message);
                LogError(originalOrder, ex, infoTxt);
            }
        }

        /// <summary>
        /// Performs partial return of order
        /// </summary>
        /// <param name="order">Modified order</param>
        /// <param name="originalOrder">Original order</param>
        public void PartialReturn(Order order, Order originalOrder)
        {
            try
            {
                KlarnaResponse response = PostToKlarna(GetAPIURL(),
                    $"ordermanagement/v1/orders/{originalOrder.TransactionNumber}/refunds", 
                    JsonSerializer.Serialize(new Refunds { refunded_amount = order.Price.PricePIP }));

                if (!response.HasError)
                {
                    OrderReturnInfo.SaveReturnOperation(OrderReturnOperationState.PartiallyReturned, "Klarna has partial refunded payment.", order.Price.Price, originalOrder);
                    LogEvent(originalOrder, "Partial return succeed");
                }
                else
                {
                    LogError(originalOrder, $"Partial return request returned the error with code: '{response.error_code}' and message: '{response.GetErrorMessage()}'");
                }
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Partial return failed with the message: {0}", ex.Message);
                LogError(originalOrder, ex, infoTxt);
            }
        }

        /// <summary>
        /// Performs full return of order
        /// </summary>
        /// <param name="order">Order that needs to be return</param>
        public void FullReturn(Order order)
        {
            try
            {
                KlarnaResponse response = PostToKlarna(GetAPIURL(), 
                    $"ordermanagement/v1/orders/{order.TransactionNumber}/refunds", 
                    JsonSerializer.Serialize(new Refunds { refunded_amount = PriceHelper.ConvertToPIP(order.Currency, order.CaptureAmount) }));

                if (!response.HasError)
                {
                    OrderReturnInfo.SaveReturnOperation(OrderReturnOperationState.FullyReturned, "Klarna has full refunded payment.", order.Price.Price, order);
                    LogEvent(order, "Full return succeed");
                }
                else
                {
                    LogError(order, $"Full return request returned the error with code: '{response.error_code}' and message: '{response.GetErrorMessage()}'");
                }
            }
            catch (Exception ex)
            {
                var infoTxt = string.Format("Full return failed with the message: {0}", ex.Message);
                LogError(order, ex, infoTxt);
            }
        }

        #endregion

        #region Private members

        private KlarnaOrder FetchKlarnaOrder()
        {
            return GetFromKlarna<KlarnaOrder>($"{Context.Current.Request.QueryString["klarna_order"]}", "");
        }

        private KlarnaOrder CreateKlarnaOrder(Order order)
        {            
            KlarnaOrder klarnaOrder = new KlarnaOrder
            {
                merchant_urls = new MerchantUrls
                {
                    terms = FullUrlToTermsAndConditions(),
                    checkout = GetBaseUrl(order),
                    confirmation = ConfirmUrl(order),
                    push = CallbackUrl(order)
                },
                options = new Options(),
                merchant_reference1 = order.Id,
                purchase_country = order.CustomerCountryCode,
                purchase_currency = order.CurrencyCode,
                locale = GetLocale(order.CustomerCountryCode),
                order_amount = order.Price.PricePIP,
                order_tax_amount = PriceHelper.ConvertToPIP(order.Currency, order.Price.VAT),
                order_lines = GetOrderLines(order)
            };

            if (!string.IsNullOrEmpty(order.ShippingMethodDescription))
            {
                klarnaOrder.options.shipping_details = order.ShippingMethodDescription.Length > 70 ? order.ShippingMethodDescription.Substring(0, 70) : order.ShippingMethodDescription;
            }

            if (AllowSeparateShippingAddresses)
            {
                klarnaOrder.options.allow_separate_shipping_address = AllowSeparateShippingAddresses;
            }

            if (AllowSwitchCustomerTypes)
            {
                klarnaOrder.options.allowed_customer_types.Add("person");
                klarnaOrder.options.allowed_customer_types.Add("organization");
            }

            if (DefaultFlowIsB2B)
            {
                klarnaOrder.customer = new Customer { type = "organization" };
            }

            return PostToKlarna<KlarnaOrder>(GetAPIURL(), "checkout/v3/orders", JsonSerializer.Serialize(klarnaOrder));
        }

        private string GetAPIURL()
        {
            return TestMode ? "https://api.playground.klarna.com/" : "https://api.klarna.com/";
        }
       
        private KlarnaResponse PostToKlarna(string url, string method, string jsonObject) => PostToKlarna<KlarnaResponse>(url, method, jsonObject);

        private T PostToKlarna<T>(string url, string method, string jsonObject) where T: KlarnaResponse
        {
            using (var client = new HttpClient())
            {
                string authenticationScheme = "Basic";
                string authenticationParameter = Convert.ToBase64String(Encoding.Default.GetBytes($"{ApiUID}:{ApiPassword}"));
                client.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue(authenticationScheme, authenticationParameter);
                var content = new StringContent(jsonObject);
                content.Headers.ContentType = MediaTypeHeaderValue.Parse("application/json");
                using (var response = client.PostAsync(url + method, content).GetAwaiter().GetResult())
                {
                    string responseText = response.Content.ReadAsStringAsync().GetAwaiter().GetResult();
                    var responseObject = string.IsNullOrEmpty(responseText) ?
                        Activator.CreateInstance<T>() :
                        JsonSerializer.Deserialize<T>(responseText);

                    if (!response.IsSuccessStatusCode && !responseObject.HasError)
                    {
                        responseObject.error_code = response.StatusCode.ToString();
                        responseObject.error_message = response.ReasonPhrase;
                    }                   

                    return responseObject;
                }
            }
        }

        private T GetFromKlarna<T>(string url, string method) where T : KlarnaResponse
        {
            using (var client = new HttpClient())
            {
                string authenticationScheme = "Basic";
                string authenticationParameter = Convert.ToBase64String(Encoding.Default.GetBytes($"{ApiUID}:{ApiPassword}"));
                client.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue(authenticationScheme, authenticationParameter);
                var content = new StringContent("");
                content.Headers.ContentType = MediaTypeHeaderValue.Parse("application/json");
                using (var response = client.GetAsync(url + method).GetAwaiter().GetResult())
                {
                    string responseText = response.Content.ReadAsStringAsync().GetAwaiter().GetResult();
                    return JsonSerializer.Deserialize<T>(responseText);
                }
            }
        }

        private static List<Model.OrderLine> GetOrderLines(Order order)
        {
            List<Model.OrderLine> result = new List<Model.OrderLine>();

            foreach (var orderLine in order.OrderLines)
            {              
                result.Add(new Model.OrderLine()
                {
                    reference = orderLine.ProductNumber,
                    name = !string.IsNullOrEmpty(orderLine.ProductName) ? orderLine.ProductName : orderLine.Id,
                    quantity = (int)orderLine.Quantity,
                    unit_price = PriceHelper.ConvertToPIP(order.Currency, orderLine.UnitPrice.PriceWithVAT),
                    tax_rate = PriceHelper.ConvertToPIP(order.Currency, orderLine.Price.VATPercent),
                    total_amount = PriceHelper.ConvertToPIP(order.Currency, orderLine.Price.PriceWithVAT),
                    total_tax_amount = PriceHelper.ConvertToPIP(order.Currency, orderLine.Price.VAT)
                });
            }

            var shipping = order.ShippingFee;
            if (shipping.Price > 0)
            {
                result.Add(new Model.OrderLine()
                {
                    type = "shipping_fee",
                    reference = order.ShippingMethodId,
                    name = order.ShippingMethod,
                    quantity = 1,
                    unit_price = PriceHelper.ConvertToPIP(order.Currency, shipping.PriceWithVAT),
                    tax_rate = PriceHelper.ConvertToPIP(order.Currency, shipping.VATPercent),
                    total_amount = PriceHelper.ConvertToPIP(order.Currency, shipping.PriceWithVAT),
                    total_tax_amount = PriceHelper.ConvertToPIP(order.Currency, shipping.VAT)
                });
            }

            var payment = order.PaymentFee;
            if (payment.Price > 0)
            {
                result.Add(new Model.OrderLine()
                {
                    reference = order.PaymentMethodId,
                    name = order.PaymentMethod,
                    quantity = 1,
                    unit_price = PriceHelper.ConvertToPIP(order.Currency, payment.PriceWithVAT),
                    tax_rate = PriceHelper.ConvertToPIP(order.Currency, payment.VATPercent),
                    total_amount = PriceHelper.ConvertToPIP(order.Currency, payment.PriceWithVAT),
                    total_tax_amount = PriceHelper.ConvertToPIP(order.Currency, payment.VAT)
                });
            }
            
            return result;
        }

        private string GetLocale(string countryCode)
        {
            string locale = null;
            if (countryCode == "SE")
            {
                locale = "sv-se";
            }
            else if (countryCode == "NO")
            {
                locale = "nb-no";
            }
            else if (countryCode == "DE")
            {
                locale = "de-de";
            }
            else if (countryCode == "AT")
            {
                locale = "de-at";
            }
            else if (countryCode == "FI")
            {
                locale = Environment.ExecutingContext.GetCulture(true).Name.ToLowerInvariant() == "fi-fi" ? "fi-fi" : "sv-fi";
            }

            return locale;
        }

        #endregion
    }
}
