﻿using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.PayPalExpressCheckout
{
    /// <summary>
    /// PayPal Express Checkout Handler
    /// </summary>
    [
        AddInName("PayPal Express Checkout"),
        AddInDescription("PayPal Express Checkout Handler")
    ]
    public class PayPalExpressCheckout : CheckoutHandler, IRemoteCapture, IDropDownOptions
    {
        #region Constants

        private const string ApiVersion = "97";
        private const string ResponseTypeRequestName = "CheckoutHandlerResponseType";
        private const string TrackingCode = "DynamicwebAS_SP_EC";

        #endregion

        #region Public Properties

        [AddInParameter("Username")]
        [AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Username { get; set; }

        [AddInParameter("Password")]
        [AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Password { get; set; }

        [AddInParameter("Signature")]
        [AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Signature { get; set; }

        [AddInParameter("Email ID")]
        [AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Email { get; set; }

        [AddInParameter("Cancel template")]
        [AddInParameterEditor(typeof(TemplateParameterEditor),
            "folder=templates/eCom7/CheckoutHandler/PayPal/ExpressCheckout/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template")]
        [AddInParameterEditor(typeof(TemplateParameterEditor),
            "folder=templates/eCom7/CheckoutHandler/PayPal/ExpressCheckout/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Enable guest checkout")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool EnableGuestCheckout { get; set; }

        [AddInParameter("Use PayPal shipping info")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool UsePayPalShippingInfo { get; set; }

        [AddInParameter("Don't override PayPal shipping address")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool DontOverrideShipping { get; set; }

        [AddInParameter("No shipping")]
        [AddInParameterEditor(typeof(RadioParameterEditor), "SortBy=Key;")]
        public string NoShipping { get; set; }

        [AddInParameter("Payment type")]
        [AddInParameterEditor(typeof(RadioParameterEditor), "")]
        public string PaymentType { get; set; }

        [AddInParameter("Test")]
        [AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool Test { get; set; }

        #endregion

        #region Properties

        /// <summary>
        /// Url for API calls.
        /// </summary>
        private string ApiUrl
        {
            get { return Test ? "sandbox.paypal.com" : "paypal.com"; }
        }

        /// <summary>
        /// Url for rederecting and IPN responding.
        /// </summary>
        private string Url
        {
            get { return string.Format("https://www.{0}/cgi-bin/webscr", ApiUrl); }
        }

        /// <summary>
        /// Returns the callback status.
        /// </summary>
        private PayPalExpressCheckoutResponseType ResponseType
        {
            get
            {
                PayPalExpressCheckoutResponseType result;
                Enum.TryParse(Dynamicweb.Context.Current.Request[ResponseTypeRequestName], true, out result);

                return result;
            }
        }

        /// <summary>
        /// Returns the current language code.
        /// </summary>
        private string LanguageCode
        {
            get
            {
                var result = "GB";
                var supportedLanguageCodes = new List<string>(
                    new[] {
                                "AU", "AI", "BE", "BR", "CA", "CH", "CN", "DE", "ES", "GB", "FR", "IT", "NL", "PL", "PT",
                                "RU", "da_DK", "he_IL", "id_ID", "jp_JP", "no_NO", "pt_BR", "ru_RU", "sv_SE", "th_TH",
                                "tr_TR", "zh_CN", "zh_HK", "zh_TW"
                    }
                );

                var cultureInfo = Dynamicweb.Environment.ExecutingContext.GetCulture(true);
                var currentRegionCode = new RegionInfo(cultureInfo.LCID);
                var code1 = currentRegionCode.TwoLetterISORegionName;
                var code2 = string.Format("{0}_{1}", cultureInfo.TwoLetterISOLanguageName, currentRegionCode.TwoLetterISORegionName);

                if (supportedLanguageCodes.Contains(code1))
                {
                    result = code1;
                }
                else if (supportedLanguageCodes.Contains(code2))
                {
                    result = code2;
                }

                return result;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public PayPalExpressCheckout()
        {
            PaymentType = "Sale";
            NoShipping = "0";
            ServicePointManager.SecurityProtocol = ServicePointManager.SecurityProtocol | SecurityProtocolType.Tls12;
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Starts checkout.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <returns>Rendered template.</returns>
        public override string StartCheckout(Order order)
        {
            var result = string.Empty;

            try
            {
                var response = SetExpressCheckout(order);

                if (response.Success)
                {
                    Context.Current.Response.Redirect(string.Format("{0}?cmd=_express-checkout&token={1}", Url, response.Data["TOKEN"]));
                }
                else
                {
                    result = OnError(order, response.Message);
                    LogEvent(order, GetLogMessage(response));
                }
            }
            catch (ThreadAbortException)
            {
                result = string.Empty;
            }
            catch (Exception ex)
            {
                result = OnError(order, "Internal error.");
                LogEvent(order, ex.Message);
            }

            return result;
        }

        /// <summary>
		/// Starts order checkout procedure
        /// </summary>
		/// <param name="order">Order to be checked out</param>
		/// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            string result;

            try
            {
                switch (ResponseType)
                {
                    case PayPalExpressCheckoutResponseType.Accepted:
                        LogEvent(order, "State accepted");
                        result = OnAccept(order, Dynamicweb.Context.Current.Request["token"]);
                        break;

                    case PayPalExpressCheckoutResponseType.Cancelled:
                        LogEvent(order, "State cancelled");
                        result = OnCancel(order);
                        break;

                    case PayPalExpressCheckoutResponseType.Notification:
                        LogEvent(order, "State notification");
                        result = OnNotification(order);
                        break;

                    default:
                        LogEvent(order, "State error");
                        result = OnError(order, "Invalid request!");
                        break;
                }
            }
            catch (ThreadAbortException)
            {
                result = string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "State failed");
                result = OnError(order, ex.Message);
            }

            return result;
        }

        /// <summary>
		/// Send capture request to transaction service
        /// </summary>
		/// <param name="order">Order to be captured</param>
		/// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order)
        {
            OrderCaptureInfo result;

            try
            {
                result = CaptureInternal(order);
            }
            catch (Exception e)
            {
                result = new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, e.Message);
            }

            return result;
        }

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="behaviorMode"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string behaviorMode)
        {
            var result = new Hashtable();
            switch (behaviorMode)
            {
                case "Payment type":
                    result.Add("Sale", "Sale");
                    result.Add("Authorization", "Authorization");
                    break;
                case "No shipping":
                    result.Add("0", " PayPal displays the shipping address on the PayPal pages");
                    result.Add("1", "PayPal does not display shipping address fields and removes shipping information from the transaction");
                    result.Add("2", "If you do not pass the shipping address, PayPal obtains it from the buyer's account profile");
                    break;

                default:
                    break;
            }

            return result;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Sets up an Express Checkout transaction.
        /// </summary>
        /// <param name="order">
        /// The current order.
        /// </param>
        /// <returns>
        /// The result of operation <see cref="bool"/>.
        /// </returns>
        private PayPalActionResult SetExpressCheckout(Order order)
        {
            var returnUrl = GetBaseUrl(order);
            var parameters = new NameValueCollection();

            parameters.Add("METHOD", "SetExpressCheckout");
            parameters.Add("USER", Username);
            parameters.Add("PWD", Password);
            parameters.Add("EMAIL", order.CustomerEmail);
            parameters.Add("SIGNATURE", Signature);
            parameters.Add("BUTTONSOURCE", TrackingCode);
            parameters.Add("VERSION", ApiVersion);
            parameters.Add("RETURNURL", string.Format("{0}&{1}={2}", returnUrl, ResponseTypeRequestName, PayPalExpressCheckoutResponseType.Accepted));
            parameters.Add("CANCELURL", string.Format("{0}&{1}={2}", returnUrl, ResponseTypeRequestName, PayPalExpressCheckoutResponseType.Cancelled));
            parameters.Add("REQCONFIRMSHIPPING", UsePayPalShippingInfo ? "1" : "0");
            parameters.Add("NOSHIPPING", NoShipping);
            parameters.Add("LOCALECODE", LanguageCode);
            parameters.Add("SOLUTIONTYPE", EnableGuestCheckout ? "Sole" : "Mark");
            parameters.Add("LANDINGPAGE", EnableGuestCheckout ? "Billing" : "Login");

            if (!UsePayPalShippingInfo)
            {
                parameters.Add("PAYMENTREQUEST_0_SHIPTONAME", order.DeliveryName);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOSTREET", order.DeliveryAddress);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOSTREET2", order.DeliveryAddress2);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOCITY", order.DeliveryCity);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOSTATE", order.DeliveryRegion);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOZIP", order.DeliveryZip);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOCOUNTRY", order.DeliveryCountryCode);
                parameters.Add("PAYMENTREQUEST_0_SHIPTOPHONENUM", order.DeliveryPhone);
                parameters.Add("ADDROVERRIDE", DontOverrideShipping ? "0" : "1");
            }

            parameters.Add("PAYMENTREQUEST_0_PAYMENTACTION", PaymentType);
            parameters.Add("PAYMENTREQUEST_0_CUSTOM", PaymentType);
            parameters.Add("PAYMENTREQUEST_0_AMT", FormatPrice(order.Price.PriceWithVAT));
            parameters.Add("PAYMENTREQUEST_0_CURRENCYCODE", order.CurrencyCode);
            parameters.Add("PAYMENTREQUEST_0_INVNUM", order.Id);

            var response = ParseResponse(DoRequest(parameters));

            return new PayPalActionResult(response);
        }

        /// <summary>
        /// Obtains details about an Express Checkout transaction.
        /// </summary>
        /// <param name="token">
        /// The token.
        /// </param>
        /// <returns>
        /// The <see cref="PayPalActionResult"/>.
        /// </returns>
        private PayPalActionResult GetExpressCheckoutDetails(string token)
        {
            var parameters = new NameValueCollection();
            parameters.Add("METHOD", "GetExpressCheckoutDetails");
            parameters.Add("USER", Username);
            parameters.Add("PWD", Password);
            parameters.Add("SIGNATURE", Signature);
            parameters.Add("BUTTONSOURCE", TrackingCode);
            parameters.Add("VERSION", ApiVersion);
            parameters.Add("TOKEN", token);

            var response = ParseResponse(DoRequest(parameters));

            return new PayPalActionResult(response);
        }

        /// <summary>
        /// Completes an Express Checkout transaction.
        /// </summary>
        /// <param name="order">
        /// The current order.
        /// </param>
        /// <param name="details">
        /// Transaction details.
        /// </param>
        /// <returns>
        /// The <see cref="PayPalActionResult"/>.
        /// </returns>
        private PayPalActionResult DoExpressCheckoutPayment(Order order, NameValueCollection details)
        {
            var parameters = new NameValueCollection(details);
            parameters.Add("METHOD", "DoExpressCheckoutPayment");
            parameters.Add("USER", Username);
            parameters.Add("PWD", Password);
            parameters.Add("SIGNATURE", Signature);
            parameters.Add("BUTTONSOURCE", TrackingCode);
            parameters.Add("PAYMENTREQUEST_0_PAYMENTACTION", string.IsNullOrWhiteSpace(details["CUSTOM"]) ? PaymentType : details["CUSTOM"]);
            parameters.Add("PAYMENTREQUEST_0_NOTIFYURL", string.Format("{0}&{1}={2}", GetBaseUrl(order), ResponseTypeRequestName, PayPalExpressCheckoutResponseType.Notification));

            var response = ParseResponse(DoRequest(parameters));

            return new PayPalActionResult(response);
        }

        /// <summary>
        /// Captures an authorized payment.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <returns>PayPalActionResult</returns>
        private PayPalActionResult DoCapture(Order order)
        {
            var parameters = new NameValueCollection();
            parameters.Add("METHOD", "DoCapture");
            parameters.Add("USER", Username);
            parameters.Add("PWD", Password);
            parameters.Add("SIGNATURE", Signature);
            parameters.Add("VERSION", ApiVersion);
            parameters.Add("AUTHORIZATIONID", order.TransactionNumber);
            parameters.Add("AMT", FormatPrice(order.Price.Price));
            parameters.Add("CURRENCYCODE", order.CurrencyCode);
            parameters.Add("INVNUM", order.Id);
            parameters.Add("COMPLETETYPE", "Complete");

            var response = ParseResponse(DoRequest(parameters));

            return new PayPalActionResult(response);
        }

        private OrderCaptureInfo CaptureInternal(Order order)
        {
            if (order == null || string.IsNullOrEmpty(order.Id))
            {
                throw new ArgumentNullException("No valid Order object set");
            }

            if (string.IsNullOrEmpty(order.TransactionNumber))
            {
                throw new ArgumentNullException("No transaction number set on the order");
            }

            OrderCaptureInfo.OrderCaptureState state;
            var captureResult = DoCapture(order);

            switch (captureResult.PaymentStatus)
            {
                case PayPalExpressCheckoutPaymentStatus.Authorized:
                    state = OrderCaptureInfo.OrderCaptureState.NotCaptured;
                    break;

                case PayPalExpressCheckoutPaymentStatus.Paid:
                    state = OrderCaptureInfo.OrderCaptureState.Success;
                    break;

                case PayPalExpressCheckoutPaymentStatus.Pending:
                    state = OrderCaptureInfo.OrderCaptureState.NotCaptured;
                    break;

                case PayPalExpressCheckoutPaymentStatus.Voided:
                    state = OrderCaptureInfo.OrderCaptureState.NotCaptured;
                    break;

                default:
                    state = OrderCaptureInfo.OrderCaptureState.Failed;
                    break;
            }

            return new OrderCaptureInfo(state, captureResult.Message);
        }

        private string DoRequest(NameValueCollection parameters)
        {
            return DoRequest(parameters, new Uri(string.Format("https://api-3t.{0}/nvp", ApiUrl)));
        }

        private string DoRequest(NameValueCollection parameters, Uri customUrl)
        {
            if (parameters == null || parameters.Count == 0)
            {
                throw new ArgumentException("Parameters collection is empty!");
            }

            if (customUrl == null || customUrl.AbsolutePath.Length == 0)
            {
                throw new ArgumentException("Invalid url!");
            }

            string result;

            using (var client = new WebClient())
            {
                client.Headers[HttpRequestHeader.ContentType] = "application/x-www-form-urlencoded";
                client.Encoding = Encoding.UTF8;
                byte[] response = client.UploadValues(customUrl, parameters);
                result = Encoding.UTF8.GetString(response);
            }

            return result;
        }

        /// <summary>
        /// Parses the response data.
        /// </summary>
        /// <param name="str">Response string.</param>
        /// <returns>Dictionary</returns>
        private NameValueCollection ParseResponse(string str)
        {
            var result = new NameValueCollection();
            foreach (var el in str.Split(Convert.ToChar("&")))
            {
                var pair = el.Split(Convert.ToChar("="));
                result.Add(
                    WebUtility.UrlDecode(pair[0].Trim()),
                    pair.Length > 1 ? WebUtility.UrlDecode(pair[1].Trim()) : WebUtility.UrlDecode(pair[0].Trim())
                );
            }

            return result;
        }

        /// <summary>
        /// On client accept event handler.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <param name="token">The transaction token.</param>
        /// <returns>Template output.</returns>
        private string OnAccept(Order order, string token)
        {
            var detailsResult = GetExpressCheckoutDetails(token);

            if (!detailsResult.Success)
            {
                throw new Exception(detailsResult.Message);
            }

            var checkoutResult = DoExpressCheckoutPayment(order, detailsResult.Data);

            if (!checkoutResult.Success)
            {
                throw new Exception(checkoutResult.Message);
            }

            if (checkoutResult.PaymentStatus != PayPalExpressCheckoutPaymentStatus.Paid &&
                checkoutResult.PaymentStatus != PayPalExpressCheckoutPaymentStatus.Authorized &&
                checkoutResult.PaymentStatus != PayPalExpressCheckoutPaymentStatus.Pending)
            {
                throw new Exception(string.Format("Unexpected payment status: {0}", checkoutResult.PaymentStatus));
            }

            order.TransactionNumber = checkoutResult.TransactionId;

            if (UsePayPalShippingInfo)
            {
                UpdateShippingInfo(order, detailsResult.Data);
            }

            if (!order.Complete)
            {
                SetOrderComplete(order);
            }

            CheckoutDone(order);

            LogEvent(order, GetLogMessage(checkoutResult));

            RedirectToCart(order);

            return null;
        }

        /// <summary>
        /// On client cancel event handler.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <returns>Template output.</returns>
        private string OnCancel(Order order)
        {
            order.TransactionStatus = "Cancelled";
            Services.Orders.Save(order);

            CheckoutDone(order);

            if (string.IsNullOrWhiteSpace(CancelTemplate))
            {
                RedirectToCart(order);
            }

            return Render(order, new Template(CancelTemplate));
        }

        /// <summary>
        /// On error event handler.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <param name="msg">The error message.</param>
        /// <returns>Template output.</returns>
        private string OnError(Order order, string msg)
        {
            order.TransactionAmount = 0;
            order.TransactionStatus = "Failed";
            order.Errors.Add(msg);
            Services.Orders.Save(order);

            Services.Orders.DowngradeToCart(order);
            order.CartV2StepIndex = 0;
            order.TransactionStatus = "";
            Common.Context.SetCart(order);

            if (string.IsNullOrWhiteSpace(ErrorTemplate))
            {
                RedirectToCart(order);
            }

            return Render(order, new Template(ErrorTemplate));
        }

        private string OnNotification(Order order)
        {
            if (order.Complete && !string.IsNullOrWhiteSpace(order.TransactionNumber))
            {
                LogEvent(order, "Order is completed already");
                return string.Empty;
            }

            if (string.Compare(Dynamicweb.Context.Current.Request["receiver_email"], Email, StringComparison.InvariantCultureIgnoreCase) != 0)
            {
                LogEvent(order, string.Format("Receiver email is wrong: {0}", Dynamicweb.Context.Current.Request["receiver_email"]));
                return string.Empty;
            }

            var ipnResponseData = new NameValueCollection();
            ipnResponseData.Add("cmd", "_notify-validate");
            ipnResponseData.Add(Context.Current.Request.Form);
            ipnResponseData.Add(ParseResponse(DoRequest(ipnResponseData, new Uri(Url))));

            var ipnVerification = new PayPalActionResult(ipnResponseData);

            string transactionType = ipnVerification.Data["txn_type"];
            string transactionId = ipnVerification.Data["txn_id"];

            if (ipnVerification.Success)
            {
                this.LogEvent(order, string.Format("Verification is succeed. Payment status: {0}", ipnVerification.PaymentStatus));

                switch (ipnVerification.PaymentStatus)
                {
                    case PayPalExpressCheckoutPaymentStatus.Authorized:
                        MarkAsSucceeded(order, transactionType, transactionId);
                        LogEvent(order, "Payment authorized.");
                        break;

                    case PayPalExpressCheckoutPaymentStatus.Paid:
                        order.CaptureInfo.State = OrderCaptureInfo.OrderCaptureState.Success;
                        MarkAsSucceeded(order, transactionType, transactionId);
                        LogEvent(order, "Payment succeeded.");
                        break;

                    case PayPalExpressCheckoutPaymentStatus.Voided:
                        MarkAsDeclined(order);
                        LogEvent(order, "Payment declined.");
                        break;

                    default:
                        break;
                }
            }
            else
            {
                LogEvent(order, "Verification is failed.");
            }

            return string.Empty;
        }

        private void MarkAsSucceeded(Order order, string transactionType, string transactionId)
        {
            order.TransactionStatus = "Succeeded";
            order.TransactionAmount = order.Price.Price;
            order.TransactionNumber = transactionId;
            order.TransactionType = transactionType;

            SetOrderComplete(order);
        }

        private void MarkAsDeclined(Order order)
        {
            order.TransactionStatus = "Decline";
            order.TransactionAmount = 0;
        }

        /// <summary>
        /// Updates the shipping information in order on accept and notification events.
        /// </summary>
        /// <param name="order">The current order.</param>
        /// <param name="data">The data from "DoEpxressCheckout" action.</param>
        private void UpdateShippingInfo(Order order, NameValueCollection data)
        {
            order.DeliveryName = data["PAYMENTREQUEST_0_SHIPTONAME"];
            order.DeliveryAddress = data["PAYMENTREQUEST_0_SHIPTOSTREET"];
            order.DeliveryAddress2 = data["PAYMENTREQUEST_0_SHIPTOSTREET2"];
            order.DeliveryCity = data["PAYMENTREQUEST_0_SHIPTOCITY"];
            order.DeliveryRegion = data["PAYMENTREQUEST_0_SHIPTOSTATE"];
            order.DeliveryZip = data["PAYMENTREQUEST_0_SHIPTOZIP"];
            order.DeliveryCountryCode = data["PAYMENTREQUEST_0_SHIPTOCOUNTRYCODE"];
            order.DeliveryPhone = data["PAYMENTREQUEST_0_SHIPTOPHONENUM"];
        }

        private string FormatPrice(double price)
        {
            return string.Format(CultureInfo.InvariantCulture, "{0:#,##0.00}", price);
        }

        private string GetLogMessage(PayPalActionResult result)
        {
            return string.Format("Correlation id: {0}. {1}", result.CorrelationId, result.Message);
        }

        #endregion

        #region Nested Types

        private class PayPalActionResult
        {
            #region Public Properties

            public NameValueCollection Data { get; private set; }

            public string Message { get; private set; }

            public bool Success { get; private set; }

            public PayPalExpressCheckoutPaymentStatus PaymentStatus { get; private set; }

            public string CorrelationId { get; private set; }

            public string TransactionId { get; private set; }

            #endregion

            #region Constructors and Destructors

            public PayPalActionResult(NameValueCollection response)
            {
                Data = response;
                Success = true;
                Message = string.Empty;
                PaymentStatus = PayPalExpressCheckoutPaymentStatus.Undefined;

                if (response.AllKeys.Contains("ACK"))
                {
                    Success = string.Compare(response["ACK"].Trim(), "Success", StringComparison.InvariantCultureIgnoreCase) == 0;
                }
                else
                {
                    Success = response.AllKeys.Contains("VERIFIED");
                }

                if (response.AllKeys.Contains("PAYMENTINFO_0_PAYMENTSTATUS"))
                {
                    PaymentStatus = GetPaymentStatus(response["PAYMENTINFO_0_PAYMENTSTATUS"], response["PAYMENTINFO_0_PENDINGREASON"]);
                }
                else if (response.AllKeys.Contains("PAYMENTSTATUS"))
                {
                    PaymentStatus = GetPaymentStatus(response["PAYMENTSTATUS"], response["PENDINGREASON"]);
                }
                else if (response.AllKeys.Contains("payment_status"))
                {
                    PaymentStatus = GetPaymentStatus(response["payment_status"], response["pending_reason"]);
                }

                if (response.AllKeys.Contains("CORRELATIONID"))
                {
                    CorrelationId = response["CORRELATIONID"];
                }

                if (response.AllKeys.Contains("PAYMENTINFO_0_TRANSACTIONID"))
                {
                    TransactionId = response["PAYMENTINFO_0_TRANSACTIONID"];
                }

                var stringBuilder = new StringBuilder();

                if (!Success)
                {
                    var errCount = response.AllKeys.Count(k => k.StartsWith("L_ERRORCODE"));

                    for (int i = 0; i < errCount; i++)
                    {
                        var numStr = i.ToString(CultureInfo.InvariantCulture);
                        stringBuilder.AppendLine(
                            string.Format(
                                "{0}. {1}. {2}.",
                                response[string.Format("L_ERRORCODE{0}", numStr)],
                                response[string.Format("L_SHORTMESSAGE{0}", numStr)],
                                response[string.Format("L_LONGMESSAGE{0}", numStr)]
                            )
                        );
                    }

                    if (stringBuilder.Length == 0)
                    {
                        stringBuilder.AppendLine("Uknown error!");
                    }
                }

                Message = stringBuilder.ToString();
            }

            #endregion

            #region Methods

            /// <summary>
            /// Gets a payment status
            /// </summary>
            /// <param name="paymentStatus">PayPal payment status</param>
            /// <param name="pendingReason">PayPal pending reason</param>
            /// <returns>Payment status</returns>
            private PayPalExpressCheckoutPaymentStatus GetPaymentStatus(string paymentStatus, string pendingReason)
            {
                PayPalExpressCheckoutPaymentStatus result;

                if (paymentStatus == null)
                {
                    paymentStatus = string.Empty;
                }

                if (pendingReason == null)
                {
                    pendingReason = string.Empty;
                }

                switch (paymentStatus.ToLowerInvariant())
                {
                    case "pending":
                        switch (pendingReason.ToLowerInvariant())
                        {
                            case "authorization":
                                result = PayPalExpressCheckoutPaymentStatus.Authorized;
                                break;
                            default:
                                result = PayPalExpressCheckoutPaymentStatus.Pending;
                                break;
                        }
                        break;
                    case "processed":
                    case "completed":
                    case "canceled_reversal":
                        result = PayPalExpressCheckoutPaymentStatus.Paid;
                        break;
                    case "denied":
                    case "expired":
                    case "failed":
                    case "voided":
                        result = PayPalExpressCheckoutPaymentStatus.Voided;
                        break;
                    case "refunded":
                    case "reversed":
                        result = PayPalExpressCheckoutPaymentStatus.Refunded;
                        break;
                    default:
                        result = PayPalExpressCheckoutPaymentStatus.Undefined;
                        break;
                }

                return result;
            }

            #endregion
        }

        private enum PayPalExpressCheckoutResponseType
        {
            Undefined,
            Accepted,
            Cancelled,
            Callback,
            Notification
        }

        private enum PayPalExpressCheckoutPaymentStatus
        {
            Undefined,
            Pending,
            Authorized,
            Paid,
            PartiallyRefunded,
            Refunded,
            Voided,
        }

        #endregion
    }
}