﻿using Dynamicweb.Caching;
using Dynamicweb.Core;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.Gateways;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Extensibility.Editors;
using Dynamicweb.Rendering;
using Dynamicweb.SystemTools;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Xml;

namespace Dynamicweb.Ecommerce.CheckoutHandlers.QuickPay3
{
    /// <summary>
    /// QuickPay Checkout Handler
    /// </summary>
    [
        AddInName("QuickPay"),
        AddInDescription("Checkout handler for QuickPay"),
        AddInDeprecated(true)
    ]
    public class QuickPay3 : CheckoutHandler, IDropDownOptions, IRemotePartialCapture
    {
        private enum PostModes { Auto, Template }

        private PostModes postMode = PostModes.Auto;
        private static object lockObject = new object();
        private const string orderCacheKey = "CheckoutHandler:Order.";

        #region Addin parameters

        [AddInParameter("Merchant"), AddInParameterEditor(typeof(TextParameterEditor), "")]
        public string Merchant { get; set; }

        [AddInParameter("Secret key"), AddInParameterEditor(typeof(TextParameterEditor), "size=80")]
        public string SecretKey { get; set; }

        [AddInParameter("Card type"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=true; nonetext=All; SortBy=Value;")]
        public string CardTypeLock { get; set; }

        [AddInParameter("Post mode"), AddInParameterEditor(typeof(DropDownParameterEditor), "NewGUI=true; none=false; SortBy=Value;")]
        public string PostModeSelection
        {
            get
            {
                return postMode.ToString();
            }
            set
            {
                switch (value)
                {
                    case "Auto":
                        postMode = PostModes.Auto;
                        break;
                    case "Template":
                        postMode = PostModes.Template;
                        break;
                }
            }
        }

        [AddInParameter("Post template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/QuickPay 3/Post")]
        public string PostTemplate { get; set; }

        [AddInParameter("Cancel template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/QuickPay 3/Cancel")]
        public string CancelTemplate { get; set; }

        [AddInParameter("Error template"), AddInParameterEditor(typeof(TemplateParameterEditor), "folder=templates/eCom7/CheckoutHandler/QuickPay 3/Error")]
        public string ErrorTemplate { get; set; }

        [AddInParameter("Test mode"), AddInParameterEditor(typeof(YesNoParameterEditor), "")]
        public bool TestMode { get; set; }

        #endregion

        #region Form properties

        private static string Protocol { get { return "7"; } }
        private static string AuthorizeMsgType { get { return "authorize"; } }
        private static string CaptureMsgType { get { return "capture"; } }
        private static string LanguageCode
        {
            get
            {
                var currentLanguageCode = Dynamicweb.Environment.ExecutingContext.GetCulture(true).TwoLetterISOLanguageName;
                var supportedLanguageCodes = new List<string>(new[] { "da", "de", "en", "fo", "fr", "it", "no", "nl", "pl", "se" });
                return supportedLanguageCodes.Contains(currentLanguageCode) ? currentLanguageCode : "en";
            }
        }

        private static string BaseUrl(Order order)
        {
            var disablePortNumber = Dynamicweb.Configuration.SystemConfiguration.Instance.GetValue("/Globalsettings/System/http/DisableBaseHrefPort") == "True";
            var portString = (Context.Current.Request.Url.IsDefaultPort || disablePortNumber) ? string.Empty : string.Format(":{0}", Context.Current.Request.Url.Port);
            var pageId = Dynamicweb.Context.Current.Request["ID"] == null ? string.Empty : string.Format("ID={0}&", Dynamicweb.Context.Current.Request["ID"]);

            return string.Format("{2}://{0}{3}/Default.aspx?{1}{5}={4}", Context.Current.Request.Url.Host, pageId, Context.Current.Request.Url.Scheme, portString, order.Id, OrderIdRequestName);
        }

        private static string ContinueUrl(Order order)
        {
            return string.Format("{0}&QuickPayState=Ok", BaseUrl(order));
        }

        private static string CancelUrl(Order order)
        {
            return string.Format("{0}&QuickPayState=Cancel", BaseUrl(order));
        }

        private static string CallbackUrl(Order order)
        {
            return string.Format("{0}&QuickPayState=Callback", BaseUrl(order));
        }

        private string AuthorizeMD5(Order order)
        {
            var valuesConcatinated = string.Format(
                "{0}{1}{2}{3}{4}{5}{6}{7}{8}{9}{10}{11}{12}{13}",
                Protocol,
                AuthorizeMsgType,
                Merchant,
                LanguageCode,
                order.Id,
                order.Price.PricePIP,
                order.Price.Currency.Code,
                ContinueUrl(order),
                CancelUrl(order),
                CallbackUrl(order),
                CardTypeLock,
                Context.Current.Request.UserHostAddress,
                (TestMode ? "1" : "0"),
                SecretKey
            );

            return StringHelper.Md5HashToString(valuesConcatinated);
        }

        #endregion

        /// <summary>
        /// Starts order checkout procedure
        /// </summary>
        /// <param name="order">Order to be checked out</param>
        /// <returns>String representation of template output</returns>
        public override string StartCheckout(Order order)
        {
            try
            {
                LogEvent(order, "Checkout started");
                var formValues = new Dictionary<string, string> {
                    { "protocol", Protocol },
                    { "msgtype", AuthorizeMsgType },
                    { "merchant", Merchant },
                    { "language", LanguageCode },
                    { "ordernumber", order.Id },
                    { "amount", order.Price.PricePIP.ToString() },
                    { "currency", order.Price.Currency.Code },
                    { "continueurl", ContinueUrl(order) },
                    { "cancelurl", CancelUrl(order) },
                    { "callbackurl", CallbackUrl(order) },
                    { "cardtypelock", CardTypeLock },
                    { "ipaddress", Context.Current.Request.UserHostAddress },
                    { "testmode", TestMode ? "1" : "0" },
                    { "md5check", AuthorizeMD5(order) }
                };

                switch (postMode)
                {
                    case PostModes.Auto:
                        LogEvent(order, "Autopost to QuickPay");
                        SubmitForm("https://secure.quickpay.dk/form/", formValues);

                        return string.Empty;

                    case PostModes.Template:
                        LogEvent(order, "Render template");

                        var formTemplate = new Template(PostTemplate);
                        foreach (var formValue in formValues)
                        {
                            formTemplate.SetTag(string.Format("QuickPay3.{0}", formValue.Key), formValue.Value);
                        }

                        return formTemplate.Output();

                    default:
                        LogError(order, "Unhandled post mode: '{0}'", postMode);
                        return PrintErrorTemplate(order);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        /// <summary>
        /// Redirects user to QuickPay Checkout step
        /// </summary>
        /// <param name="order">Order for processing</param>
        /// <returns>String representation of template output</returns>
        public override string Redirect(Order order)
        {
            try
            {
                LogEvent(null, "Redirected to QuickPay CheckoutHandler");

                var quickPayState = Context.Current.Request["QuickPayState"];
                if (quickPayState == null)
                {
                    LogError(null, "The request variable 'QuickPayState' is not present");
                    return string.Empty;
                }

                switch (Converter.ToString(quickPayState))
                {
                    case "Ok":
                        return StateOk(order);

                    case "Cancel":
                        return StateCancel(order);

                    case "Callback":
                        Callback(order);
                        Context.Current.Response.End();
                        return null;

                    default:
                        LogError(order, "Unknown QuickPayState: '{0}'", Context.Current.Request["QuickPayState"]);
                        return PrintErrorTemplate(order);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {
                return string.Empty;
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unhandled exception with message: {0}", ex.Message);
                return PrintErrorTemplate(order);
            }
        }

        private string StateOk(Order order)
        {
            LogEvent(order, "State ok");

            var startTime = DateTime.Now;
            while (!order.Complete)
            {
                if (DateTime.Now.Subtract(startTime).TotalSeconds > 120)
                {
                    break;
                }

                if (Cache.Current[orderCacheKey + order.Id] != null)
                {
                    order = (Order)Cache.Current[orderCacheKey + order.Id];
                    break;
                }

                System.Threading.Thread.Sleep(500);
            }

            lock (lockObject)
            {
                if (order.Complete)
                {
                    RedirectToCart(order);
                    return null;
                }

                if (Common.Context.Cart == null)
                {
                    CheckoutDone(order);
                }
            }

            LogError(order, "Called State ok, but order is not set complete - this should have happened in the callback.");

            return PrintErrorTemplate(order);
        }

        private string StateCancel(Order order)
        {
            LogEvent(order, "State cancel");

            lock (lockObject)
            {
                order.TransactionStatus = "Cancelled";
                Services.Orders.Save(order);

                CheckoutDone(order);
            }

            var cancelTemplate = new Template(CancelTemplate);
            var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(cancelTemplate, order, true);

            return cancelTemplate.Output();
        }

        private static void SetOrderSucceeded(Order order, bool success)
        {
            if (success)
            {
                order.TransactionAmount = order.Price.PricePIP / 100d; // it may be don't equal order.Price.Price;
                order.TransactionStatus = "Succeeded";
            }
            else
            {
                order.TransactionStatus = "Failed";
            }

            Services.Orders.Save(order);
        }

        /// <remarks>
        /// These are the fields that QuickPay should send
        /// msgtype  	    /^[a-z]$/  	            Defines which action was performed - Each message type is described in detail later
        /// ordernumber 	/^[a-zA-Z0-9]{4,20}$/ 	A value specified by merchant in the initial request.
        /// amount 	        /^[0-9]{1,10}$/ 	    The amount defined in the request in its smallest unit. In example, 1 EUR is written 100.
        /// currency 	    /^[A-Z]{3}$/ 	        The transaction currency as the 3-letter ISO 4217 alphabetical code.
        /// time 	        /^[0-9]{12}$/ 	        The time of which the message was handled. Format is YYMMDDHHIISS.
        /// state 	        /^[1-9]{1,2}$/ 	        The current state of the transaction. See Appendix D.
        /// qpstat 	        /^[0-9]{3}$/ 	        Return code from QuickPay. See Appendix A.
        /// qpstatmsg 	    /^[\w \-\.]{1,}$/ 	    A message detailing errors and warnings if any.
        /// chstat 	        /^[0-9]{3}$/ 	        Return code from the clearing house. Please refer to the clearing house documentation.
        /// chstatmsg 	    /^[\w \-\.]{1,}$/ 	    A message from the clearing house detailing errors and warnings if any.
        /// merchant 	    /^[\w \-\.]{1,100}$/ 	The QuickPay merchant name
        /// merchantemail 	/^[\w_\-\.\@]{6,}$/ 	The QuickPay merchant email/username
        /// transaction 	/^[0-9]{1,32}$/ 	    The id assigned to the current transaction.
        /// cardtype 	    /^[\w\-]{1,32}$/ 	    The card type used to authorize the transaction
        /// cardnumber 	    /^[\w\s]{,32}$/ 	    A truncated version of the card number - eg. 'XXXX XXXX XXXX 1234'.
        /// md5check 	    /^[a-z0-9]{32}$/ 	    A MD5 checksum to ensure data integrity. See Section 3.1 for more information.
        /// </remarks>
        private void Callback(Order order)
        {
            LogEvent(order, "Callback started");

            lock (lockObject)
            {
                var requiredFields = new List<string> {
                    "msgtype",
                    "ordernumber",
                    "amount",
                    "currency",
                    "time",
                    "state",
                    "qpstat",
                    "qpstatmsg",
                    "chstat",
                    "chstatmsg",
                    "merchant",
                    "merchantemail",
                    "transaction",
                    "cardtype",
                    "cardnumber",
                    "cardhash",
                    "acquirer",
                    "splitpayment",
                    "fraudprobability",
                    "fraudremarks",
                    "fraudreport",
                    "fee",
                    "md5check"
                };

                var quickpayResponse = Context.Current.Request.Form.AllKeys.ToDictionary(x => x.ToLower(), x => Converter.ToString(Context.Current.Request[x]));

                var errorOccured = false;

                foreach (var key in requiredFields.Where(x => !quickpayResponse.ContainsKey(x) || quickpayResponse[x] == null))
                {
                    LogError(order, "The expected parameter from QuickPay '{0}' was not send", key);
                    errorOccured = true;
                }

                if (quickpayResponse["msgtype"] != AuthorizeMsgType)
                {
                    LogError(order, "The message type is not valid: '{0}'", quickpayResponse["msgtype"]);
                    errorOccured = true;
                }

                if (quickpayResponse["ordernumber"] != order.Id)
                {
                    LogError(
                        order,
                        "The ordernumber returned from callback does not match with the ordernumber set on the order: Callback: '{0}', order: '{1}'",
                        quickpayResponse["ordernumber"], order.Id
                    );

                    errorOccured = true;
                }

                if (quickpayResponse["amount"] != order.Price.PricePIP.ToString())
                {
                    LogError(
                        order,
                        "The amount returned from callback does not match the amount set on the order: Callback: {0}, order: {1}",
                        quickpayResponse["amount"], order.Price.PricePIP
                    );

                    errorOccured = true;
                }

                if (quickpayResponse["currency"] != order.Price.Currency.Code)
                {
                    LogError(
                        order,
                        "The currency return from callback does not match the amount set on the order: Callback: {0}, order: {1}",
                        quickpayResponse["currency"], order.Price.Currency.Code
                    );

                    errorOccured = true;
                }

                var calculatedMD5 = StringHelper.Md5HashToString(string.Join(string.Empty, requiredFields.Where(x => x != "md5check").Select(x => quickpayResponse[x])) + SecretKey);
                if (!calculatedMD5.Equals(quickpayResponse["md5check"], StringComparison.CurrentCultureIgnoreCase))
                {
                    LogError(
                        order,
                        "The MD5 checksum returned from callback does not match: Callback: {0}, calculated: {1}",
                        quickpayResponse["md5check"], calculatedMD5
                    );

                    errorOccured = true;
                }

                if (errorOccured)
                {
                    LogError(order, "At least one callback validation error exists - exiting callback routine.");
                }
                else
                {
                    if (quickpayResponse["state"] == "1")
                    {
                        // 1: Authorized
                        switch (quickpayResponse["qpstat"])
                        {
                            // 000  Approved.
                            // 001 	Rejected by clearing house. See field 'chstat' and 'chstatmsg' for further explanation.
                            // 002 	Communication error.
                            // 003 	Card expired.
                            // 004 	Transition is not allowed for transaction current state.
                            // 005 	Authorization is expired.
                            // 006 	Error reported by clearing house.
                            // 007 	Error reported by QuickPay.
                            // 008 	Error in request data.

                            case "000":
                                // Set complete
                                LogEvent(order, "Payment succeeded with transaction number {0}", quickpayResponse["transaction"]);
                                order.TransactionCardType = quickpayResponse["cardtype"];
                                order.TransactionCardNumber = HideCardNumber(quickpayResponse["cardnumber"]);
                                SetOrderComplete(order, quickpayResponse["transaction"]);

                                SetOrderSucceeded(order, true);
                                CheckoutDone(order);
                                break;

                            case "001":
                                LogEvent(order, "Not approved: QuickPay response: 'Rejected by clearing house', qpstat: '001', chstat: '{0}', chstatmsg: '{1}'.", quickpayResponse["chstat"], quickpayResponse["chstatmsg"]);
                                break;

                            default:
                                LogEvent(order, "Not approved: QuickPay response: Status code (qpstat): '{0}', status message (qpstatmsg): '{1}'", quickpayResponse["qpstat"], quickpayResponse["qpstatmsg"]);
                                break;
                        }
                    }
                    else
                    {
                        var stateNames = new Dictionary<string, string> {
                        { "1", "Authorized" },
                        { "3", "Captured" },
                        { "5", "Cancelled" },
                        { "7", "Refunded" },
                        { "9", "Subscribed" }
                    };
                        var stateName = stateNames.ContainsKey(quickpayResponse["state"]) ? stateNames[quickpayResponse["state"]] : string.Format("Unknown (code: '{0}')", quickpayResponse["state"]);
                        LogError(
                            order,
                            "QuickPay returned unexpected state: '{0}', state code: '{1}', status code (qpstat): '{2}', status message (qpstatmsg): '{3}'.",
                            stateName, quickpayResponse["state"], quickpayResponse["qpstat"], quickpayResponse["qpstatmsg"]
                        );
                    }
                }

                if (!order.Complete)
                {
                    SetOrderSucceeded(order, false);
                }

                Cache.Current.Set(orderCacheKey + order.Id, order, new CacheItemPolicy { AbsoluteExpiration = DateTime.Now.AddMinutes(10) });
            }
        }

        private string PrintErrorTemplate(Order order)
        {
            LogEvent(order, "Printing error template");

            var errorTemplate = new Template(ErrorTemplate);
            var orderRenderer = new Dynamicweb.Ecommerce.Frontend.Renderer();
            orderRenderer.RenderOrderDetails(errorTemplate, order, true);

            return errorTemplate.Output();
        }

        #region IDropDownOptions

        /// <summary>
        /// Gets options according to behavior mode
        /// </summary>
        /// <param name="behaviorMode"></param>
        /// <returns>Key-value pairs of settings</returns>
        public Hashtable GetOptions(string behaviorMode)
        {
            try
            {
                switch (behaviorMode)
                {
                    case "Post mode":
                        return new Hashtable {
                            { "Auto", Translate.Translate("Auto post (does not use the template)") },
                            { "Template", Translate.Translate("Render template") }
                        };

                    case "Card type":
                        var cardTypes = new Dictionary<string, string> {
                            { "creditcard", "Payments to any card" },
                            { "american-express", "American Express" },
                            { "american-express-dk", "American Express (Danish card)" },
                            { "dankort", "Dankort" },
                            { "danske-dk", "Danske Netbank" },
                            { "diners", "Diners Club" },
                            { "diners-dk", "Diners Club (Danish card)" },
                            { "edankort", "eDankort" },
                            { "fbg1886", "Forbrugsforeningen af 1886" },
                            { "jcb", "JCB" },
                            { "mastercard", "Mastercard" },
                            { "mastercard-dk", "Mastercard (Danish card)" },
                            { "mastercard-debet-dk", "Mastercard-Debet" },
                            { "nordea-dk", "Nordea Netbank" },
                            { "visa", "Visa" },
                            { "visa-dk", "Visa (Danish card)" },
                            { "visa-electron", "Visa Electron" },
                            { "visa-electron-dk", "Visa Electron (Danish card)" },
                            { "paypal", "PayPal" },
                            { "sofort", "Sofort" },
                            { "viabill", "ViaBill" },
                            { "3d-creditcard", "3D-secure payments" },
                            { "3d-jcb", "JCB 3D-Secure" },
                            { "3d-maestro", "Maestro 3D-Secure" },
                            { "3d-maestro-dk", "Maestro 3D-Secure (Danish card)" },
                            { "3d-mastercard", "Mastercard 3D-Secure" },
                            { "3d-mastercard-dk", "Mastercard 3D-Secure (Danish card)" },
                            { "3d-mastercard-debet-dk", "Mastercard-Debet 3D-Secure (Danish card)" },
                            { "3d-visa", "Visa 3D-Secure" },
                            { "3d-visa-dk", "Visa 3D-Secure (Danish card)" },
                            { "3d-visa-electron", "Visa Electron 3D-Secure" },
                            { "3d-visa-electron-dk", "Visa Electron 3D-Secure (Danish card)" }
                        }.ToDictionary(
                            x => x.Key,
                            x => Translate.Translate(x.Value)
                        );

                        var returnHashTable = new Hashtable();
                        foreach (var key in cardTypes.Keys)
                        {
                            returnHashTable.Add(key, cardTypes[key]);
                        }

                        return returnHashTable;

                    default:
                        throw new ArgumentException(string.Format("Unknown dropdown name: '{0}'", behaviorMode));
                }

            }
            catch (Exception ex)
            {
                LogError(null, ex, "Unhandled exception with message: {0}", ex.Message);
                return null;
            }
        }

        #endregion

        #region IRemoteCapture

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order)
        {
            return Capture(order, order.Price.PricePIP, false);
        }

        /// <summary>
        /// Send capture request to transaction service
        /// </summary>
        /// <param name="order">Order to be captured</param>
        /// <param name="amount">Amount to be transferred. The amount must be given in smallest possible unit, e.g. cents 100 = USD 1.</param>
        /// <param name="final">Shows that this step is final</param>
        /// <returns>Response from transaction service</returns>
        public OrderCaptureInfo Capture(Order order, long amount, bool final)
        {
            try
            {
                if (order == null)
                {
                    LogError(null, "Order not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Order not set");
                }
                else if (string.IsNullOrEmpty(order.Id))
                {
                    LogError(null, "Order id not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Order id not set");
                }
                else if (string.IsNullOrEmpty(order.TransactionNumber))
                {
                    LogError(null, "Transaction number not set");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Transaction number not set");
                }
                else if (order.Price.PricePIP < amount)
                {
                    LogError(null, "Amount to capture should be less of order total");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Amount to capture should be less of order total");
                }

                var request = WebRequest.Create("https://secure.quickpay.dk/api");
                var splitpayment = order.Price.PricePIP != amount;

                var httpContent = Encoding.UTF8.GetBytes(
                    string.Format(
                        "protocol={0}&msgtype={1}&merchant={2}&amount={3}&splitpayment={4}&finalize={5}&transaction={6}&md5check={7}",
                        Protocol, CaptureMsgType, Merchant, amount, splitpayment ? "1" : "0", final ? "1" : "0", order.TransactionNumber, CaptureMD5(order, amount, splitpayment, final)
                    )
                );

                request.Method = "POST";
                request.ContentType = "application/x-www-form-urlencoded";
                request.ContentLength = httpContent.Length;

                string responseText;
                try
                {
                    using (var stream = request.GetRequestStream())
                    {
                        stream.Write(httpContent, 0, httpContent.Length);
                    }

                    var response = (HttpWebResponse)request.GetResponse();
                    LogEvent(
                        order,
                        "Remote capture response: HttpStatusCode = {0}, HttpStatusDescription = {1}",
                        response.StatusCode, response.StatusDescription
                    );

                    var reader = new StreamReader(response.GetResponseStream());
                    responseText = reader.ReadToEnd();
                    LogEvent(order, "Remote capture ResponseText: {0}", responseText);
                }
                catch (Exception ex)
                {
                    LogError(order, ex, "Unable to make http request to QuickPay: {0}", ex.Message);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Unable to make http request to QuickPay");
                }

                var xml = new XmlDocument();
                xml.LoadXml(responseText);

                var rootNode = xml["response"];
                if (rootNode == null)
                {
                    LogError(order, "Unable to read response from QuickPay capture");
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Unable to read response from QuickPay capture");
                }

                var requiredFields = new List<string> {
                "msgtype",
                "ordernumber",
                "amount",
                "currency",
                "time",
                "state",
                "qpstat",
                "qpstatmsg",
                "chstat",
                "chstatmsg",
                "merchant",
                "transaction",
                "md5check"
            };

                var quickPayResponse = rootNode.ChildNodes.OfType<XmlNode>().ToDictionary(x => x.LocalName, x => rootNode[x.LocalName] == null ? null : rootNode[x.LocalName].InnerText);

                foreach (var key in requiredFields.Where(x => !quickPayResponse.ContainsKey(x) || quickPayResponse[x] == null))
                {
                    LogError(order, "The expected parameter from QuickPay '{0}' was not send", key);
                    return new OrderCaptureInfo(
                        OrderCaptureInfo.OrderCaptureState.Failed,
                        string.Format("The expected parameter from QuickPay '{0}' was not send", key)
                    );
                }

                if (quickPayResponse["msgtype"] != CaptureMsgType)
                {
                    LogError(
                        order,
                        "Response expected: 'capture'. Response received: '{0}'.",
                        quickPayResponse["msgtype"]
                    );
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Wrong msgtype received");
                }

                var calculatedMD5 = StringHelper.Md5HashToString(string.Join(string.Empty, quickPayResponse.Where(x => x.Key != "md5check").Select(x => x.Value).ToArray()) + SecretKey);
                if (!calculatedMD5.Equals(quickPayResponse["md5check"], StringComparison.CurrentCultureIgnoreCase))
                {
                    LogError(
                        order,
                        "The MD5 checksum returned from callback does not match: Callback: {0}, calculated: {1}",
                        quickPayResponse["md5check"], calculatedMD5
                    );
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "The MD5 checksum returned from callback does not match");
                }

                if (quickPayResponse["state"] == "3" && quickPayResponse["qpstat"] == "000")
                {
                    if (final)
                    {
                        LogEvent(order, String.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture(final)", amount / 100f), DebuggingInfoType.CaptureResult);
                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Split capture successful");
                    }
                    else
                    {
                        LogEvent(order, "Capture successful", DebuggingInfoType.CaptureResult);
                        return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Success, "Capture successful");
                    }
                }
                else if (quickPayResponse["state"] == "1" && quickPayResponse["qpstat"] == "000" && splitpayment)
                {
                    LogEvent(order, String.Format("Message=\"{0}\" Amount=\"{1:f2}\"", "Split capture", amount / 100f), DebuggingInfoType.CaptureResult);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Split, "Split capture successful");
                }
                else
                {
                    LogEvent(order, "Incorrect state and state code received from QuickPay: state: {0}, qpstat: {1}", quickPayResponse["state"], quickPayResponse["qpstat"], DebuggingInfoType.CaptureResult);
                    return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Incorrect state and state code received from QuickPay");
                }
            }
            catch (Exception ex)
            {
                LogError(order, ex, "Unexpected error during capture: {0}", ex.Message);
                return new OrderCaptureInfo(OrderCaptureInfo.OrderCaptureState.Failed, "Unexpected error during capture");
            }
        }

        /// <summary>
        /// Shows if capture supported
        /// </summary>
        /// <param name="order">This object is not used in current implementation</param>
        /// <returns>This method always return 'true' value</returns>
        public bool CaptureSupported(Order order)
        {
            return true;
        }

        /// <summary>
        /// Shows if partial capture of the order supported
        /// </summary>
        /// <param name="order">Instance of order</param>
        /// <returns>True, if partial capture of the order is supported</returns>
        public bool SplitCaptureSupported(Order order)
        {
            return order.TransactionCardType == "dankort";
        }

        private string CaptureMD5(Order order, long amount, bool splitpayment, bool final)
        {
            var valuesConcatinated = Protocol + CaptureMsgType + Merchant + amount + (splitpayment ? "1" : "0") + (final ? "1" : "0") + order.TransactionNumber + SecretKey;
            return StringHelper.Md5HashToString(valuesConcatinated);
        }

        #endregion
    }
}
